@@ -59,19 +59,46 @@
    containing 01. 
    DCJM 27/9/00.
  */
+
+#if HAVE_STDINT_H
+#  include <stdint.h>
+#endif
+
+#if HAVE_INTTYPES_H
+#  ifndef __STDC_FORMAT_MACROS
+#    define __STDC_FORMAT_MACROS
+#  endif
+#  include <inttypes.h>
+#endif
+
+#ifdef HAVE_STDDEF_H
+#  include <stddef.h>
+#endif
+
 #if (defined(HOSTARCHITECTURE_SPARC) && !defined(INTERPRETED))
 #define POLY_TAGSHIFT    2
 #else
 #define POLY_TAGSHIFT    1
 #endif
 
-#if defined(WINDWOWS_PC)
-#include <windows.h>
-typedef INT_PTR         POLYSIGNED;
-typedef UINT_PTR        POLYUNSIGNED;
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#  include <windows.h>
+#endif
+
+typedef intptr_t        POLYSIGNED;
+typedef uintptr_t       POLYUNSIGNED;
+
+// libpolyml uses printf-style I/O instead of C++ standard IOstreams,
+// so we need specifier to format POLYUNSIGNED/POLYSIGNED values.
+#ifdef PRIuPTR
+#  define POLYUFMT PRIuPTR
+#  define POLYSFMT PRIdPTR
+#elif (defined(_MSC_VER) && (SIZEOF_VOIDP == 8))
+#  define POLYUFMT "llu"
+#  define POLYSFMT "lld"
 #else
-typedef long            POLYSIGNED;
-typedef unsigned long   POLYUNSIGNED;
+#  define POLYUFMT "lu"         // as before.  Cross your fingers.
+#  define POLYSFMT "ld"         // idem.
 #endif
 
 typedef unsigned char   byte;
@@ -184,9 +211,14 @@ SPF 24/1/95
 // Zero bits mean ordinary word object containing addresses or tagged integers.
 #define F_BYTE_OBJ                  0x01  /* byte object (contains no pointers) */
 #define F_CODE_OBJ                  0x02  /* code object (mixed bytes and words) */
-#define F_STACK_OBJ                 0x03  /* stack object - may contain internal pointers */
+#define F_GC_MARK                   0x04  // Used during the GC marking phase
 #define F_NO_OVERWRITE              0x08  /* don't overwrite when loading - mutables only. */
-#define F_NEGATIVE_BIT              0x10  /* sign bit for arbitrary precision ints */
+// This bit is overloaded and has different meanings depending on what other bits are set.
+// For byte objects it is the sign bit for arbitrary precision ints.
+// For other data it indicates either that the object is a profile block or contains
+// information for allocation profiling.
+#define F_NEGATIVE_BIT              0x10  // Sign bit for arbitrary precision ints (byte segs only)
+#define F_PROFILE_BIT               0x10  // Object has a profile pointer (word segs only)
 #define F_WEAK_BIT                  0x20  /* object contains weak references to option values. */
 #define F_MUTABLE_BIT               0x40  /* object is mutable */
 #define F_PRIVATE_GC_BIT            0x80  /* object is (pointer or depth) tombstone */
@@ -194,10 +226,10 @@ SPF 24/1/95
 
 #define _OBJ_BYTE_OBJ                _TOP_BYTE(0x01)  /* byte object (contains no pointers) */
 #define _OBJ_CODE_OBJ                _TOP_BYTE(0x02)  /* code object (mixed bytes and words) */
-#define _OBJ_STACK_OBJ               _TOP_BYTE(0x03)  /* stack object - may contain internal pointers */
-
+#define _OBJ_GC_MARK                 _TOP_BYTE(0x04)  // Mark bit
 #define _OBJ_NO_OVERWRITE            _TOP_BYTE(0x08)  /* don't overwrite when loading - mutables only. */
 #define _OBJ_NEGATIVE_BIT            _TOP_BYTE(0x10)  /* sign bit for arbitrary precision ints */
+#define _OBJ_PROFILE_BIT             _TOP_BYTE(0x10)  /* sign bit for arbitrary precision ints */
 #define _OBJ_WEAK_BIT                _TOP_BYTE(0x20)
 #define _OBJ_MUTABLE_BIT             _TOP_BYTE(0x40)  /* object is mutable */
 #define _OBJ_PRIVATE_GC_BIT          _TOP_BYTE(0x80)  /* object is (pointer or depth) tombstone */
@@ -217,9 +249,9 @@ inline byte GetTypeBits(POLYUNSIGNED L)             { return (byte)(L >> OBJ_PRI
 inline POLYUNSIGNED OBJ_OBJECT_LENGTH(POLYUNSIGNED L)   { return L & _OBJ_PRIVATE_LENGTH_MASK; }
 inline bool OBJ_IS_BYTE_OBJECT(POLYUNSIGNED L)          { return (GetTypeBits(L) == F_BYTE_OBJ); }
 inline bool OBJ_IS_CODE_OBJECT(POLYUNSIGNED L)          { return (GetTypeBits(L) == F_CODE_OBJ); }
-inline bool OBJ_IS_STACK_OBJECT(POLYUNSIGNED L)         { return (GetTypeBits(L) == F_STACK_OBJ); }
 inline bool OBJ_IS_NO_OVERWRITE(POLYUNSIGNED L)         { return ((L & _OBJ_NO_OVERWRITE) != 0); }
 inline bool OBJ_IS_NEGATIVE(POLYUNSIGNED L)             { return ((L & _OBJ_NEGATIVE_BIT) != 0); }
+inline bool OBJ_HAS_PROFILE(POLYUNSIGNED L)             { return ((L & _OBJ_PROFILE_BIT) != 0); }
 inline bool OBJ_IS_MUTABLE_OBJECT(POLYUNSIGNED L)       { return ((L & _OBJ_MUTABLE_BIT) != 0); }
 inline bool OBJ_IS_WEAKREF_OBJECT(POLYUNSIGNED L)       { return ((L & _OBJ_WEAK_BIT) != 0); }
 
@@ -245,21 +277,6 @@ inline PolyObject *ObjCodePtrToPtr(byte *cp)
     return (PolyObject *)((byte *)wp - byte_offset);
 }
 
-/* We have to very careful here, because the RTS code segments
-   are malformed - they don't contain the usual count of the
-   number of constants.  This means we can't just use a combination
-   of the OBJ_CODEPTR_TO_PTR and OBJ_PTR_TO_CONSTS_PTR macros.
-    SPF 26/7/96
-*/
-
-inline void OBJ_CODEPTR_TO_CONSTS_PTR(PolyWord cp, PolyWord * &xp)
-{
-    POLYUNSIGNED *wp = (POLYUNSIGNED *)(cp.AsUnsigned() & ~(sizeof(PolyWord)-1));
-    while (*wp != 0) wp++;
-    xp = (PolyWord*)wp;
-}
-
-
 /* Don't need to worry about whether shift is signed, 
    because OBJ_PRIVATE_USER_FLAGS_MASK removes the sign bit.
    We don't want the GC bit (which should be 0) anyway.
@@ -277,6 +294,10 @@ inline POLYUNSIGNED OBJ_SET_POINTER(PolyObject *pt) { return ((POLYUNSIGNED)pt >
 inline bool OBJ_IS_DEPTH(POLYUNSIGNED L)  { return (L & _OBJ_PRIVATE_DEPTH_MASK) == _OBJ_PRIVATE_DEPTH_MASK; }
 inline POLYUNSIGNED OBJ_GET_DEPTH(POLYUNSIGNED L) { return L & ~_OBJ_PRIVATE_DEPTH_MASK; }
 inline POLYUNSIGNED OBJ_SET_DEPTH(POLYUNSIGNED n) { return n | _OBJ_PRIVATE_DEPTH_MASK; }
+// or sharing chain
+inline bool OBJ_IS_CHAINED(POLYUNSIGNED L)  { return (L & _OBJ_PRIVATE_DEPTH_MASK) == _OBJ_PRIVATE_DEPTH_MASK; }
+inline PolyObject *OBJ_GET_CHAIN(POLYUNSIGNED L) { return (PolyObject*)(( L & ~_OBJ_PRIVATE_DEPTH_MASK) <<2); }
+inline POLYUNSIGNED OBJ_SET_CHAIN(PolyObject *pt) { return ((POLYUNSIGNED)pt >> 2) | _OBJ_PRIVATE_DEPTH_MASK; }
 
 // An object i.e. a piece of allocated memory in the heap.  In the simplest case this is a
 // tuple, a list cons cell, a string or a ref.  Every object has a length word in the word before
@@ -300,7 +321,6 @@ public:
 
     bool IsByteObject(void) const { return OBJ_IS_BYTE_OBJECT(LengthWord()); }
     bool IsCodeObject(void) const { return OBJ_IS_CODE_OBJECT(LengthWord()); }
-    bool IsStackObject(void) const { return OBJ_IS_STACK_OBJECT(LengthWord()); }
     bool IsWordObject(void) const { return OBJ_IS_WORD_OBJECT(LengthWord()); }
     bool IsMutable(void) const { return OBJ_IS_MUTABLE_OBJECT(LengthWord()); }
     bool IsWeakRefObject(void) const { return OBJ_IS_WEAKREF_OBJECT(LengthWord()); }
@@ -310,6 +330,10 @@ public:
     PolyObject *GetForwardingPtr(void) const { return OBJ_GET_POINTER(LengthWord()); }
     void SetForwardingPtr(PolyObject *newp) { ((PolyWord*)this)[-1] = PolyWord::FromUnsigned(OBJ_SET_POINTER(newp)); }
 
+    bool ContainsShareChain(void) const { return OBJ_IS_CHAINED(LengthWord()); }
+    PolyObject *GetShareChain(void) const { return OBJ_GET_CHAIN(LengthWord()); }
+    void SetShareChain(PolyObject *newp) { ((PolyWord*)this)[-1] = PolyWord::FromUnsigned(OBJ_SET_CHAIN(newp)); }
+
     bool ContainsNormalLengthWord(void) const { return OBJ_IS_LENGTH(LengthWord()); }
 
     // Find the start of the constant section for a piece of code.
@@ -367,29 +391,11 @@ public:
     PolyWord    ex_id; /* Exc identifier */
     PolyWord    ex_name;/* Exc name */
     PolyWord    arg; /* Exc arguments */
+    PolyWord    ex_location; // Location of "raise".  Always zero for RTS exceptions.
 };
 
 typedef PolyException poly_exn;
 
-/*********************************************************************
- *
- * Stack object 
- *
- *********************************************************************/
-class StackObject: public PolyObject {
-public:
-    /* space available */
-    POLYUNSIGNED    p_space;
-    POLYCODEPTR     p_pc;
-    /* stack pointer */
-    PolyWord        *p_sp;
-    /* handler pointer */
-    PolyWord        *p_hr;
-    /* number of checked registers */
-    POLYUNSIGNED    p_nreg;
-    PolyWord        p_reg[1];
-};
-
 /* 
  * Stream tokens are pointers to UNTAGGED C integers 
  * (stored in byte objects) 
@@ -397,7 +403,7 @@ public:
 class StreamToken: public PolyObject
 {
 public:
-    POLYUNSIGNED    streamNo;
+    unsigned    streamNo;
 };
 
 /***********************************************************************
@@ -451,12 +457,11 @@ typedef enum { NoMoreChildren,CannotCreate,CreatedOk } CStatus ;
 #define NONE_VALUE      (TAGGED(0))
 /* SOME x is represented by a single word cell containing x. */
 
-#if (defined(WIN32))
+#if (defined(_WIN32))
 /* Windows doesn't include 0x in %p format. */
 #define ZERO_X  "0x"
 #else
 #define ZERO_X  ""
 #endif
 
-
 #endif
