@@ -21,10 +21,12 @@
 
 */
 
-#ifdef WIN32
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#elif defined(_WIN32)
 #include "winconfig.h"
 #else
-#include "config.h"
+#error "No configuration file"
 #endif
 
 #ifdef HAVE_STDIO_H
@@ -72,8 +74,9 @@
 #define arg3    (pc[4] + pc[5]*256)
 #define arg4    (pc[6] + pc[7]*256)
 
-#define True    TAGGED(1)
-#define False   TAGGED(0)
+const PolyWord True = TAGGED(1);
+const PolyWord False = TAGGED(0);
+const PolyWord Zero = TAGGED(0);
 
 #define CHECKED_REGS 2
 #define UNCHECKED_REGS 0
@@ -113,7 +116,7 @@ public:
     // Create a task data object.
     virtual MDTaskData *CreateTaskData(void) { return new MDTaskData(); }
 
-    virtual void InitStackFrame(TaskData *taskData, Handle stack, Handle proc, Handle arg);
+    virtual void InitStackFrame(TaskData *taskData, StackSpace *space, Handle proc, Handle arg);
     // Switch to Poly and return with the io function to call.
     virtual int SwitchToPoly(TaskData *taskData);
     virtual void SetForRetry(TaskData *taskData, int ioCall) {} // Nothing to do
@@ -142,12 +145,12 @@ public:
     virtual void CallCodeTupled(TaskData *taskData) { ASSERT(false); }
 };
 
-void Interpreter::InitStackFrame(TaskData *taskData, Handle stackh, Handle proc, Handle arg)
+void Interpreter::InitStackFrame(TaskData *taskData, StackSpace *space, Handle proc, Handle arg)
 /* Initialise stack frame. */
 {
-    StackObject *stack = (StackObject *)DEREFWORDHANDLE(stackh);
+    StackObject *stack = (StackObject *)space->stack();
     PolyObject *closure = DEREFWORDHANDLE(proc);
-    POLYUNSIGNED stack_size = stack->Length();
+    POLYUNSIGNED stack_size = space->spaceSize();
     stack->p_space = OVERFLOW_STACK_SIZE;
     stack->p_pc = *(byte**)(closure);
     stack->p_sp  = (PolyWord*)stack + stack_size-3; /* sp */
@@ -162,7 +165,7 @@ void Interpreter::InitStackFrame(TaskData *taskData, Handle stackh, Handle proc,
     stack->p_sp--;
     *(stack->p_sp) = PolyWord::FromStackAddr(stack->p_sp);
     *(--stack->p_sp) = SPECIAL_PC_END_THREAD; /* Default return address. */
-    *(--stack->p_sp) = TAGGED(0); /* Default handler. */
+    *(--stack->p_sp) = Zero; /* Default handler. */
     stack->p_hr = stack->p_sp;
 
     /* If this function takes an argument store it on the stack. */
@@ -186,8 +189,8 @@ void Interpreter::InterruptCode(TaskData *taskData)
 void Interpreter::SetException(TaskData *taskData, poly_exn *exc)
 /* Set up the stack of a process to raise an exception. */
 {
-    taskData->stack->p_reg[1] = TAGGED(INSTR_raise_ex);
-    *(--taskData->stack->p_sp) = (PolyWord)exc; /* push exception data */
+    taskData->stack->stack()->p_reg[1] = TAGGED(INSTR_raise_ex);
+    *(--taskData->stack->stack()->p_sp) = (PolyWord)exc; /* push exception data */
 }
 
 int Interpreter::SwitchToPoly(TaskData *taskData)
@@ -205,8 +208,7 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
 
     RESTART: /* Load or reload the registers and run the code. */
 
-    if (taskData->allocPointer <= taskData->allocLimit + storeWords ||
-        (userOptions.debug & DEBUG_FORCEGC))
+    if (taskData->allocPointer <= taskData->allocLimit + storeWords)
     {
         if (taskData->allocPointer < taskData->allocLimit)
             Crash ("Bad length in heap overflow trap");
@@ -219,10 +221,10 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
             taskData->allocPointer += storeWords;
     }
     storeWords = 0;
-    sp = taskData->stack->p_sp; /* Reload these. */
-    pc = taskData->stack->p_pc;
-    li = UNTAGGED(taskData->stack->p_reg[1]);
-    sl = (PolyWord*)taskData->stack+OVERFLOW_STACK_SIZE;
+    sp = taskData->stack->stack()->p_sp; /* Reload these. */
+    pc = taskData->stack->stack()->p_pc;
+    li = (unsigned)UNTAGGED(taskData->stack->stack()->p_reg[1]);
+    sl = (PolyWord*)taskData->stack->stack()+OVERFLOW_STACK_SIZE;
 
     if (li != 256) goto RETRY; /* Re-execute instruction if necessary. */
 
@@ -234,21 +236,27 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
         RETRY:
 //      sprintf(buff, "PC=%x, i=%x\n", pc, li);
 //      OutputDebugString(buff);
-        /* Check for stack overflow and interrupts. These can be done less
-           frequently than every instruction. */
-        if (sp < sl) {
-            taskData->stack->p_sp = sp;
-            taskData->stack->p_pc = pc;
-            taskData->stack->p_reg[1] = TAGGED(li);
+        // Check for stack overflow and interrupts. These can be done less
+        // frequently than every instruction.
+        // Don't do this if we're raising an exception: we may be raising
+        // it because we can't grow the stack.  Once we've processed the
+        // exception and produced any exception trace the stack will have been
+        // reset to the last handler.
+        if (sp < sl && li != INSTR_raise_ex) {
+            taskData->stack->stack()->p_sp = sp;
+            taskData->stack->stack()->p_pc = pc;
+            taskData->stack->stack()->p_reg[1] = TAGGED(li);
+            Handle marker = taskData->saveVec.mark();
             CheckAndGrowStack(taskData, sp);
+            taskData->saveVec.reset(marker);
             goto RESTART;
         }
 
         if (interrupt_requested) {
             interrupt_requested = 0;
-            taskData->stack->p_sp = sp;
-            taskData->stack->p_pc = pc;
-            taskData->stack->p_reg[1] = TAGGED(li);
+            taskData->stack->stack()->p_sp = sp;
+            taskData->stack->stack()->p_pc = pc;
+            taskData->stack->stack()->p_reg[1] = TAGGED(li);
             return -1;
         }
 
@@ -271,23 +279,23 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
         case INSTR_jump: pc += *pc + 1; break;
 
         case INSTR_push_handler: /* Save the old handler value. */
-            *(--sp) = PolyWord::FromStackAddr(taskData->stack->p_hr); /* Push old handler */
+            *(--sp) = PolyWord::FromStackAddr(taskData->stack->stack()->p_hr); /* Push old handler */
             break;
 
         case INSTR_set_handler: /* Set up a handler */
             *(--sp) = PolyWord::FromCodePtr(pc + *pc + 1); /* Address of handler */
-            taskData->stack->p_hr = sp-1; /*Point to identifier about to be pushed*/
+            taskData->stack->stack()->p_hr = sp-1; /*Point to identifier about to be pushed*/
             pc += 1;
             break;
 
         case INSTR_del_handler: /* Delete handler retaining the result. */
             {
                 PolyWord u = *sp++;
-                sp = taskData->stack->p_hr;
+                sp = taskData->stack->stack()->p_hr;
                 PolyWord *t;
-                PolyWord *endStack = taskData->stack->Offset(taskData->stack->Length());
+                PolyWord *endStack = taskData->stack->top;
                 while ((t = (*sp).AsStackAddr()) < sp || t > endStack) sp++;
-                taskData->stack->p_hr = t;
+                taskData->stack->stack()->p_hr = t;
                 *sp = u;
                 pc += *pc + 1; /* Skip the handler */
                 break;
@@ -312,7 +320,7 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
                 byte *u = pc + *pc + 1;
                 *(--sp) = /* Address of handler */
                     PolyWord::FromCodePtr(u + u[0] + u[1]*256 + 2);
-                taskData->stack->p_hr = sp-1;
+                taskData->stack->stack()->p_hr = sp-1;
                 pc += 1;
                 break;
             }
@@ -321,10 +329,10 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
             {
                 PolyWord u = *sp++;
                 PolyWord *t;
-                sp = taskData->stack->p_hr;
-                PolyWord *endStack = taskData->stack->Offset(taskData->stack->Length());
+                sp = taskData->stack->stack()->p_hr;
+                PolyWord *endStack = taskData->stack->top;
                 while((t = (*sp).AsStackAddr()) < sp || t > endStack) sp++;
-                taskData->stack->p_hr = t;
+                taskData->stack->stack()->p_hr = t;
                 *sp = u;
                 pc += *pc + 1; /* Skip the handler */
                 pc += arg1 + 2;
@@ -427,7 +435,7 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
                         {
                             t = (*sp++).AsStackAddr();
                             PolyWord v = t[1]; /* Arguments. */
-                            if (v != TAGGED(0)) /* No args. */
+                            if (v != Zero) /* No args. */
                             {
                                 PolyWord *vv = v.AsStackAddr();
                                 POLYUNSIGNED u = v.AsObjPtr()->Length(); /* No. of args. */
@@ -483,7 +491,7 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
                             /* Store the length word of a string. */
                             POLYUNSIGNED len = UNTAGGED(*sp++);
                             ((PolyStringObject*)(*sp).AsObjPtr())->length = len;
-                            *sp = TAGGED(0);
+                            *sp = Zero;
                             break; 
                         }
 
@@ -527,7 +535,7 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
                         {
                             PolyWord u = *sp++;
                             if (UNTAGGED_UNSIGNED(u) > sizeof(PolyWord)*8)
-                                *sp = TAGGED(0);
+                                *sp = Zero;
                             else
                                 *sp = TAGGED(UNTAGGED_UNSIGNED(*sp) << UNTAGGED_UNSIGNED(u));
                             break;
@@ -537,7 +545,7 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
                         {
                             PolyWord u = *sp++;
                             if (UNTAGGED_UNSIGNED(u) > sizeof(PolyWord)*8)
-                                *sp = TAGGED(0);
+                                *sp = Zero;
                             else
                                 *sp = TAGGED(UNTAGGED_UNSIGNED(*sp) >> UNTAGGED_UNSIGNED(u));
                             break;
@@ -548,12 +556,12 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
                             PolyWord u = *sp++;
                             if (UNTAGGED_UNSIGNED(u) > sizeof(PolyWord)*8)
                             {
-                                if (UNTAGGED_UNSIGNED(*sp) < 0)
+                                if (UNTAGGED(*sp) < 0)
                                     *sp = TAGGED(-1);
-                                else *sp = TAGGED(0);
+                                else *sp = Zero;
                             }
                             else
-                                *sp = TAGGED(UNTAGGED_UNSIGNED(*sp) >> UNTAGGED_UNSIGNED(u));
+                                *sp = TAGGED(UNTAGGED(*sp) >> UNTAGGED(u));
                             break;
                         }
 
@@ -576,7 +584,7 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
                             POLYUNSIGNED t = UNTAGGED(*sp++);
                             POLYUNSIGNED u = UNTAGGED(*sp++);
                             (*sp).AsCodePtr()[u] = (byte)t;
-                            *sp = TAGGED(0);
+                            *sp = Zero;
                             break; 
                         }
 
@@ -585,7 +593,7 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
                             PolyWord t = *sp++;
                             POLYUNSIGNED u = UNTAGGED(*sp++);
                             (*sp).AsStackAddr()[u] = t;
-                            *sp = TAGGED(0);
+                            *sp = Zero;
                             break;
                         }
 
@@ -595,7 +603,7 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
                             POLYUNSIGNED lengthW = obj->LengthWord();
                             /* Clear the mutable bit. */
                             obj->SetLengthWord(lengthW & ~_OBJ_MUTABLE_BIT);
-                            *sp = TAGGED(0);
+                            *sp = Zero;
                             break;
                         }
 
@@ -605,46 +613,150 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
                         break;
 
                     case POLY_SYS_is_short:
-                        *sp = IS_INT(*sp) ? TAGGED(1) : TAGGED(0); break;
+                        *sp = IS_INT(*sp) ? True : False; break;
 
                     case POLY_SYS_io_operation:
-                        /* The io_operation call has changed between the old Poly
-                           version and the new ML version.  In Poly it took two
-                           parameters, the first always being an empty type.
-                           In ML it takes just one parameter. */
-                        if (*sp == TAGGED(0)) sp++;
-                        *sp = (PolyObject*)IoEntry(UNTAGGED(*sp));
+                        *sp = (PolyObject*)IoEntry((unsigned)UNTAGGED(*sp));
                         break;
 
                     case POLY_SYS_exception_trace:
                         u = *sp; /* Procedure to call. */
                         *(--sp) = PolyWord::FromCodePtr(pc); /* Push a return address. */
-                        *(--sp) = PolyWord::FromStackAddr(taskData->stack->p_hr); /* Push old handler */
+                        *(--sp) = PolyWord::FromStackAddr(taskData->stack->stack()->p_hr); /* Push old handler */
                         *(--sp) = SPECIAL_PC_TRACE_EX; /* Marks exception trace. */
-                        *(--sp) = TAGGED(0); /* Catch everything. */
-                        taskData->stack->p_hr = sp; /* Handler is here. */
+                        *(--sp) = Zero; /* Catch everything. */
+                        taskData->stack->stack()->p_hr = sp; /* Handler is here. */
                         pc = (SPECIAL_PC_TRACE_EX).AsCodePtr(); /* Special return address. */
-                        *(--sp) = TAGGED(0); /* Unit argument to the function. */
+                        *(--sp) = Zero; /* Unit argument to the function. */
                         *(--sp) = u; /* Push the procedure. */
                         goto CALL_CLOSURE;
 
                     case POLY_SYS_is_big_endian: {
                         union { unsigned long wrd; char chrs[sizeof(unsigned long)]; } endian;
                         endian.wrd = 1;
-                        *(--sp) = endian.chrs[0] == 0 ? TAGGED(1) : TAGGED(0);
+                        *(--sp) = endian.chrs[0] == 0 ? True : False;
                         break;
                     }
 
                     case POLY_SYS_bytes_per_word:
                         *(--sp) = TAGGED(sizeof(PolyWord)); break;
  
+                    case POLY_SYS_raisex:
+                        goto RAISE_EXCEPTION;
+
+                    case POLY_SYS_aplus:
+                        {
+                            PolyWord x = sp[0];
+                            PolyWord y = sp[1];
+                            // If they're both short and no overflow.
+                            if (IS_INT(x) && IS_INT(y))
+                            {
+                                POLYSIGNED t = UNTAGGED(x) + UNTAGGED(y);
+                                if (t <= MAXTAGGED && t >= -MAXTAGGED-1)
+                                {
+                                    sp++;
+                                    *sp = TAGGED(t);
+                                    break;
+                                }
+                                else goto FullRTSCall;
+                            }
+                            else goto FullRTSCall;
+                        }
+
+                    case POLY_SYS_aminus:
+                        {
+                            PolyWord x = sp[0];
+                            PolyWord y = sp[1];
+                            // If they're both short and no overflow.
+                            if (IS_INT(x) && IS_INT(y))
+                            {
+                                POLYSIGNED t = UNTAGGED(y) - UNTAGGED(x);
+                                if (t <= MAXTAGGED && t >= -MAXTAGGED-1)
+                                {
+                                    sp++;
+                                    *sp = TAGGED(t);
+                                    break;
+                                }
+                                else goto FullRTSCall;
+                            }
+                            else goto FullRTSCall;
+                        }
+
+                    case POLY_SYS_equala:
+                        {
+                            PolyWord x = sp[0];
+                            PolyWord y = sp[1];
+                            // If either argument is short the values are only equal
+                            // if the words contain the same bit pattern.
+                            if (IS_INT(x) || IS_INT(y))
+                            {
+                                sp++;
+                                *sp = y == x ? True : False;
+                                break;
+                            }
+                            else goto FullRTSCall;
+                        }
+
+                    case POLY_SYS_int_geq:
+                        {
+                            PolyWord x = sp[0];
+                            PolyWord y = sp[1];
+                            if (IS_INT(x) && IS_INT(y))
+                            {
+                                sp++;
+                                *sp = UNTAGGED(y) >= UNTAGGED(x) ? True : False;
+                                break;
+                            }
+                            else goto FullRTSCall;
+                        }
+
+                    case POLY_SYS_int_leq:
+                        {
+                            PolyWord x = sp[0];
+                            PolyWord y = sp[1];
+                            if (IS_INT(x) && IS_INT(y))
+                            {
+                                sp++;
+                                *sp = UNTAGGED(y) <= UNTAGGED(x) ? True : False;
+                                break;
+                            }
+                            else goto FullRTSCall;
+                        }
+
+                    case POLY_SYS_int_gtr:
+                        {
+                            PolyWord x = sp[0];
+                            PolyWord y = sp[1];
+                            if (IS_INT(x) && IS_INT(y))
+                            {
+                                sp++;
+                                *sp = UNTAGGED(y) > UNTAGGED(x) ? True : False;
+                                break;
+                            }
+                            else goto FullRTSCall;
+                        }
+
+                    case POLY_SYS_int_lss:
+                        {
+                            PolyWord x = sp[0];
+                            PolyWord y = sp[1];
+                            if (IS_INT(x) && IS_INT(y))
+                            {
+                                sp++;
+                                *sp = UNTAGGED(y) < UNTAGGED(x) ? True : False;
+                                break;
+                            }
+                            else goto FullRTSCall;
+                        }
+
                     default:
+                    FullRTSCall:
                         // For all the calls that aren't built in ...
                         /* Save the state so that the instruction can be retried if necessary. */
-                        taskData->stack->p_pc = pc; /* Pc value after instruction. */
-                        taskData->stack->p_reg[1] = TAGGED(li); /* Previous instruction. */
-                        taskData->stack->p_sp = sp-1; /* Include the closure address. */
-                        return uu;
+                        taskData->stack->stack()->p_pc = pc; /* Pc value after instruction. */
+                        taskData->stack->stack()->p_reg[1] = TAGGED(li); /* Previous instruction. */
+                        taskData->stack->stack()->p_sp = sp-1; /* Include the closure address. */
+                        return (int)uu;
                     }
                 } /* End of system calls. */
                 else {
@@ -672,7 +784,7 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
                     /* Return from a call to exception_trace when an exception
                        has not been raised. */
                     sp += 1;
-                    taskData->stack->p_hr = (sp[1]).AsStackAddr();
+                    taskData->stack->stack()->p_hr = (sp[1]).AsStackAddr();
                     *sp = result;
                     returnCount = 1;
                     goto RETURN;
@@ -691,15 +803,16 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
 
         case INSTR_raise_ex:
             {
+            RAISE_EXCEPTION:
                 PolyException *exn = (PolyException*)((*sp).AsObjPtr());
-                taskData->stack->p_reg[0] = exn; /* Get exception data */
+                taskData->stack->stack()->p_reg[0] = exn; /* Get exception data */
                 PolyWord exId = exn->ex_id; /* Get exception identifier. */
-                taskData->stack->p_sp = sp; /* Save this in case of trace. */
-                PolyWord *t = taskData->stack->p_hr;  /* First handler */
-                PolyWord *endStack = taskData->stack->Offset(taskData->stack->Length());
+                taskData->stack->stack()->p_sp = sp; /* Save this in case of trace. */
+                PolyWord *t = taskData->stack->stack()->p_hr;  /* First handler */
+                PolyWord *endStack = taskData->stack->top;
                 /* Handlers consist of one or more pairs of identifier and
                    code address, followed by the address of the next handler. */
-                while (*t != TAGGED(0) && *t != exId) {
+                while (*t != Zero && *t != exId) {
                     /* Loop until we find an ELSE handler or one that matches */
                     t += 2; /* Go on to next. */
                     /* If it points into stack it must be a pointer to the next
@@ -715,9 +828,9 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
                     PolyWord *nextHandler = t;
                     Handle marker = taskData->saveVec.mark();
                     try {
-                        ex_tracec(taskData, taskData->saveVec.push(taskData->stack->p_reg[0]),
+                        ex_tracec(taskData, taskData->saveVec.push(taskData->stack->stack()->p_reg[0]),
                                   taskData->saveVec.push(PolyWord::FromStackAddr(nextHandler)));
-					}
+                    }
                     catch (IOException) {
                     }
                     taskData->saveVec.reset(marker);
@@ -726,15 +839,15 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
                 }
                 else if (*t == SPECIAL_PC_END_THREAD)
                     exitThread(taskData);  // Default handler for thread.
-                taskData->stack->p_pc = (*t).AsCodePtr();
+                taskData->stack->stack()->p_pc = (*t).AsCodePtr();
                 /* Now remove this handler. */
                 sp = t;
                 while ((t = (*sp).AsStackAddr()) < sp || t > endStack)
                     sp++;
-                taskData->stack->p_hr = t; /* Restore old handler */
+                taskData->stack->stack()->p_hr = t; /* Restore old handler */
                 sp++; /* Remove that entry. */
-                taskData->stack->p_sp = sp;
-                taskData->stack->p_reg[1] = TAGGED(256); /* Get the next instruction. */
+                taskData->stack->stack()->p_sp = sp;
+                taskData->stack->stack()->p_reg[1] = TAGGED(256); /* Get the next instruction. */
                 goto RESTART; /* Restart in case pc is persistent (??? Still relevant????). */
             }
 
@@ -748,9 +861,9 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
                 if (taskData->allocPointer < taskData->allocLimit)
                 {
                     taskData->allocPointer += storeWords;
-                    taskData->stack->p_sp = sp;
-                    taskData->stack->p_pc = pc;
-                    taskData->stack->p_reg[1] = TAGGED(li);
+                    taskData->stack->stack()->p_sp = sp;
+                    taskData->stack->stack()->p_pc = pc;
+                    taskData->stack->stack()->p_reg[1] = TAGGED(li);
                     goto RESTART;
                 }
                 pc += instrBytes;
@@ -776,9 +889,9 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
                 taskData->allocPointer -= storeWords;
                 if (taskData->allocPointer < taskData->allocLimit) {
                     taskData->allocPointer += storeWords;
-                    taskData->stack->p_sp = sp;
-                    taskData->stack->p_pc = pc;
-                    taskData->stack->p_reg[1] = TAGGED(li);
+                    taskData->stack->stack()->p_sp = sp;
+                    taskData->stack->stack()->p_pc = pc;
+                    taskData->stack->stack()->p_reg[1] = TAGGED(li);
                     goto RESTART;
                 }
                 pc += instrBytes;
@@ -866,7 +979,7 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
             pc += 1;
             break;
 
-        case INSTR_const_nil: *(--sp) = (PolyWord)TAGGED(0); break;
+        case INSTR_const_nil: *(--sp) = Zero; break;
 
         case INSTR_jump_back: pc -= *pc + 1; break;
 
@@ -877,7 +990,7 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
                 break;
             }
 
-        case INSTR_ldexc: *(--sp) = taskData->stack->p_reg[0]; break;
+        case INSTR_ldexc: *(--sp) = taskData->stack->stack()->p_reg[0]; break;
 
         case INSTR_local_b: { PolyWord u = sp[*pc]; *(--sp) = u; pc += 1; break; }
 
@@ -930,7 +1043,7 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
         case INSTR_indirect_5:
             *sp = (*sp).AsObjPtr()->Get(5); break;
 
-        case INSTR_const_0: *(--sp) = TAGGED(0); break;
+        case INSTR_const_0: *(--sp) = Zero; break;
         case INSTR_const_1: *(--sp) = TAGGED(1); break;
         case INSTR_const_2: *(--sp) = TAGGED(2); break;
         case INSTR_const_3: *(--sp) = TAGGED(3); break;
@@ -1053,7 +1166,7 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
                 PolyWord u = *sp++;
                 POLYUNSIGNED uu = UNTAGGED(*sp++);
                 (*sp).AsObjPtr()->Set(uu, u);
-                *sp = TAGGED(0);
+                *sp = Zero;
                 break;
             }
 
@@ -1065,15 +1178,15 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
                 taskData->allocPointer -= storeWords;
                 if (taskData->allocPointer < taskData->allocLimit) {
                     taskData->allocPointer += storeWords;
-                    taskData->stack->p_sp = sp;
-                    taskData->stack->p_pc = pc;
-                    taskData->stack->p_reg[1] = TAGGED(li);
+                    taskData->stack->stack()->p_sp = sp;
+                    taskData->stack->stack()->p_pc = pc;
+                    taskData->stack->stack()->p_reg[1] = TAGGED(li);
                     goto RESTART;
                 }
                 storeWords--;
                 PolyObject *t = (PolyObject*)(taskData->allocPointer+1);
                 t->SetLengthWord(storeWords, F_MUTABLE_BIT);
-                for(; storeWords > 0; ) t->Set(--storeWords, TAGGED(0));
+                for(; storeWords > 0; ) t->Set(--storeWords, Zero);
                 *(--sp) = t; /* Push the address of the container. */
                 pc += 2;
                 break;
@@ -1098,9 +1211,9 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
                 taskData->allocPointer -= storeWords;
                 if (taskData->allocPointer < taskData->allocLimit) {
                     taskData->allocPointer += storeWords;
-                    taskData->stack->p_sp = sp;
-                    taskData->stack->p_pc = pc;
-                    taskData->stack->p_reg[1] = TAGGED(li);
+                    taskData->stack->stack()->p_sp = sp;
+                    taskData->stack->stack()->p_pc = pc;
+                    taskData->stack->stack()->p_reg[1] = TAGGED(li);
                     goto RESTART;
                 }
                 storeWords--;
@@ -1120,7 +1233,7 @@ int Interpreter::SwitchToPoly(TaskData *taskData)
 
         } /* switch */
      } /* for */
-	 return 0;
+     return 0;
 } /* MD_switch_to_poly */
 
 
@@ -1128,83 +1241,83 @@ void Interpreter::CallIO0(TaskData *taskData, Handle(*ioFun)(TaskData *))
 {
     try {
         Handle result = (*ioFun)(taskData);
-        *(taskData->stack->p_sp) = result->Word();
-        taskData->stack->p_reg[1] = TAGGED(256); /* Take next instruction. */
-	}
+        *(taskData->stack->stack()->p_sp) = result->Word();
+        taskData->stack->stack()->p_reg[1] = TAGGED(256); /* Take next instruction. */
+    }
     catch (IOException) {
     }
 }
 
 void Interpreter::CallIO1(TaskData *taskData, Handle(*ioFun)(TaskData *, Handle))
 {
-    Handle funarg = taskData->saveVec.push(taskData->stack->p_sp[1]);
+    Handle funarg = taskData->saveVec.push(taskData->stack->stack()->p_sp[1]);
     try {
         Handle result = (*ioFun)(taskData, funarg);
-        *(++taskData->stack->p_sp) = result->Word();
-        taskData->stack->p_reg[1] = TAGGED(256); /* Take next instruction. */
-	}
+        *(++taskData->stack->stack()->p_sp) = result->Word();
+        taskData->stack->stack()->p_reg[1] = TAGGED(256); /* Take next instruction. */
+    }
     catch (IOException) {
     }
 }
 
 void Interpreter::CallIO2(TaskData *taskData, Handle(*ioFun)(TaskData *, Handle, Handle))
 {
-    Handle funarg1 = taskData->saveVec.push(taskData->stack->p_sp[1]);
-    Handle funarg2 = taskData->saveVec.push(taskData->stack->p_sp[2]);
+    Handle funarg1 = taskData->saveVec.push(taskData->stack->stack()->p_sp[1]);
+    Handle funarg2 = taskData->saveVec.push(taskData->stack->stack()->p_sp[2]);
     try {
         Handle result = (*ioFun)(taskData, funarg1, funarg2);
-        taskData->stack->p_sp += 2;
-        *(taskData->stack->p_sp) = DEREFWORD(result);
-        taskData->stack->p_reg[1] = TAGGED(256); /* Take next instruction. */
-	}
+        taskData->stack->stack()->p_sp += 2;
+        *(taskData->stack->stack()->p_sp) = DEREFWORD(result);
+        taskData->stack->stack()->p_reg[1] = TAGGED(256); /* Take next instruction. */
+    }
     catch (IOException) {
     }
 }
 
 void Interpreter::CallIO3(TaskData *taskData, Handle(*ioFun)(TaskData *, Handle, Handle, Handle))
 {
-    Handle funarg1 = taskData->saveVec.push(taskData->stack->p_sp[1]);
-    Handle funarg2 = taskData->saveVec.push(taskData->stack->p_sp[2]);
-    Handle funarg3 = taskData->saveVec.push(taskData->stack->p_sp[3]);
+    Handle funarg1 = taskData->saveVec.push(taskData->stack->stack()->p_sp[1]);
+    Handle funarg2 = taskData->saveVec.push(taskData->stack->stack()->p_sp[2]);
+    Handle funarg3 = taskData->saveVec.push(taskData->stack->stack()->p_sp[3]);
     try {
         Handle result = (*ioFun)(taskData, funarg1, funarg2, funarg3);
-        taskData->stack->p_sp += 3;
-        *(taskData->stack->p_sp) = DEREFWORD(result);
-        taskData->stack->p_reg[1] = TAGGED(256); /* Take next instruction. */
-	}
+        taskData->stack->stack()->p_sp += 3;
+        *(taskData->stack->stack()->p_sp) = DEREFWORD(result);
+        taskData->stack->stack()->p_reg[1] = TAGGED(256); /* Take next instruction. */
+    }
     catch (IOException) {
     }
 }
 
 void Interpreter::CallIO4(TaskData *taskData, Handle(*ioFun)(TaskData *, Handle, Handle, Handle, Handle))
 {
-    Handle funarg1 = taskData->saveVec.push(taskData->stack->p_sp[1]);
-    Handle funarg2 = taskData->saveVec.push(taskData->stack->p_sp[2]);
-    Handle funarg3 = taskData->saveVec.push(taskData->stack->p_sp[3]);
-    Handle funarg4 = taskData->saveVec.push(taskData->stack->p_sp[4]);
+    Handle funarg1 = taskData->saveVec.push(taskData->stack->stack()->p_sp[1]);
+    Handle funarg2 = taskData->saveVec.push(taskData->stack->stack()->p_sp[2]);
+    Handle funarg3 = taskData->saveVec.push(taskData->stack->stack()->p_sp[3]);
+    Handle funarg4 = taskData->saveVec.push(taskData->stack->stack()->p_sp[4]);
     try {
         Handle result = (*ioFun)(taskData, funarg1, funarg2, funarg3, funarg4);
-        taskData->stack->p_sp += 4;
-        *(taskData->stack->p_sp) = DEREFWORD(result);
-        taskData->stack->p_reg[1] = TAGGED(256); /* Take next instruction. */
-	}
+        taskData->stack->stack()->p_sp += 4;
+        *(taskData->stack->stack()->p_sp) = DEREFWORD(result);
+        taskData->stack->stack()->p_reg[1] = TAGGED(256); /* Take next instruction. */
+    }
     catch (IOException) {
     }
 }
 
 void Interpreter::CallIO5(TaskData *taskData, Handle(*ioFun)(TaskData *, Handle, Handle, Handle, Handle, Handle))
 {
-    Handle funarg1 = taskData->saveVec.push(taskData->stack->p_sp[1]);
-    Handle funarg2 = taskData->saveVec.push(taskData->stack->p_sp[2]);
-    Handle funarg3 = taskData->saveVec.push(taskData->stack->p_sp[3]);
-    Handle funarg4 = taskData->saveVec.push(taskData->stack->p_sp[4]);
-    Handle funarg5 = taskData->saveVec.push(taskData->stack->p_sp[5]);
+    Handle funarg1 = taskData->saveVec.push(taskData->stack->stack()->p_sp[1]);
+    Handle funarg2 = taskData->saveVec.push(taskData->stack->stack()->p_sp[2]);
+    Handle funarg3 = taskData->saveVec.push(taskData->stack->stack()->p_sp[3]);
+    Handle funarg4 = taskData->saveVec.push(taskData->stack->stack()->p_sp[4]);
+    Handle funarg5 = taskData->saveVec.push(taskData->stack->stack()->p_sp[5]);
     try {
         Handle result = (*ioFun)(taskData, funarg1, funarg2, funarg3, funarg4, funarg5);
-        taskData->stack->p_sp += 5;
-        *(taskData->stack->p_sp) = DEREFWORD(result);
-        taskData->stack->p_reg[1] = TAGGED(256); /* Take next instruction. */
-	}
+        taskData->stack->stack()->p_sp += 5;
+        *(taskData->stack->stack()->p_sp) = DEREFWORD(result);
+        taskData->stack->stack()->p_reg[1] = TAGGED(256); /* Take next instruction. */
+    }
     catch (IOException) {
     }
 }
@@ -1212,13 +1325,15 @@ void Interpreter::CallIO5(TaskData *taskData, Handle(*ioFun)(TaskData *, Handle,
 // Return the callback result.  The current ML process (thread) terminates.
 Handle Interpreter::CallBackResult(TaskData *taskData)
 {
-    return taskData->saveVec.push(taskData->stack->p_sp[1]);
+    return taskData->saveVec.push(taskData->stack->stack()->p_sp[1]);
 }
 
 void Interpreter::InitInterfaceVector(void)
 {
     add_word_to_io_area(POLY_SYS_exit, TAGGED(POLY_SYS_exit));
     add_word_to_io_area(POLY_SYS_alloc_store, TAGGED(POLY_SYS_alloc_store));
+    add_word_to_io_area(POLY_SYS_alloc_uninit, TAGGED(POLY_SYS_alloc_uninit));
+    add_word_to_io_area(POLY_SYS_raisex, TAGGED(POLY_SYS_raisex));
     add_word_to_io_area(POLY_SYS_chdir, TAGGED(POLY_SYS_chdir));
     add_word_to_io_area(POLY_SYS_get_length, TAGGED(POLY_SYS_get_length));
     add_word_to_io_area(POLY_SYS_get_flags, TAGGED(POLY_SYS_get_flags));
@@ -1234,6 +1349,7 @@ void Interpreter::InitInterfaceVector(void)
     add_word_to_io_area(POLY_SYS_profiler, TAGGED(POLY_SYS_profiler));
     add_word_to_io_area(POLY_SYS_is_short, TAGGED(POLY_SYS_is_short));
 //    add_word_to_io_area(POLY_SYS_raiseexception, TAGGED(POLY_SYS_raiseexception));
+    add_word_to_io_area(POLY_SYS_quotrem, TAGGED(POLY_SYS_quotrem));
     add_word_to_io_area(POLY_SYS_aplus, TAGGED(POLY_SYS_aplus));
     add_word_to_io_area(POLY_SYS_aminus, TAGGED(POLY_SYS_aminus));
     add_word_to_io_area(POLY_SYS_amul, TAGGED(POLY_SYS_amul));
@@ -1258,7 +1374,7 @@ void Interpreter::InitInterfaceVector(void)
     add_word_to_io_area(POLY_SYS_Sub_real, TAGGED(POLY_SYS_Sub_real));
     add_word_to_io_area(POLY_SYS_Mul_real, TAGGED(POLY_SYS_Mul_real));
     add_word_to_io_area(POLY_SYS_Div_real, TAGGED(POLY_SYS_Div_real));
-//    add_word_to_io_area(POLY_SYS_Comp_real, TAGGED(POLY_SYS_Comp_real));
+    add_word_to_io_area(POLY_SYS_Abs_real, TAGGED(POLY_SYS_Abs_real));
     add_word_to_io_area(POLY_SYS_Neg_real, TAGGED(POLY_SYS_Neg_real));
     add_word_to_io_area(POLY_SYS_Repr_real, TAGGED(POLY_SYS_Repr_real));
     add_word_to_io_area(POLY_SYS_conv_real, TAGGED(POLY_SYS_conv_real));
@@ -1318,6 +1434,7 @@ void Interpreter::InitInterfaceVector(void)
     add_word_to_io_area(POLY_SYS_set_string_length, TAGGED(POLY_SYS_set_string_length));
     add_word_to_io_area(POLY_SYS_get_first_long_word, TAGGED(POLY_SYS_get_first_long_word));
     add_word_to_io_area(POLY_SYS_poly_specific, TAGGED(POLY_SYS_poly_specific));
+    add_word_to_io_area(POLY_SYS_bytevec_eq, TAGGED(POLY_SYS_bytevec_eq));
     add_word_to_io_area(POLY_SYS_shrink_stack,     TAGGED(POLY_SYS_shrink_stack));
     add_word_to_io_area(POLY_SYS_code_flags,        TAGGED(POLY_SYS_code_flags));
     add_word_to_io_area(POLY_SYS_shift_right_arith_word, TAGGED(POLY_SYS_shift_right_arith_word));
