@@ -2,7 +2,7 @@
     Title:     Write out a database as a Mach object file
     Author:    David Matthews.
 
-    Copyright (c) 2006-7 David C. J. Matthews
+    Copyright (c) 2006-7, 2011-2 David C. J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -57,6 +57,7 @@
 #include <mach-o/reloc.h>
 #include <mach-o/nlist.h>
 #include <mach-o/ppc/reloc.h>
+#include <mach-o/x86_64/reloc.h>
 
 #ifdef HAVE_STRING_H
 #include <string.h>
@@ -78,12 +79,12 @@
 #include "version.h"
 #include "polystring.h"
 
-// The data section consists of one area beginning with the descriptors
-// and followed by each of the memory sections in turn.  We have to adjust
-// offsets to match that.
+
+// Mach-O seems to require each section to have a discrete virtual address range
+// so we have to adjust various offsets to fit.
 void MachoExport::adjustOffset(unsigned area, POLYUNSIGNED &offset)
 {
-     // Add in the offset.  If sect is memTableEntries it's actually the
+    // Add in the offset.  If sect is memTableEntries it's actually the
     // descriptors so doesn't have any additional offset.
     if (area != memTableEntries)
     {
@@ -98,7 +99,6 @@ void MachoExport::setRelocationAddress(void *p, int32_t *reloc)
 {
     unsigned area = findArea(p);
     POLYUNSIGNED offset = (char*)p - (char*)memTable[area].mtAddr;
-    adjustOffset(area, offset);
     *reloc = offset;
 }
 
@@ -114,10 +114,15 @@ PolyWord MachoExport::createRelocation(PolyWord p, void *relocAddr)
     // with GENERIC_RELOC_VANILLA types.
     struct relocation_info relInfo;
     setRelocationAddress(relocAddr, &relInfo.r_address);
-    relInfo.r_symbolnum = 1; // Section numbers start at 1
+    relInfo.r_symbolnum = addrArea+1; // Section numbers start at 1
     relInfo.r_pcrel = 0;
+#if (SIZEOF_VOIDP == 8)
+    relInfo.r_length = 3; // 8 bytes
+    relInfo.r_type = X86_64_RELOC_UNSIGNED;
+#else
     relInfo.r_length = 2; // 4 bytes
     relInfo.r_type = GENERIC_RELOC_VANILLA;
+#endif
     relInfo.r_extern = 0; // r_symbolnum is a section number.  It should be 1 if we make the IO area a common.
 
     fwrite(&relInfo, sizeof(relInfo), 1, exportFile);
@@ -148,10 +153,15 @@ void MachoExport::ScanConstant(byte *addr, ScanRelocationKind code)
         {
             struct relocation_info reloc;
             setRelocationAddress(addr, &reloc.r_address);
-            reloc.r_symbolnum = 1; // Section numbers start at 1
+            reloc.r_symbolnum = aArea+1; // Section numbers start at 1
             reloc.r_pcrel = 0;
+#if (defined(HOSTARCHITECTURE_X86_64))
+            reloc.r_length = 3; // 8 bytes
+            reloc.r_type = X86_64_RELOC_UNSIGNED;
+#else
             reloc.r_length = 2; // 4 bytes
             reloc.r_type = GENERIC_RELOC_VANILLA;
+#endif
             reloc.r_extern = 0; // r_symbolnum is a section number.  It should be 1 if we make the IO area a common.
 
             for (unsigned i = 0; i < sizeof(PolyWord); i++)
@@ -163,20 +173,33 @@ void MachoExport::ScanConstant(byte *addr, ScanRelocationKind code)
             relocationCount++;
         }
         break;
-#if(defined(HOSTARCHITECTURE_X86) || defined(HOSTARCHITECTURE_X86_64))
+#if (defined(HOSTARCHITECTURE_X86))
      case PROCESS_RELOC_I386RELATIVE:         // 32 bit relative address
         {
-            // We don't need a relocation since everything is in the same segment
-            // but we still need to recalculate the offset.
             unsigned addrArea = findArea(addr);
-            POLYUNSIGNED addrOffset = (char*)addr - (char*)memTable[addrArea].mtAddr;
-            adjustOffset(addrArea, addrOffset);
-            offset -= addrOffset + 4;
-
-            for (unsigned i = 0; i < sizeof(PolyWord); i++)
+            // If it's in the same area we don't need a relocation because the
+            // relative offset will be unchanged.
+            if (addrArea != aArea)
             {
-                addr[i] = (byte)(offset & 0xff);
-                offset >>= 8;
+                struct relocation_info reloc;
+                setRelocationAddress(addr, &reloc.r_address);
+                reloc.r_symbolnum = aArea+1; // Section numbers start at 1
+                reloc.r_pcrel = 1;
+                reloc.r_length = 2; // 4 bytes
+                reloc.r_type = GENERIC_RELOC_VANILLA;
+                reloc.r_extern = 0; // r_symbolnum is a section number.
+                fwrite(&reloc, sizeof(reloc), 1, exportFile);
+                relocationCount++;
+
+                POLYUNSIGNED addrOffset = (char*)addr - (char*)memTable[addrArea].mtAddr;
+                adjustOffset(addrArea, addrOffset);
+                offset -= addrOffset + 4;
+
+                for (unsigned i = 0; i < sizeof(PolyWord); i++)
+                {
+                    addr[i] = (byte)(offset & 0xff);
+                    offset >>= 8;
+                }
             }
         }
         break;
@@ -193,7 +216,7 @@ void MachoExport::ScanConstant(byte *addr, ScanRelocationKind code)
             // relocations for each.  It stores one half of the address in the instruction
             // itself and the other half is carried in a PPC_RELOC_PAIR relocation entry.
             // We need four relocations here in total.
-            reloc.r_symbolnum = 1; // Section numbers start at 1
+            reloc.r_symbolnum = aArea+1; // Section numbers start at 1
             reloc.r_extern = 0; // r_symbolnum is a section number.
             reloc.r_pcrel = 0;
             reloc.r_length = 2; // 4 bytes
@@ -212,7 +235,7 @@ void MachoExport::ScanConstant(byte *addr, ScanRelocationKind code)
 
             // Now the low-order part.
             setRelocationAddress(addr+sizeof(PolyWord), &reloc.r_address);
-            reloc.r_symbolnum = 1; // Section numbers start at 1
+            reloc.r_symbolnum = aArea+1; // Section numbers start at 1
             reloc.r_extern = 0; // r_symbolnum is a section number.
             reloc.r_pcrel = 0;
             reloc.r_length = 2; // 4 bytes
@@ -244,11 +267,15 @@ void MachoExport::ScanConstant(byte *addr, ScanRelocationKind code)
 
 void MachoExport::writeSymbol(const char *symbolName, unsigned char nType, unsigned char nSect, unsigned long offset)
 {
+#if (SIZEOF_VOIDP == 8)
+    struct nlist_64 symbol;
+#else
     struct nlist symbol;
+#endif
     memset(&symbol, 0, sizeof(symbol)); // Zero unused fields
     symbol.n_un.n_strx = stringTable.makeEntry(symbolName);
     symbol.n_type = nType;
-    symbol.n_sect = 1; // Sections count from 1.
+    symbol.n_sect = nSect+1; // Sections count from 1.
     symbol.n_desc = REFERENCE_FLAG_DEFINED;
     adjustOffset(nSect, offset);
     symbol.n_value = offset;
@@ -270,10 +297,15 @@ void MachoExport::createStructsRelocation(unsigned sect, POLYUNSIGNED offset)
 {
     struct relocation_info reloc;
     reloc.r_address = offset;
-    reloc.r_symbolnum = 1; // Section numbers start at 1
+    reloc.r_symbolnum = sect+1; // Section numbers start at 1
     reloc.r_pcrel = 0;
+#if (SIZEOF_VOIDP == 8)
+    reloc.r_length = 3; // 8 bytes
+    reloc.r_type = X86_64_RELOC_UNSIGNED;
+#else
     reloc.r_length = 2; // 4 bytes
     reloc.r_type = GENERIC_RELOC_VANILLA;
+#endif
     reloc.r_extern = 0; // r_symbolnum is a section number.
 
     fwrite(&reloc, sizeof(reloc), 1, exportFile);
@@ -283,39 +315,57 @@ void MachoExport::createStructsRelocation(unsigned sect, POLYUNSIGNED offset)
 void MachoExport::exportStore(void)
 {
     PolyWord    *p;
+#if (SIZEOF_VOIDP == 8)
+    struct mach_header_64 fhdr;
+    struct segment_command_64 sHdr;
+    struct section_64 *sections = new section_64[memTableEntries+1];
+    size_t sectionSize = sizeof(section_64);
+#else
     struct mach_header fhdr;
     struct segment_command sHdr;
+    struct section *sections = new section[memTableEntries+1];
+    size_t sectionSize = sizeof(section);
+#endif
     struct symtab_command symTab;
-    struct section theSection;
     unsigned i;
 
     // Write out initial values for the headers.  These are overwritten at the end.
     // File header
     memset(&fhdr, 0, sizeof(fhdr));
-    fhdr.magic = MH_MAGIC; // Feed Face (0xfeedface)
     fhdr.filetype = MH_OBJECT;
     fhdr.ncmds = 2; // One for the segment and one for the symbol table.
-    fhdr.sizeofcmds = sizeof(struct segment_command) + sizeof(struct section)*(memTableEntries+1)
-        + sizeof(struct symtab_command);
+    fhdr.sizeofcmds = sizeof(sHdr) + sectionSize * (memTableEntries+1) + sizeof(symTab);
     fhdr.flags = 0;
     // The machine needs to match the machine we're compiling for
     // even if this is actually portable code.
+#if (SIZEOF_VOIDP == 8)
+    fhdr.magic = MH_MAGIC_64; // (0xfeedfacf) 64-bit magic number
+#else
+    fhdr.magic = MH_MAGIC; // Feed Face (0xfeedface)
+#endif
 #if defined(HOSTARCHITECTURE_X86)
     fhdr.cputype = CPU_TYPE_I386;
     fhdr.cpusubtype = CPU_SUBTYPE_I386_ALL;
 #elif defined(HOSTARCHITECTURE_PPC)
     fhdr.cputype = CPU_TYPE_POWERPC;
     fhdr.cpusubtype = CPU_SUBTYPE_POWERPC_ALL;
+#elif defined(HOSTARCHITECTURE_X86_64)
+    fhdr.cputype = CPU_TYPE_X86_64;
+    fhdr.cpusubtype = CPU_SUBTYPE_X86_64_ALL;
 #else
 #error "No support for exporting on this architecture"
 #endif
     fwrite(&fhdr, sizeof(fhdr), 1, exportFile); // Write it for the moment.
 
     // Segment header.
-    memset(&sHdr, 0, sizeof(struct segment_command));
+    memset(&sHdr, 0, sizeof(sHdr));
+#if (SIZEOF_VOIDP == 8)
+    sHdr.cmd = LC_SEGMENT_64;
+#else
     sHdr.cmd = LC_SEGMENT;
-    sHdr.nsects = 1;
-    sHdr.cmdsize = sizeof(struct segment_command) + sizeof(struct section) * sHdr.nsects;
+#endif
+    sHdr.nsects = memTableEntries+1; // One for each entry plus one for the tables.
+    sHdr.cmdsize = sizeof(sHdr) + sectionSize * sHdr.nsects;
     // Add up the sections to give the file size
     sHdr.filesize = 0;
     for (i = 0; i < memTableEntries; i++)
@@ -330,29 +380,65 @@ void MachoExport::exportStore(void)
     // Write it initially.
     fwrite(&sHdr, sizeof(sHdr), 1, exportFile);
 
-    // Section header.
-    memset(&theSection, 0, sizeof(struct section));
-    sprintf(theSection.sectname, "poly");
-    sprintf(theSection.segname, "POLY");
-    //theSection.offset is set later
-    //theSection.reloff is set later
-    //theSection.nreloc is set later
-    theSection.align = 3; // 8 byte alignment
-    // theSection.size is set later
-    theSection.flags = S_ATTR_LOC_RELOC | S_ATTR_SOME_INSTRUCTIONS | S_REGULAR; // 
+    // Section header for each entry in the table
+    POLYUNSIGNED sectAddr = sizeof(exportDescription)+sizeof(memoryTableEntry)*memTableEntries;
+    for (i = 0; i < memTableEntries; i++)
+    {
+        memset(&(sections[i]), 0, sectionSize);
 
-    // Write it out for the moment.
-    fwrite(&theSection, sizeof(struct section), 1, exportFile);
+        if (memTable[i].mtFlags & MTF_WRITEABLE)
+        {
+            // Mutable areas
+            ASSERT(!(memTable[i].mtFlags & MTF_EXECUTABLE)); // Executable areas can't be writable.
+            sprintf(sections[i].sectname, "__data");
+            sprintf(sections[i].segname, "__DATA");
+            sections[i].flags = S_ATTR_LOC_RELOC | S_REGULAR;
+        }
+        else if (memTable[i].mtFlags & MTF_EXECUTABLE)
+        {
+            sprintf(sections[i].sectname, "__text");
+            sprintf(sections[i].segname, "__TEXT");
+            sections[i].flags = S_ATTR_LOC_RELOC | S_ATTR_SOME_INSTRUCTIONS | S_REGULAR;
+        }
+        else
+        {
+            sprintf(sections[i].sectname, "__const");
+            sprintf(sections[i].segname, "__DATA");
+            sections[i].flags = S_ATTR_LOC_RELOC | S_REGULAR;
+        }
+
+        sections[i].addr = sectAddr;
+        sections[i].size = memTable[i].mtLength;
+        sectAddr += memTable[i].mtLength;
+        //sections[i].offset is set later
+        //sections[i].reloff is set later
+        //sections[i].nreloc is set later
+        sections[i].align = 3; // 8 byte alignment
+        // theSection.size is set later
+
+    }
+    // For the tables.
+    memset(&(sections[memTableEntries]), 0, sectionSize);
+    sprintf(sections[memTableEntries].sectname, "__const");
+    sprintf(sections[memTableEntries].segname, "__DATA");
+    sections[memTableEntries].addr = 0;
+    sections[memTableEntries].size = sizeof(exportDescription)+sizeof(memoryTableEntry)*memTableEntries;
+    sections[memTableEntries].align = 3; // 8 byte alignment
+    // theSection.size is set later
+    sections[memTableEntries].flags = S_ATTR_LOC_RELOC | S_ATTR_SOME_INSTRUCTIONS | S_REGULAR;
+
+    // Write them out for the moment.
+    fwrite(sections, sectionSize * (memTableEntries+1), 1, exportFile);
 
     // Symbol table header.
-    memset(&symTab, 0, sizeof(struct symtab_command));
+    memset(&symTab, 0, sizeof(symTab));
     symTab.cmd = LC_SYMTAB;
-    symTab.cmdsize = sizeof(struct symtab_command);
+    symTab.cmdsize = sizeof(symTab);
     //symTab.symoff is set later
     //symTab.nsyms is set later
     //symTab.stroff is set later
     //symTab.strsize is set later
-    fwrite(&symTab, sizeof(struct symtab_command), 1, exportFile);
+    fwrite(&symTab, sizeof(symTab), 1, exportFile);
 
     // Create the symbol table first before we mess up the addresses by turning them
     // into relocations.
@@ -370,7 +456,10 @@ void MachoExport::exportStore(void)
             char buff[50];
             sprintf(buff, "area%0d", i);
             writeSymbol(buff, N_SECT, i, 0);
+#if (SIZEOF_VOIDP == 8)
             // See if we can find the names of any functions.
+            // This seems to break on 32-bit Mac OS X.  It seems to align
+            // some relocations onto an 8-byte boundary so we just disable it.
             char *start = (char*)memTable[i].mtAddr;
             char *end = start + memTable[i].mtLength;
             for (p = (PolyWord*)start; p < (PolyWord*)end; )
@@ -388,15 +477,16 @@ void MachoExport::exportStore(void)
                 }
                 p += length;
             }
+#endif
         }
     }
     symTab.nsyms = symbolCount;
 
     // Create and write out the relocations.
-    theSection.reloff = ftell(exportFile);
-    relocationCount = 0;
     for (i = 0; i < memTableEntries; i++)
     {
+        sections[i].reloff = ftell(exportFile);
+        relocationCount = 0;
         if (i != ioMemEntry) // Don't relocate the IO area
         {
             // Create the relocation table and turn all addresses into offsets.
@@ -413,9 +503,12 @@ void MachoExport::exportStore(void)
                 p += length;
             }
         }
+        sections[i].nreloc = relocationCount;
     }
 
     // Additional relocations for the descriptors.
+    sections[memTableEntries].reloff = ftell(exportFile);
+    relocationCount = 0;
 
     // Address of "memTable" within "exports". We can't use createRelocation because
     // the position of the relocation is not in either the mutable or the immutable area.
@@ -433,7 +526,7 @@ void MachoExport::exportStore(void)
         createStructsRelocation(i,
             sizeof(exportDescription) + i * sizeof(memoryTableEntry) + offsetof(memoryTableEntry, mtAddr));
     }
-    theSection.nreloc = relocationCount;
+    sections[memTableEntries].nreloc = relocationCount;
 
     // The symbol name table
     symTab.stroff = ftell(exportFile);
@@ -456,7 +549,7 @@ void MachoExport::exportStore(void)
     exports.architecture = machineDependent->MachineArchitecture();
     exports.rtsVersion = POLY_version_number;
 
-    theSection.offset = ftell(exportFile);
+    sections[memTableEntries].offset = ftell(exportFile);
     fwrite(&exports, sizeof(exports), 1, exportFile);
     POLYUNSIGNED addrOffset = sizeof(exports)+sizeof(memoryTableEntry)*memTableEntries;
     for (i = 0; i < memTableEntries; i++)
@@ -467,19 +560,21 @@ void MachoExport::exportStore(void)
         fwrite(&memTable[i], sizeof(memoryTableEntry), 1, exportFile);
         memTable[i].mtAddr = save;
     }
-    theSection.size = addrOffset;
 
     // Now the binary data.
     for (i = 0; i < memTableEntries; i++)
     {
+        alignFile(4);
+        sections[i].offset = ftell(exportFile);
         fwrite(memTable[i].mtAddr, 1, memTable[i].mtLength, exportFile);
     }
-
     // Rewind to rewrite the headers with the actual offsets.
     rewind(exportFile);
     fwrite(&fhdr, sizeof(fhdr), 1, exportFile); // File header
     fwrite(&sHdr, sizeof(sHdr), 1, exportFile); // Segment header
-    fwrite(&theSection, sizeof(struct section), 1, exportFile); // Section headers
-    fwrite(&symTab, sizeof(struct symtab_command), 1, exportFile); // Symbol table header
+    fwrite(sections, sectionSize * (memTableEntries+1), 1, exportFile); // Section headers
+    fwrite(&symTab, sizeof(symTab), 1, exportFile); // Symbol table header
     fclose(exportFile); exportFile = NULL;
+
+    delete[](sections);
 }
