@@ -23,27 +23,15 @@
 #define _MPOLY_H_DEFINED
 
 #include "noreturn.h"
-#include "polyexports.h"
+#include "../polyexports.h"
 
 extern struct _userOptions {
     unsigned    user_arg_count;
     char        **user_arg_strings;
     const char  *programName;
-    unsigned    debug;              /* debugging  flags                       */
-    bool        noDisplay;          /* X display flag                         */
-    // Not all of these can be set any longer.
-    unsigned long    heapSize, immutableSegSize, mutableSegSize;
-    unsigned long    immutableFreeSpace, mutableFreeSpace;
-    unsigned long    immutableMinFree, mutableMinFree; // Probably remove
+    unsigned    gcthreads;    // Number of threads to use for gc
 } userOptions;
 
-// Values for debugging flags
-#define DEBUG_CHECK_OBJECTS 1       // Check that addresses are valid.
-#define DEBUG_REGION_CHECK  2       // VERY EXPENSIVE checking
-#define DEBUG_NOGROW        4       // Don't allocate new heap segments
-#define DEBUG_FORCEGC       16      // Force a GC every allocation
-#define DEBUG_X             128     // X-Windows debugging
-
 class PolyWord;
 
 // Return the entry in the io vector corresponding to the Poly system call.
@@ -53,6 +41,6 @@ NORETURNFN(extern void finish(int n));
 
 extern char *RTSArgHelp(void);
 
-extern UNSIGNEDADDR exportTimeStamp;
+extern time_t exportTimeStamp;
 
 #endif /* _MPOLY_H_DEFINED */
