@@ -18,10 +18,12 @@
     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
 */
-#ifdef WIN32
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#elif defined(_WIN32)
 #include "winconfig.h"
 #else
-#include "config.h"
+#error "No configuration file"
 #endif
 
 #ifdef HAVE_STDIO_H
@@ -104,7 +106,7 @@ typedef int socklen_t;
 #endif
 
 
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
 #ifdef USEWINSOCK2
 #include <winsock2.h>
 #else
@@ -125,6 +127,7 @@ typedef int socklen_t;
 #include "save_vec.h"
 #include "rts_module.h"
 #include "machine_dep.h"
+#include "errors.h"
 
 #ifndef MAXHOSTNAMELEN
 #define MAXHOSTNAMELEN 256
@@ -135,7 +138,7 @@ typedef int socklen_t;
 #define ALLOC(n) alloc_and_save(taskData, n)
 #define SIZEOF(x) (sizeof(x)/sizeof(PolyWord))
 
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
 static int winsock_init = 0; /* Check that it has been initialised. */
 
 #else
@@ -282,13 +285,12 @@ struct sk_tab_struct {
     { "DGRAM",      SOCK_DGRAM },
     { "RAW",        SOCK_RAW },
     { "RDM",        SOCK_RDM },
-    { "SEQPACKET",      SOCK_SEQPACKET }
+    { "SEQPACKET",  SOCK_SEQPACKET }
 };
 
 static Handle makeHostEntry(TaskData *taskData, struct hostent *host);
 static Handle makeProtoEntry(TaskData *taskData, struct protoent *proto);
 static Handle makeServEntry(TaskData *taskData, struct servent *proto);
-static Handle makeNetEntry(TaskData *taskData, struct netent *net);
 static Handle makeList(TaskData *taskData, int count, char *p, int size, void *arg,
                        Handle (mkEntry)(TaskData *, void*, char*));
 static Handle mkAftab(TaskData *taskData, void*, char *p);
@@ -298,47 +300,6 @@ static Handle getSocketOption(TaskData *taskData, Handle args, int level, int op
 static Handle getSocketInt(TaskData *taskData, Handle args, int level, int opt);
 static Handle selectCall(TaskData *taskData, Handle args, int blockType);
 
-#ifdef WINDOWS_PC
-/* To allow for portable code we map Windows socket errors to
-   errno-style errors if there is a suitable entry in errno.h.
-   If there isn't we simply return the negative value and handle
-   it later. N.B. WSAEWOULDBLOCK is one of those which does not
-   have an equivalent.  (It does seem to be defined in Windows CE). */
-static int MapError(int err)
-{
-    switch (err)
-    {
-#ifdef EINTR
-    case WSAEINTR: return EINTR;
-#endif
-#ifdef EBADF
-    case WSAEBADF: return EBADF;
-#endif
-#ifdef EACCES
-    case WSAEACCES: return EACCES;
-#endif
-#ifdef EFAULT
-    case WSAEFAULT: return EFAULT;
-#endif
-#ifdef EINVAL
-    case WSAEINVAL: return EINVAL;
-#endif
-#ifdef EMFILE
-    case WSAEMFILE: return EMFILE;
-#endif
-#ifdef ENAMETOOLONG
-    case WSAENAMETOOLONG: return ENAMETOOLONG;
-#endif
-#ifdef ENOTEMPTY
-    case WSAENOTEMPTY: return ENOTEMPTY;
-#endif
-#ifdef EWOULDBLOCK
-	case WSAEWOULDBLOCK: return EWOULDBLOCK;
-#endif
-    default: return -err;
-    }
-}
-
 /* If these are not defined define them as negative because GetError returns
    negative values for socket library errors which do not have
    equivalents as errno-style errors. */
@@ -362,21 +323,67 @@ static int MapError(int err)
 #define EBADF       (-WSAEBADF)
 #endif
 
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
+/* To allow for portable code we map Windows socket errors to
+   errno-style errors if there is a suitable entry in errno.h.
+   If there isn't we simply return the negative value and handle
+   it later. N.B. WSAEWOULDBLOCK is one of those which does not
+   have an equivalent.  (It does seem to be defined in Windows CE). */
+int mapWindowsErrorCode(int err)
+{
+#ifdef EWOULDBLOCK
+    // This is very common so we treat it specially.
+    if (err == WSAEWOULDBLOCK)
+        return EWOULDBLOCK;
+#endif
+    const char *errText = stringFromErrorCode(-err);
+    int newErr = 0;
+    if (errText != 0 && errorCodeFromString(errText, &newErr))
+        return newErr;
+    else return -err;
+}
+
 static int GetError()
 {
-    return MapError(WSAGetLastError());
+    return mapWindowsErrorCode(WSAGetLastError());
 }
+
 #define GETERROR    (GetError())
-#define MAPERROR(x) (MapError(x))
+#define MAPERROR(x) (mapWindowsErrorCode(x))
 #else
 #define GETERROR    (errno)
 #define MAPERROR(x) (x)
 #endif
 
+class WaitNet: public Waiter {
+public:
+    WaitNet(SOCKET sock, bool isOOB = false) : m_sock(sock), m_isOOB(isOOB) {}
+    void Wait(unsigned maxMillisecs);
+private:
+    SOCKET m_sock;
+    bool m_isOOB;
+};
+
+// Use "select" in both Windows and Unix.  In Windows that means we
+// don't watch hWakeupEvent but that's only a hint.
+void WaitNet::Wait(unsigned maxMillisecs)
+{
+    fd_set readFds, writeFds, exceptFds;
+    struct timeval toWait = { 0, 0 };
+    // Mac OS X requires the usec field to be less than a million
+    toWait.tv_sec = maxMillisecs / 1000;
+    toWait.tv_usec = (maxMillisecs % 1000) * 1000;
+    FD_ZERO(&readFds);
+    FD_ZERO(&writeFds);
+    FD_ZERO(&exceptFds);
+    FD_SET(m_sock, m_isOOB ? &exceptFds : &readFds);
+    int result = select(FD_SETSIZE, &readFds, &writeFds, &exceptFds, &toWait);
+    ASSERT(result >= 0 || errno == EINTR); // The only "error" should be an interrupt.
+}
 
 Handle Net_dispatch_c(TaskData *taskData, Handle args, Handle code)
 {
-    int c = get_C_long(taskData, DEREFWORDHANDLE(code));
+    int c = get_C_int(taskData, DEREFWORDHANDLE(code));
 TryAgain:
     switch (c)
     {
@@ -392,13 +399,10 @@ TryAgain:
         {
             /* Look up a host name. */
             char hostName[MAXHOSTNAMELEN];
-            int length;
-            struct hostent *host;
-            length = Poly_string_to_C(DEREFWORD(args),
-                        hostName, MAXHOSTNAMELEN);
+            POLYUNSIGNED length = Poly_string_to_C(DEREFWORD(args), hostName, MAXHOSTNAMELEN);
             if (length > MAXHOSTNAMELEN)
                 raise_syscall(taskData, "Host name too long", ENAMETOOLONG);
-            host = gethostbyname(hostName);
+            struct hostent *host = gethostbyname(hostName);
             if (host == NULL)
                 raise_syscall(taskData, "gethostbyname failed", GETERROR);
             return makeHostEntry(taskData, host);
@@ -408,7 +412,7 @@ TryAgain:
         {
             /* Look up entry by address. */
             unsigned long addr =
-                htonl(get_C_ulong(taskData, DEREFWORDHANDLE(args)));
+                htonl(get_C_unsigned(taskData, DEREFWORDHANDLE(args)));
             struct hostent *host;
             /* Look up a host name given an address. */
             host = gethostbyaddr((char*)&addr, sizeof(addr), AF_INET);
@@ -421,13 +425,10 @@ TryAgain:
         {
             /* Look up protocol entry. */
             char protoName[MAXHOSTNAMELEN];
-            int length;
-            struct protoent *proto;
-            length = Poly_string_to_C(DEREFWORD(args),
-                        protoName, MAXHOSTNAMELEN);
+            POLYUNSIGNED length = Poly_string_to_C(DEREFWORD(args), protoName, MAXHOSTNAMELEN);
             if (length > MAXHOSTNAMELEN)
                 raise_syscall(taskData, "Protocol name too long", ENAMETOOLONG);
-            proto = getprotobyname(protoName);
+            struct protoent *proto = getprotobyname(protoName);
             if (proto == NULL)
                 raise_syscall(taskData, "getprotobyname failed", GETERROR);
             return makeProtoEntry(taskData, proto);
@@ -437,7 +438,7 @@ TryAgain:
         {
             /* Look up protocol entry. */
             struct protoent *proto;
-            int pNum = get_C_ulong(taskData, DEREFWORDHANDLE(args));
+            int pNum = get_C_int(taskData, DEREFWORDHANDLE(args));
             proto = getprotobynumber(pNum);
             if (proto == NULL)
                 raise_syscall(taskData, "getprotobynumber failed", GETERROR);
@@ -448,13 +449,10 @@ TryAgain:
         {
             /* Get service given service name only. */
             char servName[MAXHOSTNAMELEN];
-            int length;
-            struct servent *serv;
-            length = Poly_string_to_C(DEREFWORD(args),
-                        servName, MAXHOSTNAMELEN);
+            POLYUNSIGNED length = Poly_string_to_C(DEREFWORD(args), servName, MAXHOSTNAMELEN);
             if (length > MAXHOSTNAMELEN)
                 raise_syscall(taskData, "Service name too long", ENAMETOOLONG);
-            serv = getservbyname (servName, NULL);
+            struct servent *serv = getservbyname (servName, NULL);
             if (serv == NULL)
                 raise_syscall(taskData, "getservbyname failed", GETERROR);
             return makeServEntry(taskData, serv);
@@ -464,15 +462,13 @@ TryAgain:
         {
             /* Get service given service name and protocol name. */
             char servName[MAXHOSTNAMELEN], protoName[MAXHOSTNAMELEN];
-            int length;
-            struct servent *serv; 
-            length = Poly_string_to_C(args->WordP()->Get(0), servName, MAXHOSTNAMELEN);
+            POLYUNSIGNED length = Poly_string_to_C(args->WordP()->Get(0), servName, MAXHOSTNAMELEN);
             if (length > MAXHOSTNAMELEN)
                 raise_syscall(taskData, "Service name too long", ENAMETOOLONG);
             length = Poly_string_to_C(args->WordP()->Get(1), protoName, MAXHOSTNAMELEN);
             if (length > MAXHOSTNAMELEN)
                 raise_syscall(taskData, "Protocol name too long", ENAMETOOLONG);
-            serv = getservbyname (servName, protoName);
+            struct servent *serv = getservbyname (servName, protoName);
             if (serv == NULL)
                 raise_syscall(taskData, "getservbyname failed", GETERROR);
             return makeServEntry(taskData, serv);
@@ -493,10 +489,9 @@ TryAgain:
         {
             /* Get service given port number and protocol name. */
             char protoName[MAXHOSTNAMELEN];
-            int length;
             struct servent *serv;
             long port = htons(get_C_ushort(taskData, DEREFHANDLE(args)->Get(0)));
-            length = Poly_string_to_C(args->WordP()->Get(1), protoName, MAXHOSTNAMELEN);
+            POLYUNSIGNED length = Poly_string_to_C(args->WordP()->Get(1), protoName, MAXHOSTNAMELEN);
             if (length > MAXHOSTNAMELEN)
                 raise_syscall(taskData, "Protocol name too long", ENAMETOOLONG);
             serv = getservbyport (port, protoName);
@@ -505,40 +500,8 @@ TryAgain:
             return makeServEntry(taskData, serv);
         }
 
-    case 9:
-        {
-            /* Get a network entry. */
-            struct netent *net = NULL;
-#ifdef HAVE_GETNETBYNAME
-            char netName[MAXHOSTNAMELEN];
-            POLYUNSIGNED length = Poly_string_to_C(DEREFWORD(args),
-                        netName, MAXHOSTNAMELEN);
-            if (length > MAXHOSTNAMELEN)
-                raise_syscall(taskData, "Network name too long", ENAMETOOLONG);
-            /* Winsock2 does not have getnetbyname.
-               Just return failure. */
-            net = getnetbyname(netName);
-#endif
-            if (net == NULL)
-                raise_syscall(taskData, "getnetbyname failed", GETERROR);
-            return makeNetEntry(taskData, net);
-        }
-
-    case 10:
-        {
-            /* Look up network entry from a number. */
-            struct netent *net = NULL;
-#ifdef HAVE_GETNETBYADDR
-            unsigned long netNum = htonl(get_C_ulong(taskData, DEREFHANDLE(args)->Get(0)));
-            long af = get_C_ulong(taskData, DEREFHANDLE(args)->Get(1));
-            /* Winsock2 does not have getnetbyaddr.
-               Just return failure. */
-            net = getnetbyaddr(netNum, af);
-#endif
-            if (net == NULL)
-                raise_syscall(taskData, "getnetbyaddr failed", GETERROR);
-            return makeNetEntry(taskData, net);
-        }
+    // 9 and 10 were used for the NetDB structure which was removed from
+    // the basis library a long time ago.
 
     case 11:
         {
@@ -563,10 +526,10 @@ TryAgain:
         {
             Handle str_token = make_stream_entry(taskData);
             PIOSTRUCT strm;
-            int stream_no = STREAMID(str_token);
-            int af = get_C_long(taskData, DEREFHANDLE(args)->Get(0));
-            int type = get_C_long(taskData, DEREFHANDLE(args)->Get(1));
-            int proto = get_C_long(taskData, DEREFHANDLE(args)->Get(2));
+            unsigned stream_no = STREAMID(str_token);
+            int af = get_C_int(taskData, DEREFHANDLE(args)->Get(0));
+            int type = get_C_int(taskData, DEREFHANDLE(args)->Get(1));
+            int proto = get_C_int(taskData, DEREFHANDLE(args)->Get(2));
             unsigned long onOff = 1;
             SOCKET skt = socket(af, type, proto);
             if (skt == INVALID_SOCKET)
@@ -587,14 +550,14 @@ TryAgain:
                 }
             }
             /* Set the socket to non-blocking mode. */
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
             if (ioctlsocket(skt, FIONBIO, &onOff) != 0)
 #else
             if (ioctl(skt, FIONBIO, &onOff) < 0)
 #endif
             {
                 free_stream_entry(stream_no);
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
                 closesocket(skt);
 #else
                 close(skt);
@@ -672,7 +635,7 @@ TryAgain:
         {
             struct linger linger;
             PIOSTRUCT strm = get_stream(DEREFHANDLE(args)->Get(0).AsObjPtr());
-            int lTime = get_C_long(taskData, DEREFHANDLE(args)->Get(1));
+            int lTime = get_C_int(taskData, DEREFHANDLE(args)->Get(1));
             /* We pass in a negative value to turn the option off,
                zero or positive to turn it on. */
             if (lTime < 0)
@@ -746,7 +709,7 @@ TryAgain:
             sockaddr.sin_family = AF_INET;
             sockaddr.sin_port = htons(get_C_ushort(taskData, DEREFHANDLE(args)->Get(0)));
             sockaddr.sin_addr.s_addr =
-                htonl(get_C_ulong(taskData, DEREFHANDLE(args)->Get(1)));
+                htonl(get_C_unsigned(taskData, DEREFHANDLE(args)->Get(1)));
             return(SAVE(Buffer_to_Poly(taskData, (char*)&sockaddr, sizeof(sockaddr))));
         }
 
@@ -775,7 +738,7 @@ TryAgain:
             PIOSTRUCT strm = get_stream(args->WordP());
             unsigned long readable;
             if (strm == NULL) raise_syscall(taskData, "Stream is closed", EBADF);
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
             if (ioctlsocket(strm->device.sock, FIONREAD, &readable) != 0)
                 raise_syscall(taskData, "ioctlsocket failed", GETERROR);
 #else
@@ -790,7 +753,7 @@ TryAgain:
             PIOSTRUCT strm = get_stream(args->WordP());
             unsigned long atMark;
             if (strm == NULL) raise_syscall(taskData, "Stream is closed", EBADF);
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
             if (ioctlsocket(strm->device.sock, SIOCATMARK, &atMark) != 0)
                 raise_syscall(taskData, "ioctlsocket failed", GETERROR);
 #else
@@ -801,6 +764,9 @@ TryAgain:
         }
 
     case 46: /* Accept a connection. */
+        // We should check for interrupts even if we're not going to block.
+        processes->TestAnyEvents(taskData);
+
     case 58: /* Non-blocking accept. */
         {
             PIOSTRUCT strm = get_stream(args->WordP());
@@ -808,7 +774,6 @@ TryAgain:
             else {
                 SOCKET sock = strm->device.sock, result;
                 struct sockaddr resultAddr;
-                int stream_no;
                 socklen_t addrLen;
                 Handle addrHandle, pair;
                 Handle str_token;
@@ -816,7 +781,7 @@ TryAgain:
                 /* Get a token for the new socket - may raise an
                    exception if it fails. */
                 str_token = make_stream_entry(taskData);
-                stream_no = STREAMID(str_token);
+                unsigned stream_no = STREAMID(str_token);
                 addrLen = sizeof(resultAddr);
                 result = accept(sock, &resultAddr, &addrLen);
 
@@ -840,8 +805,10 @@ TryAgain:
                         /* If the socket is in non-blocking mode we pass
                            this back to the caller.  If it is blocking we
                            suspend this process and try again later. */
-                        if (c == 46 /* blocking version. */)
-                            processes->BlockAndRestart(taskData, strm->device.sock, false, POLY_SYS_network);
+                        if (c == 46 /* blocking version. */) {
+                            WaitNet waiter(strm->device.sock);
+                            processes->BlockAndRestart(taskData, &waiter, false, POLY_SYS_network);
+                        }
                         /* else drop through. */
                     default:
                         raise_syscall(taskData, "accept failed", GETERROR);
@@ -867,12 +834,14 @@ TryAgain:
             PolyStringObject * psAddr = (PolyStringObject *)args->WordP()->Get(1).AsObjPtr();
             struct sockaddr *psock = (struct sockaddr *)&psAddr->chars;
             if (strm == NULL) raise_syscall(taskData, "Stream is closed", EBADF);
-            if (bind(strm->device.sock, psock, psAddr->length) != 0)
+            if (bind(strm->device.sock, psock, (int)psAddr->length) != 0)
                 raise_syscall(taskData, "bind failed", GETERROR);
             return Make_arbitrary_precision(taskData, 0);
         }
 
     case 48: /* Connect to an address. */
+        // We should check for interrupts even if we're not going to block.
+        processes->TestAnyEvents(taskData);
     case 59: /* Non-blocking connect. */
         {
             PIOSTRUCT strm = get_stream(DEREFHANDLE(args)->Get(0).AsObjPtr());
@@ -911,7 +880,7 @@ TryAgain:
                     else if (sel == 0)
                     {
                         /* Nothing yet */
-                        processes->BlockAndRestart(taskData, -1, false, POLY_SYS_network);
+                        processes->BlockAndRestart(taskData, NULL, false, POLY_SYS_network);
                             /* -1 => not for reading. */
                     }
                     else /* Definite result. */
@@ -928,7 +897,7 @@ TryAgain:
                 else
                 {
                     int err;
-                    res = connect(strm->device.sock, psock, psAddr->length);
+                    res = connect(strm->device.sock, psock, (int)psAddr->length);
                     if (res == 0) return Make_arbitrary_precision(taskData, 0); /* OK */
                     /* It isn't clear that EINTR can ever occur with
                        connect, but just to be safe, we retry. */
@@ -936,7 +905,7 @@ TryAgain:
                     if ((err == EWOULDBLOCK || err == EINPROGRESS) && c == 48 /*blocking version*/)
                     {
                         strm->ioBits |= IO_BIT_INPROGRESS;
-                        processes->BlockAndRestart(taskData, -1, false, POLY_SYS_network);
+                        processes->BlockAndRestart(taskData, NULL, false, POLY_SYS_network);
                             /* -1 => not for reading. */
                     }
                     else if (err != EINTR)
@@ -949,7 +918,7 @@ TryAgain:
     case 49: /* Put socket into listening mode. */
         {
             PIOSTRUCT strm = get_stream(DEREFHANDLE(args)->Get(0).AsObjPtr());
-            int backlog = get_C_ulong(taskData, DEREFHANDLE(args)->Get(1));
+            int backlog = get_C_int(taskData, DEREFHANDLE(args)->Get(1));
             if (strm == NULL) raise_syscall(taskData, "Stream is closed", EBADF);
             if (listen(strm->device.sock, backlog) != 0)
                 raise_syscall(taskData, "listen failed", GETERROR);
@@ -973,15 +942,17 @@ TryAgain:
         }
 
     case 51: /* Send data on a socket. */
+        // We should check for interrupts even if we're not going to block.
+        processes->TestAnyEvents(taskData);
     case 60: /* Non-blocking send. */
         {
             PIOSTRUCT strm = get_stream(DEREFHANDLE(args)->Get(0).AsObjPtr());
             PolyWord pBase = DEREFHANDLE(args)->Get(1);
             char    ch, *base;
-            unsigned int offset = get_C_ulong(taskData, DEREFHANDLE(args)->Get(2));
-            unsigned int length = get_C_ulong(taskData, DEREFHANDLE(args)->Get(3));
-            unsigned int dontRoute = get_C_ulong(taskData, DEREFHANDLE(args)->Get(4));
-            unsigned int outOfBand = get_C_ulong(taskData, DEREFHANDLE(args)->Get(5));
+            unsigned int offset = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(2));
+            unsigned int length = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(3));
+            unsigned int dontRoute = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(4));
+            unsigned int outOfBand = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(5));
             int flags = 0, sent;
             if (dontRoute != 0) flags |= MSG_DONTROUTE;
             if (outOfBand != 0) flags |= MSG_OOB;
@@ -1008,7 +979,7 @@ TryAgain:
                 err = GETERROR;
                 if (err == EWOULDBLOCK && c == 51 /* blocking */)
                 {
-                    processes->BlockAndRestart(taskData, -1, false, POLY_SYS_network);
+                    processes->BlockAndRestart(taskData, NULL, false, POLY_SYS_network);
                         /* -1 => not for reading */
                     ASSERT(0); /* Must not have returned. */
                 }
@@ -1019,16 +990,18 @@ TryAgain:
         }
 
     case 52: /* Send data on a socket to a given address. */
+        // We should check for interrupts even if we're not going to block.
+        processes->TestAnyEvents(taskData);
     case 61: /* Non-blocking send. */
         {
             PIOSTRUCT strm = get_stream(DEREFHANDLE(args)->Get(0).AsObjPtr());
             PolyStringObject * psAddr = (PolyStringObject *)args->WordP()->Get(1).AsObjPtr();
             PolyWord pBase = DEREFHANDLE(args)->Get(2);
             char    ch, *base;
-            unsigned int offset = get_C_ulong(taskData, DEREFHANDLE(args)->Get(3));
-            unsigned int length = get_C_ulong(taskData, DEREFHANDLE(args)->Get(4));
-            unsigned int dontRoute = get_C_ulong(taskData, DEREFHANDLE(args)->Get(5));
-            unsigned int outOfBand = get_C_ulong(taskData, DEREFHANDLE(args)->Get(6));
+            unsigned int offset = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(3));
+            unsigned int length = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(4));
+            unsigned int dontRoute = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(5));
+            unsigned int outOfBand = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(6));
             int flags = 0, sent;
             if (dontRoute != 0) flags |= MSG_DONTROUTE;
             if (outOfBand != 0) flags |= MSG_OOB;
@@ -1047,7 +1020,7 @@ TryAgain:
             {
                 int err;
                 sent = sendto(strm->device.sock, base+offset, length, flags,
-                            (struct sockaddr *)psAddr->chars, psAddr->length);
+                            (struct sockaddr *)psAddr->chars, (int)psAddr->length);
                 /* It isn't clear that EINTR can ever occur with
                    send but just to be safe we deal with that case and
                    retry the send. */
@@ -1056,7 +1029,7 @@ TryAgain:
                 err = GETERROR;
                 if (err == EWOULDBLOCK && c == 52 /* blocking */)
                 {
-                    processes->BlockAndRestart(taskData, -1, false, POLY_SYS_network);
+                    processes->BlockAndRestart(taskData, NULL, false, POLY_SYS_network);
                     ASSERT(0); /* Must not have returned. */
                 }
                 else if (err != EINTR)
@@ -1066,14 +1039,16 @@ TryAgain:
         }
 
     case 53: /* Receive data into an array. */
+        // We should check for interrupts even if we're not going to block.
+        processes->TestAnyEvents(taskData);
     case 62: /* Non-blocking receive. */
         {
             PIOSTRUCT strm = get_stream(DEREFHANDLE(args)->Get(0).AsObjPtr());
             char *base = (char*)DEREFHANDLE(args)->Get(1).AsObjPtr()->AsBytePtr();
-            unsigned int offset = get_C_ulong(taskData, DEREFHANDLE(args)->Get(2));
-            unsigned int length = get_C_ulong(taskData, DEREFHANDLE(args)->Get(3));
-            unsigned int peek = get_C_ulong(taskData, DEREFHANDLE(args)->Get(4));
-            unsigned int outOfBand = get_C_ulong(taskData, DEREFHANDLE(args)->Get(5));
+            unsigned int offset = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(2));
+            unsigned int length = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(3));
+            unsigned int peek = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(4));
+            unsigned int outOfBand = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(5));
             int flags = 0, recvd;
             if (peek != 0) flags |= MSG_PEEK;
             if (outOfBand != 0) flags |= MSG_OOB;
@@ -1092,8 +1067,8 @@ TryAgain:
                 if (err == EWOULDBLOCK && c == 53 /* blocking */)
                 {
                     /* Block until something arrives. */
-                    processes->BlockAndRestart(taskData, outOfBand ? -1 : strm->device.sock,
-                        false, POLY_SYS_network);
+                    WaitNet waiter(strm->device.sock, outOfBand != 0);
+                    processes->BlockAndRestart(taskData, &waiter, false, POLY_SYS_network);
                     ASSERT(0); /* Must not have returned. */
                 }
                 else if (err != EINTR)
@@ -1105,14 +1080,16 @@ TryAgain:
     case 54: /* Receive data into an array and return the sender's
                 address along with the length.  In Windows this can
                 only be used with datagrams. */
+        // We should check for interrupts even if we're not going to block.
+        processes->TestAnyEvents(taskData);
     case 63: /* Non-blocking receive. */
         {
             PIOSTRUCT strm = get_stream(DEREFHANDLE(args)->Get(0).AsObjPtr());
             char *base = (char*)DEREFHANDLE(args)->Get(1).AsObjPtr()->AsBytePtr();
-            unsigned int offset = get_C_ulong(taskData, DEREFHANDLE(args)->Get(2));
-            unsigned int length = get_C_ulong(taskData, DEREFHANDLE(args)->Get(3));
-            unsigned int peek = get_C_ulong(taskData, DEREFHANDLE(args)->Get(4));
-            unsigned int outOfBand = get_C_ulong(taskData, DEREFHANDLE(args)->Get(5));
+            unsigned int offset = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(2));
+            unsigned int length = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(3));
+            unsigned int peek = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(4));
+            unsigned int outOfBand = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(5));
             int flags = 0, recvd;
             socklen_t addrLen;
             struct sockaddr resultAddr;
@@ -1139,8 +1116,8 @@ TryAgain:
                 }
                 if (err == EWOULDBLOCK && c == 54 /* blocking */)
                 {
-                    processes->BlockAndRestart(taskData, outOfBand ? -1 : strm->device.sock,
-                        false, POLY_SYS_network);
+                    WaitNet waiter(strm->device.sock, outOfBand != 0);
+                    processes->BlockAndRestart(taskData, &waiter, false, POLY_SYS_network);
                     ASSERT(0); /* Must not have returned. */
                 }
                 else if (err != EINTR)
@@ -1150,7 +1127,7 @@ TryAgain:
         }
 
     case 55: /* Create a socket pair. */
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
         /* Not implemented. */
         raise_syscall(taskData, "socketpair not implemented", -WSAEAFNOSUPPORT);
 #else
@@ -1159,8 +1136,8 @@ TryAgain:
             Handle str_token2 = make_stream_entry(taskData);
             Handle pair;
             PIOSTRUCT strm1, strm2;
-            int stream_no1 = STREAMID(str_token1);
-            int stream_no2 = STREAMID(str_token2);
+            unsigned stream_no1 = STREAMID(str_token1);
+            unsigned stream_no2 = STREAMID(str_token2);
             int af = get_C_long(taskData, DEREFHANDLE(args)->Get(0));
             int type = get_C_long(taskData, DEREFHANDLE(args)->Get(1));
             int proto = get_C_long(taskData, DEREFHANDLE(args)->Get(2));
@@ -1211,20 +1188,18 @@ TryAgain:
 #endif
 
     case 56: /* Create a Unix socket address from a string. */
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
         /* Not implemented. */
         raise_syscall(taskData, "Unix addresses not implemented", -WSAEAFNOSUPPORT);
 #else
         {
             struct sockaddr_un addr;
-            int length;
             memset(&addr, 0, sizeof(addr));
             addr.sun_family = AF_UNIX;
 #ifdef HAVE_STRUCT_SOCKADDR_UN_SUN_LEN
             addr.sun_len = sizeof(addr); // Used in FreeBSD only.
 #endif
-            length = Poly_string_to_C(DEREFWORD(args),
-                        addr.sun_path, sizeof(addr.sun_path));
+            POLYUNSIGNED length = Poly_string_to_C(DEREFWORD(args), addr.sun_path, sizeof(addr.sun_path));
             if (length > (int)sizeof(addr.sun_path))
                 raise_syscall(taskData, "Address too long", ENAMETOOLONG);
             return SAVE(Buffer_to_Poly(taskData, (char*)&addr, sizeof(addr)));
@@ -1232,7 +1207,7 @@ TryAgain:
 #endif
 
     case 57: /* Get the file name from a Unix socket address. */
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
         /* Not implemented. */
         raise_syscall(taskData, "Unix addresses not implemented", -WSAEAFNOSUPPORT);
 #else
@@ -1258,7 +1233,7 @@ TryAgain:
             char msg[100];
             sprintf(msg, "Unknown net function: %d", c);
             raise_exception_string(taskData, EXC_Fail, msg);
-			return 0;
+            return 0;
         }
     }
 }
@@ -1389,34 +1364,6 @@ static Handle makeServEntry(TaskData *taskData, struct servent *serv)
     return result;
 }
 
-static Handle makeNetEntry(TaskData *taskData, struct netent *net)
-{
-    int i;
-    char **p;
-    Handle aliases, name, addrType, result, network;
-
-    /* Canonical name. */
-    name = SAVE(C_string_to_Poly(taskData, net->n_name));
-
-    /* Aliases. */
-    for (i=0, p = net->n_aliases; *p != NULL; p++, i++);
-    aliases = convert_string_list(taskData, i, net->n_aliases);
-
-    /* Address type. */
-    addrType = Make_unsigned(taskData, net->n_addrtype);
-
-    /* Protocol name. */
-    network = Make_unsigned(taskData, ntohl(net->n_net));
-
-    /* Make the result structure. */
-    result = ALLOC(4);
-    DEREFHANDLE(result)->Set(0, DEREFWORDHANDLE(name));
-    DEREFHANDLE(result)->Set(1, DEREFWORDHANDLE(aliases));
-    DEREFHANDLE(result)->Set(2, DEREFWORDHANDLE(addrType));
-    DEREFHANDLE(result)->Set(3, DEREFWORDHANDLE(network));
-    return result;
-}
-
 static Handle mkAftab(TaskData *taskData, void *arg, char *p)
 {
     struct af_tab_struct *af = (struct af_tab_struct *)p;
@@ -1447,7 +1394,7 @@ static Handle mkSktab(TaskData *taskData, void *arg, char *p)
 static Handle setSocketOption(TaskData *taskData, Handle args, int level, int opt)
 {
     PIOSTRUCT strm = get_stream(DEREFHANDLE(args)->Get(0).AsObjPtr());
-    int onOff = get_C_long(taskData, DEREFHANDLE(args)->Get(1));
+    int onOff = get_C_int(taskData, DEREFHANDLE(args)->Get(1));
     if (strm == NULL) raise_syscall(taskData, "Stream is closed", EBADF);
     if (setsockopt(strm->device.sock, level, opt,
         (char*)&onOff, sizeof(int)) != 0)
@@ -1486,9 +1433,9 @@ static Handle getSelectResult(TaskData *taskData, Handle args, int offset, fd_se
 {
     /* Construct the result vectors. */
     PolyObject *inVec = DEREFHANDLE(args)->Get(offset).AsObjPtr();
-    int nVec = OBJECT_LENGTH(inVec);
+    POLYUNSIGNED nVec = OBJECT_LENGTH(inVec);
     int nRes = 0;
-    int i;
+    POLYUNSIGNED i;
     for (i = 0; i < nVec; i++) {
         PIOSTRUCT strm = get_stream(inVec->Get(i).AsObjPtr());
         if (FD_ISSET(strm->device.sock, pFds)) nRes++;
@@ -1513,9 +1460,12 @@ static Handle getSelectResult(TaskData *taskData, Handle args, int offset, fd_se
    minus one.  */
 static Handle selectCall(TaskData *taskData, Handle args, int blockType)
 {
+    // We should check for interrupts even if we're not going to block.
+    processes->TestAnyEvents(taskData);
     fd_set readers, writers, excepts;
     struct timeval timeout;
-    int i, nVec, selectRes;
+    int selectRes;
+    POLYUNSIGNED i, nVec;
     Handle rdResult, wrResult, exResult, result;
     /* Set up the bitmaps for the select call from the arrays. */
     PolyObject *readVec = DEREFHANDLE(args)->Get(0).AsObjPtr();
@@ -1553,12 +1503,12 @@ static Handle selectCall(TaskData *taskData, Handle args, int blockType)
         case 0: /* Check the timeout. */
             {
             /* The time argument is an absolute time. */
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
             FILETIME ftTime, ftNow;
             /* Get the file time. */
             get_C_pair(taskData, DEREFHANDLE(args)->Get(3),
                 &ftTime.dwHighDateTime, &ftTime.dwLowDateTime);
-			GetSystemTimeAsFileTime(&ftNow);
+            GetSystemTimeAsFileTime(&ftNow);
             /* If the timeout time is earlier than the current time
                we must return, otherwise we block. */
             if (CompareFileTime(&ftTime, &ftNow) <= 0)
@@ -1585,7 +1535,7 @@ static Handle selectCall(TaskData *taskData, Handle args, int blockType)
 #endif
         }
         case 1: /* Block until one of the descriptors is ready. */
-            processes->BlockAndRestart(taskData, -1, false, POLY_SYS_network);
+            processes->BlockAndRestart(taskData, NULL, false, POLY_SYS_network);
             /*NOTREACHED*/
         case 2: /* Just a simple poll - drop through. */
             break;
@@ -1608,7 +1558,7 @@ class Networking: public RtsModule
 {
 public:
     virtual void Init(void);
-    virtual void Uninit(void);
+    virtual void Stop(void);
 };
 
 // Declare this.  It will be automatically added to the table.
@@ -1616,7 +1566,7 @@ static Networking networkingModule;
 
 void Networking::Init(void)
 {
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
 #ifdef USEWINSOCK2
 #define WINSOCK_MAJOR_VERSION   2
 #define WINSOCK_MINOR_VERSION   2
@@ -1637,9 +1587,9 @@ void Networking::Init(void)
 #endif
 }
 
-void Networking::Uninit(void)
+void Networking::Stop(void)
 {
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
     if (winsock_init) WSACleanup();
     winsock_init = 0;
 #endif
