@@ -18,10 +18,12 @@
     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
 */
-#ifdef WIN32
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#elif defined(_WIN32)
 #include "winconfig.h"
 #else
-#include "config.h"
+#error "No configuration file"
 #endif
 
 #ifdef HAVE_STDIO_H
@@ -52,12 +54,12 @@
 #include <sys/wait.h>
 #endif
 
-#if (defined(__CYGWIN__) || defined(WINDOWS_PC))
+#if (defined(__CYGWIN__) || defined(_WIN32))
 #include <process.h>
 #endif
 
 // Include this next before errors.h since in WinCE at least the winsock errors are defined there.
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
 #ifdef USEWINSOCK2
 #include <winsock2.h>
 #else
@@ -72,7 +74,6 @@
 #include "process_env.h"
 #include "arb.h"
 #include "mpoly.h"
-#include "errors.h"
 #include "gc.h"
 #include "scanaddrs.h"
 #include "polystring.h"
@@ -82,6 +83,7 @@
 #include "machine_dep.h"
 #include "processes.h"
 #include "locking.h"
+#include "errors.h"
 
 #include "poly_specific.h" // For the functions that have been moved.
 
@@ -92,7 +94,7 @@
 #define MAXPATHLEN MAX_PATH
 #endif
 
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
 #define ISPATHSEPARATOR(c)  ((c) == '\\' || (c) == '/')
 #define DEFAULTSEPARATOR    "\\"
 #else
@@ -100,6 +102,11 @@
 #define DEFAULTSEPARATOR    "/"
 #endif
 
+#ifdef _MSC_VER
+// Don't tell me about ISO C++ changes.
+#pragma warning(disable:4996)
+#endif
+
 // "environ" is declared in the headers on some systems but not all.
 // Oddly, declaring it within process_env_dispatch_c causes problems
 // on mingw where "environ" is actually a function.
@@ -113,9 +120,10 @@ static bool exiting = false;
 
 static PLock atExitLock; // Thread lock for above.
 
+
 Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
 {
-    int c = get_C_long(mdTaskData, DEREFWORDHANDLE(code));
+    int c = get_C_int(mdTaskData, DEREFWORDHANDLE(code));
     switch (c)
     {
     case 0: /* Return the program name. */
@@ -128,9 +136,9 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
         {
             char buff[MAXPATHLEN], *res;
             /* Get the string. */
-            int length =
+            POLYUNSIGNED length =
                 Poly_string_to_C(DEREFWORDHANDLE(args), buff, sizeof(buff));
-            if (length >= (int)sizeof(buff)) raise_syscall(mdTaskData, "Not Found", 0);
+            if (length >= sizeof(buff)) raise_syscall(mdTaskData, "Not Found", 0);
             res = getenv(buff);
             if (res == NULL) raise_syscall(mdTaskData, "Not Found", 0);
             else return SAVE(C_string_to_Poly(mdTaskData, res));
@@ -155,18 +163,18 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
             char buff[MAXPATHLEN];
             /* Get the string. */
             int res = -1;
-            int length =
+            POLYUNSIGNED length =
                 Poly_string_to_C(DEREFWORD(args), buff, sizeof(buff));
-            if (length >= (int)sizeof(buff))
+            if (length >= sizeof(buff))
                 raise_syscall(mdTaskData, "Command too long", ENAMETOOLONG);
-#if (defined(WINDOWS_PC))
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
             // Windows.
             char *argv[4];
             argv[0] = getenv("COMSPEC"); // Default CLI.
             if (argv[0] == 0)
             {
-                if (_osver & 0x8000) argv[0] = "command.com"; // Win 95 etc.
-                else argv[0] = "cmd.exe"; // Win NT etc.
+                if (GetVersion() & 0x80000000) argv[0] = (char*)"command.com"; // Win 95 etc.
+                else argv[0] = (char*)"cmd.exe"; // Win NT etc.
             }
             argv[1] = (char*)"/c";
             argv[2] = buff;
@@ -174,7 +182,7 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
             // If _P_NOWAIT is given the result is the process handle.
             // spawnvp does any necessary path searching if argv[0]
             // does not contain a full path.
-            int pid = spawnvp(_P_NOWAIT, argv[0], argv);
+            intptr_t pid = spawnvp(_P_NOWAIT, argv[0], argv);
             if (pid == -1)
                 raise_syscall(mdTaskData, "Function system failed", errno);
 #else
@@ -206,11 +214,6 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
                 sigemptyset(&sigset);
                 sigprocmask(SIG_SETMASK, &sigset, 0);
                 // Reset other signals?
-                char *argv[4];
-                argv[0] = (char*)"sh";
-                argv[1] = (char*)"-c";
-                argv[2] = buff;
-                argv[3] = NULL;
                 execve("/bin/sh", argv, environ);
                 _exit(1);
             }
@@ -221,7 +224,7 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
                 try
                 {
                 // Test to see if the child has returned.
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
                     switch (WaitForSingleObject((HANDLE)pid, 0))
                     {
                     case WAIT_OBJECT_0:
@@ -236,6 +239,9 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
                     case WAIT_FAILED:
                         raise_syscall(mdTaskData, "Function system failed", -(int)GetLastError());
                     }
+                    // Wait for the process to exit or for the timeout
+                    WaitHandle waiter((HANDLE)pid);
+                    processes->ThreadPauseForIO(mdTaskData, &waiter);
 #else
                     int wRes = waitpid(pid, &res, WNOHANG);
                     if (wRes > 0)
@@ -244,15 +250,19 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
                     {
                         raise_syscall(mdTaskData, "Function system failed", errno);
                     }
-#endif
+                    // In Unix the best we can do is wait.  This may be interrupted
+                    // by SIGCHLD depending on where signals are processed.
+                    // One possibility is for the main thread to somehow wake-up
+                    // the thread when it processes a SIGCHLD.
                     processes->ThreadPause(mdTaskData);
+#endif
                 }
                 catch (...)
                 {
                     // Either IOException or KillException.
                     // We're abandoning the wait.  This will leave
                     // a zombie in Unix.
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
                     CloseHandle((HANDLE)pid);
 #endif
                     throw;
@@ -292,8 +302,8 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
         {
             /* I don't like terminating without some sort of clean up
                but we'll do it this way for the moment. */
-            int i = get_C_long(mdTaskData, DEREFWORDHANDLE(args));
-			_exit(i);
+            int i = get_C_int(mdTaskData, DEREFWORDHANDLE(args));
+            _exit(i);
         }
 
         /************ Error codes **************/
@@ -301,17 +311,16 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
     case 2: /* Get the name of a numeric error message. */
         {
             char buff[40];
-            int e = get_C_long(mdTaskData, DEREFWORDHANDLE(args));
+            int e = get_C_int(mdTaskData, DEREFWORDHANDLE(args));
             Handle  res;
-            unsigned i;
             /* First look to see if we have the name in
                the error table. They should generally all be
                there. */
-            for (i = 0; i < sizeof(errortable)/sizeof(errortable[0]); i++)
-                if (errortable[i].errorNum == e)
-                    return SAVE(C_string_to_Poly(mdTaskData, errortable[i].errorString));
+            const char *errorMsg = stringFromErrorCode(e);
+            if (errorMsg != NULL)
+                return SAVE(C_string_to_Poly(mdTaskData, errorMsg));
             /* We get here if there's an error which isn't in the table. */
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
             /* In the Windows version we may have both errno values
                and also GetLastError values.  We convert the latter into
                negative values before returning them. */
@@ -332,42 +341,18 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
 
     case 3: /* Get the explanatory message for an error. */
         {
-            int e = get_C_long(mdTaskData, DEREFWORDHANDLE(args));
-            Handle  res;
-#ifdef WINDOWS_PC
-            /* In the Windows version we may have both errno values
-               and also GetLastError values.  We convert the latter into
-               negative values before returning them. */
-            if (e < 0)
-            {
-                LPTSTR lpMsg = NULL;
-                TCHAR *p;
-                if (FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM |
-                        FORMAT_MESSAGE_ALLOCATE_BUFFER |
-                        FORMAT_MESSAGE_IGNORE_INSERTS,
-                        NULL, (DWORD)(-e), 0, (LPTSTR)&lpMsg, 1, NULL) > 0)
-                {
-                    /* The message is returned with CRLF at the end.  Remove them. */
-                    for (p = lpMsg; *p != '\0' && *p != '\n' && *p != '\r'; p++);
-                    *p = '\0';
-                    res = SAVE(C_string_to_Poly(mdTaskData, lpMsg));
-                    LocalFree(lpMsg);
-                    return res;
-                }
-            }
-#endif
-            res = SAVE(C_string_to_Poly(mdTaskData, strerror(e)));
-            return res;
+            return errorMsg(mdTaskData, get_C_int(mdTaskData, DEREFWORDHANDLE(args)));
         }
+
     case 4: /* Try to convert an error string to an error number. */
         {
             char buff[40];
             /* Get the string. */
             Poly_string_to_C(DEREFWORD(args), buff, sizeof(buff));
             /* Look the string up in the table. */
-            for (unsigned i = 0; i < sizeof(errortable)/sizeof(errortable[0]); i++)
-                if (strcmp(buff, errortable[i].errorString) == 0)
-                    return Make_arbitrary_precision(mdTaskData, errortable[i].errorNum);
+            int err = 0;
+            if (errorCodeFromString(buff, &err))
+                return Make_arbitrary_precision(mdTaskData, err);
             /* If we don't find it then it may have been a constructed
                error name. */
             if (strncmp(buff, "ERROR", 5) == 0)
@@ -375,7 +360,7 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
                 int i = atoi(buff+5);
                 if (i > 0) return Make_arbitrary_precision(mdTaskData, i);
             }
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
             if (strncmp(buff, "WINERROR", 8) == 0)
             {
                 int i = atoi(buff+8);
@@ -399,14 +384,14 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
 
     case 8: /* Test the character to see if it matches a separator. */
         {
-            int e = get_C_long(mdTaskData, DEREFWORDHANDLE(args));
+            int e = get_C_int(mdTaskData, DEREFWORDHANDLE(args));
             if (ISPATHSEPARATOR(e))
                 return Make_arbitrary_precision(mdTaskData, 1);
             else return Make_arbitrary_precision(mdTaskData, 0);
         }
 
     case 9: /* Are names case-sensitive? */
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
         /* Windows - no. */
         return Make_arbitrary_precision(mdTaskData, 0);
 #else
@@ -433,10 +418,9 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
                (e.g. A:b\c) which means the file b\c relative to the
                currently selected directory on the volume A.
             */
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
             char buff[MAXPATHLEN];
-            int length;
-            length = Poly_string_to_C(path, buff, MAXPATHLEN);
+            POLYUNSIGNED length = Poly_string_to_C(path, buff, MAXPATHLEN);
             /* Ignore the case where the whole path is too long. */
             if (length >= 2 && buff[1] == ':')
             { /* Volume name? */
@@ -498,7 +482,7 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
     case 12: /* Construct a name from a volume and whether it is
                 absolute. */
         {
-            unsigned isAbs = get_C_ulong(mdTaskData, DEREFHANDLE(args)->Get(1));
+            unsigned isAbs = get_C_unsigned(mdTaskData, DEREFHANDLE(args)->Get(1));
             PolyWord volName = DEREFHANDLE(args)->Get(0);
             /* In Unix the volume name will always be empty. */
             if (isAbs == 0)
@@ -516,7 +500,7 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
             if (IS_INT(volName))
             {
                 char ch = (char)UNTAGGED(volName);
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
                 if (ch == '<' || ch == '>' || ch == '|' ||
                     ch == '"' || ch < ' ')
 #else
@@ -532,7 +516,7 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
                 for (POLYUNSIGNED i = 0; i < volume->length; i++)
                 {
                     char ch = volume->chars[i];
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
                     if (ch == '<' || ch == '>' || ch == '|' ||
                         ch == '"' || ch < ' ')
 #else
@@ -554,7 +538,7 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
             char msg[100];
             sprintf(msg, "Unknown environment function: %d", c);
             raise_exception_string(mdTaskData, EXC_Fail, msg);
-			return 0;
+            return 0;
         }
     }
 }
@@ -562,7 +546,7 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
 /* Terminate normally with a result code. */
 Handle finishc(TaskData *taskData, Handle h)
 {
-    int i = get_C_long(taskData, DEREFWORDHANDLE(h));
+    int i = get_C_int(taskData, DEREFWORDHANDLE(h));
     // Cause the other threads to exit.
     processes->Exit(i);
     // Exit this thread
