@@ -4,6 +4,7 @@
 
     Copyright (c) 2000-7
         Cambridge University Technical Services Limited
+    Further development copyright (c) David C.J. Matthews 2011
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -21,10 +22,12 @@
 
 */
 
-#ifdef WIN32
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#elif defined(_WIN32)
 #include "winconfig.h"
 #else
-#include "config.h"
+#error "No configuration file"
 #endif
 
 #ifdef HAVE_STDIO_H
@@ -92,7 +95,6 @@
 #include "profiling.h"
 #include "save_vec.h"
 #include "rts_module.h"
-#include "gc.h" // For gc_phase
 #include "memmgr.h"
 #include "scanaddrs.h"
 #include "locking.h"
@@ -102,19 +104,56 @@
 /*      PROFILING SHARED DATA                                                 */
 /*                                                                            */
 /******************************************************************************/
-static POLYUNSIGNED gc_count1 = 0;
-static POLYUNSIGNED gc_count2 = 0;
-static POLYUNSIGNED gc_count3 = 0;
+static POLYUNSIGNED mainThreadCounts[MTP_MAXENTRY];
+static const char* mainThreadText[MTP_MAXENTRY] =
+{
+    "UNKNOWN",
+    "GARBAGE COLLECTION (sharing phase)",
+    "GARBAGE COLLECTION (mark phase)",
+    "GARBAGE COLLECTION (copy phase)",
+    "GARBAGE COLLECTION (update phase)",
+    "GARBAGE COLLECTION (minor collection)",
+    "Common data sharing",
+    "Exporting",
+    "Saving state",
+    "Loading saved state",
+    "Profiling",
+    "Setting signal handler"
+};
+
+// Entries for store profiling
+enum _extraStore {
+    EST_CODE = 0,
+    EST_STRING,
+    EST_BYTE,
+    EST_WORD,
+    EST_MUTABLE,
+    EST_MUTABLEBYTE,
+    EST_MAX_ENTRY
+};
+
+static POLYUNSIGNED extraStoreCounts[EST_MAX_ENTRY];
+static const char * extraStoreText[EST_MAX_ENTRY] =
+{
+    "Function code",
+    "Strings",
+    "Byte data (long precision ints etc)",
+    "Unidentified word data",
+    "Unidentified mutable data",
+    "Mutable byte data (profiling counts)"
+};
+
 static POLYUNSIGNED total_count = 0;
-static POLYUNSIGNED unknown_count  = 0;
+
+/* This type is coercible to PolyString *. We can use static "pstrings"
+   since we are not going to garbage collect while we are printing out
+   the profiling information. */
+static struct {
+    POLYUNSIGNED length; char chars[40];
+} psStrings[MTP_MAXENTRY+EST_MAX_ENTRY], psGCTotal;
 
 ProfileMode profileMode;
 
-/* If Accumulated is added to one of the above profiling is only turned off
-   by the next call with PROF_ACCUMULATED+PROF_OFF.  This allows us to profile
-   the compiler. */
-#define PROF_ACCUMULATED    4
-
 /******************************************************************************/
 /*                                                                            */
 /*      PROFILING FUNCTIONS                                                   */
@@ -146,17 +185,31 @@ static struct
 // running then.
 static PLock countLock;
 
+// Get the profile object associated with a piece of code.  Returns null if
+// there isn't one, in particular if this is in the old format.
+static PolyObject *getProfileObjectForCode(PolyObject *code)
+{
+    ASSERT(code->IsCodeObject());
+    PolyWord *consts;
+    POLYUNSIGNED constCount;
+    code->GetConstSegmentForCode(consts, constCount);
+    if (constCount < 3 || ! consts[2].IsDataPtr()) return 0;
+    PolyObject *profObject = consts[2].AsObjPtr();
+    if (profObject->IsMutable() && profObject->IsByteObject() && profObject->Length() == 1)
+        return profObject;
+    else return 0;
+}
+
 // Adds incr to the profile count for the function pointed at by
 // pc or by one of its callers.
 // This is called from a signal handler in the case of time profiling.
-void add_count(TaskData *taskData, POLYCODEPTR fpc, PolyWord *sp, int incr)
+void add_count(TaskData *taskData, POLYCODEPTR fpc, PolyWord *sp, POLYUNSIGNED incr)
 {
 
     /* The first PC may be valid even if it's not a code pointer */
     bool is_code = true;
     PolyWord pc = PolyWord::FromCodePtr(fpc);
-    StackObject *stack = taskData->stack;
-    PolyWord *endStack = stack->Offset(stack->Length());
+    PolyWord *endStack = taskData->stack->top;
     
     /* First try the pc value we have been given - if that fails search down
        the stack to see if there is a return address we can use. */
@@ -172,16 +225,12 @@ void add_count(TaskData *taskData, POLYCODEPTR fpc, PolyWord *sp, int incr)
             MemSpace *space = gMem.SpaceForAddress(pc.AsAddress());
             if (space != 0)
             {
-                PolyObject *ptr = ObjCodePtrToPtr(pc.AsCodePtr());
-                ASSERT(ptr->IsCodeObject());
-                PolyWord *consts = ptr->ConstPtrForCode();
-                if (consts[0] != TAGGED(0)) // Skip anonymous code.
-                {
-                    PLocker locker(&countLock);
-                    ((POLYUNSIGNED*)consts)[-1] += incr;
-                    total_count += incr;
-                    return;
-                }
+                PolyObject *profObject = getProfileObjectForCode(ObjCodePtrToPtr(pc.AsCodePtr()));
+                PLocker locker(&countLock);
+                if (profObject)
+                    profObject->Set(0, PolyWord::FromUnsigned(profObject->Get(0).AsUnsigned() + incr));
+                total_count += incr;
+                return;
             }
             /* else just fall through and try next candidate address */
         } /* OBJ_IS_CODEPTR(pc) */
@@ -192,7 +241,7 @@ void add_count(TaskData *taskData, POLYCODEPTR fpc, PolyWord *sp, int incr)
         else /* Reached bottom of stack without finding valid code pointer */
         {
             PLocker locker(&countLock);
-            unknown_count += incr;
+            mainThreadCounts[MTP_USER_CODE] += incr;
             total_count += incr;
             return;
         }
@@ -266,7 +315,7 @@ static void writeProfileResults(void)
         for (i = 0; i < P.used; i++)
         {
             PPROFENTRY pEnt = &P.pTab[i];
-            fprintf(stdout, "%10lu ", pEnt->count);
+            fprintf(stdout, "%10" POLYUFMT " ", pEnt->count);
             print_string(pEnt->functionName);
             putc ('\n', stdout);
         }
@@ -298,42 +347,36 @@ static void PrintProfileCounts(PolyWord *bottom, PolyWord *top)
         else
         {
             ASSERT(obj->ContainsNormalLengthWord());
-            
+
             if (obj->IsCodeObject())
             {
                 PolyWord *firstConstant = obj->ConstPtrForCode();
                 PolyWord name = firstConstant[0];
-                // The word before is an untagged count
-                POLYUNSIGNED    count = firstConstant[-1].AsUnsigned();
-                
-                if (count != 0)
+                PolyObject *profCount = getProfileObjectForCode(obj);
+                if (profCount)
                 {
-                    if (name != TAGGED(0))
+                    POLYUNSIGNED count = profCount->Get(0).AsUnsigned();
+                
+                    if (count != 0)
                     {
-                        PPROFENTRY pEnt = newProfileEntry();
-                        pEnt->count = count;
-                        // This should be a valid string which may be a single character.
-                        ASSERT(name.IsTagged() || name.AsObjPtr()->IsByteObject());
-                        pEnt->functionName = name;
-                    }
+                        if (name != TAGGED(0))
+                        {
+                            PPROFENTRY pEnt = newProfileEntry();
+                            pEnt->count = count;
+                            pEnt->functionName = name;
+                        }
                     
-                    firstConstant[-1] = PolyWord::FromUnsigned(0);
-                    P.total += count;
-                } /* count != 0 */
+                        profCount->Set(0, PolyWord::FromUnsigned(0));
+                        P.total += count;
+                    }
+                }
             } /* code object */
             ptr += obj->Length();
         } /* else */
     } /* while */
 }
 
-/* This type is coercible to PolyString *. We can use static "pstrings"
-   since we are not going to garbage collect while we are printing out
-   the profiling information. */
-static struct {
-    POLYUNSIGNED length; char chars[40];
-} psMarkPhase, psCopyPhase, psUpdatePhase, psGCTotal, psUnknown;
-
-static void printprofile(void)
+void printprofile(void)
 /* Print profiling information and reset profile counts.    */
 /* Profile counts are also reset by commit so that objects  */
 /* written to the persistent store always have zero counts. */
@@ -358,63 +401,64 @@ static void printprofile(void)
         {
             LocalMemSpace *space = gMem.lSpaces[j];
             // Local areas only have objects from the allocation pointer to the top.
-            PrintProfileCounts(space->pointer, space->top);
+            PrintProfileCounts(space->bottom, space->lowerAllocPtr);
+            PrintProfileCounts(space->upperAllocPtr, space->top);
         }
     } // else if we haven't actually had an interrupt avoid expensive scan of memory.
-    
-    if (gc_count1 || gc_count2 || gc_count3)
+
     {
-        int gc_count = gc_count1 + gc_count2 + gc_count3;
-        P.total     += gc_count;
-        total_count += gc_count;
-        
-        pEnt = newProfileEntry();
-        strcpy(psMarkPhase.chars, "GARBAGE COLLECTION (mark phase)");
-        psMarkPhase.length = strlen(psMarkPhase.chars);
-        pEnt->count = gc_count1;
-        pEnt->functionName = PolyWord::FromUnsigned((POLYUNSIGNED)&psMarkPhase);
-        
-        pEnt = newProfileEntry();
-        strcpy(psCopyPhase.chars, "GARBAGE COLLECTION (copy phase)");
-        psCopyPhase.length = strlen(psCopyPhase.chars);
-        pEnt->count = gc_count2;
-        pEnt->functionName = PolyWord::FromUnsigned((POLYUNSIGNED)&psCopyPhase);
-        
-        pEnt = newProfileEntry();
-        strcpy(psUpdatePhase.chars, "GARBAGE COLLECTION (update phase)");
-        psUpdatePhase.length = strlen(psUpdatePhase.chars);
-        pEnt->count = gc_count3;
-        pEnt->functionName = PolyWord::FromUnsigned((POLYUNSIGNED)&psUpdatePhase);
-        
-        pEnt = newProfileEntry();
-        strcpy(psGCTotal.chars, "GARBAGE COLLECTION (total)");
-        psGCTotal.length = strlen(psGCTotal.chars);
-        pEnt->count = gc_count;
-        pEnt->functionName = PolyWord::FromUnsigned((POLYUNSIGNED)&psGCTotal);
-        
-        gc_count1 = 0;
-        gc_count2 = 0;
-        gc_count3 = 0;
+        POLYUNSIGNED gc_count =
+            mainThreadCounts[MTP_GCPHASESHARING]+
+            mainThreadCounts[MTP_GCPHASEMARK]+
+            mainThreadCounts[MTP_GCPHASECOMPACT] +
+            mainThreadCounts[MTP_GCPHASEUPDATE] +
+            mainThreadCounts[MTP_GCQUICK];
+        if (gc_count)
+        {
+            pEnt = newProfileEntry();
+            strcpy(psGCTotal.chars, "GARBAGE COLLECTION (total)");
+            psGCTotal.length = strlen(psGCTotal.chars);
+            pEnt->count = gc_count;
+            pEnt->functionName = PolyWord::FromUnsigned((POLYUNSIGNED)&psGCTotal);
+        }
     }
-    
-    if (unknown_count)
+
+    for (unsigned k = 0; k < MTP_MAXENTRY; k++)
     {
-        total_count += unknown_count;
-        P.total += unknown_count;
-        pEnt = newProfileEntry();
-        strcpy(psUnknown.chars, "UNKNOWN");
-        psUnknown.length = strlen(psUnknown.chars);
-        pEnt->count = unknown_count;
-        pEnt->functionName = PolyWord::FromUnsigned((POLYUNSIGNED)&psUnknown);
-        unknown_count = 0;
+        if (mainThreadCounts[k])
+        {
+            P.total     += mainThreadCounts[k];
+            total_count += mainThreadCounts[k];
+            pEnt = newProfileEntry();
+            strcpy(psStrings[k].chars, mainThreadText[k]);
+            psStrings[k].length = strlen(psStrings[k].chars);
+            pEnt->count = mainThreadCounts[k];
+            pEnt->functionName = PolyWord::FromUnsigned((POLYUNSIGNED)&psStrings[k]);
+            mainThreadCounts[k] = 0;
+        }
     }
-    
+
+    for (unsigned l = 0; l < EST_MAX_ENTRY; l++)
+    {
+        if (extraStoreCounts[l])
+        {
+            P.total     += extraStoreCounts[l];
+            total_count += extraStoreCounts[l];
+            pEnt = newProfileEntry();
+            strcpy(psStrings[MTP_MAXENTRY+l].chars, extraStoreText[l]);
+            psStrings[MTP_MAXENTRY+l].length = strlen(psStrings[MTP_MAXENTRY+l].chars);
+            pEnt->count = extraStoreCounts[l];
+            pEnt->functionName = PolyWord::FromUnsigned((POLYUNSIGNED)&psStrings[MTP_MAXENTRY+l]);
+            extraStoreCounts[l] = 0;
+        }
+    }
+
     writeProfileResults();
     
     if (total_count)
     {
         
-        printf("\nTotal: %lu; Counted: %lu; Uncounted: %lu",
+        printf("\nTotal: %" POLYUFMT "; Counted: %" POLYUFMT "; Uncounted: %" POLYUFMT,
             total_count, P.total, total_count - P.total);
         
         total_count = 0;
@@ -429,46 +473,80 @@ void handleProfileTrap(TaskData *taskData, SIGNALCONTEXT *context)
 {
     /* If we are in the garbage-collector add the count to "gc_count"
         otherwise try to find out where we are. */
-    switch (gc_phase)
+    if (mainThreadPhase == MTP_USER_CODE)
     {
-    case 0 : 
+        if (taskData)
         {
-            if (taskData)
-            {
-                PolyWord *sp;
-                POLYCODEPTR pc;
-                if (machineDependent->GetPCandSPFromContext(taskData, context, sp, pc))
-                    add_count(taskData, pc, sp, 1);
-                else unknown_count++;
-            }
-            else unknown_count++;
-            // On Mac OS X all virtual timer interrupts seem to be directed to the root thread
-            // so all the counts will be "unknown".
+            PolyWord *sp;
+            POLYCODEPTR pc;
+            if (machineDependent->GetPCandSPFromContext(taskData, context, sp, pc))
+                add_count(taskData, pc, sp, 1);
+            else mainThreadCounts[MTP_USER_CODE]++;
         }
-        break;
-        
-    case 1 :
-        gc_count1++;
-        break;
-        
-    case 2 :
-        gc_count2++;
-        break;
-        
-    case 3 :
-        gc_count3++;
-        break;
-        
-    default :
-        unknown_count++;
-        break;
+        else mainThreadCounts[MTP_USER_CODE]++;
+        // On Mac OS X all virtual timer interrupts seem to be directed to the root thread
+        // so all the counts will be "unknown".
     }
+    else mainThreadCounts[mainThreadPhase]++;
 }
 
+// Called from the GC when allocation profiling is on.
+void AddObjectProfile(PolyObject *obj)
+{
+    ASSERT(obj->ContainsNormalLengthWord());
+    POLYUNSIGNED length = obj->Length();
+
+    if (obj->IsWordObject() && OBJ_HAS_PROFILE(obj->LengthWord()))
+    {
+        // It has a profile pointer.  The last word should point to the
+        // closure or code of the allocating function.  Add the size of this to the count.
+        ASSERT(length != 0);
+        PolyWord profWord = obj->Get(length-1);
+        ASSERT(profWord.IsDataPtr());
+        PolyObject *profObject = profWord.AsObjPtr();
+        ASSERT(profObject->IsMutable() && profObject->IsByteObject() && profObject->Length() == 1);
+        profObject->Set(0, PolyWord::FromUnsigned(profObject->Get(0).AsUnsigned() + length + 1));
+        total_count += length+1;
+    }
+    // If it doesn't have a profile pointer add it to the appropriate count.
+    else if (obj->IsMutable())
+    {
+        if (obj->IsByteObject())
+            extraStoreCounts[EST_MUTABLEBYTE] += length+1;
+        else extraStoreCounts[EST_MUTABLE] += length+1;
+    }
+    else if (obj->IsCodeObject())
+        extraStoreCounts[EST_CODE] += length+1;
+    else if (obj->IsByteObject())
+    {
+        // Try to separate strings from other byte data.  This is only
+        // approximate.
+        if (OBJ_IS_NEGATIVE(obj->LengthWord()))
+            extraStoreCounts[EST_BYTE] += length+1;
+        else
+        {
+            PolyStringObject *possString = (PolyStringObject*)obj;
+            POLYUNSIGNED bytes = length * sizeof(PolyWord);
+            // If the length of the string as given in the first word is sufficient
+            // to fit in the exact number of words then it's probably a string.
+            if (length >= 2 &&
+                possString->length <= bytes - sizeof(POLYUNSIGNED) &&
+                possString->length > bytes - 2 * sizeof(POLYUNSIGNED))
+                    extraStoreCounts[EST_STRING] += length+1;
+            else
+            {
+                extraStoreCounts[EST_BYTE] += length+1;
+            }
+        }
+    }
+    else extraStoreCounts[EST_WORD] += length+1;
+}
+
+
 class ProfileRequest: public MainThreadRequest
 {
 public:
-    ProfileRequest(unsigned prof): mode(prof) {}
+    ProfileRequest(unsigned prof): MainThreadRequest(MTP_PROFILING), mode(prof) {}
 
     virtual void Perform();
     unsigned mode;
@@ -490,17 +568,12 @@ Handle profilerc(TaskData *taskData, Handle mode_handle)
    if the parameter is 2 it produces store profiling.
    3 - arbitrary precision emulation traps. */
 {
-    unsigned mode = get_C_ulong(taskData, DEREFWORDHANDLE(mode_handle));
+    unsigned mode = get_C_unsigned(taskData, DEREFWORDHANDLE(mode_handle));
     {
         PLocker locker(&profLock);
         if (mode == profile_mode) // No change in mode = no-op
             return taskData->saveVec.push(TAGGED(0));
     
-        /* profiling 0 turns off profiling and prints results,
-           but is ignored for profile_modes > 4. To get the
-           accumulated results of these, set profiling to 4. */
-        if (mode == 0 && profile_mode > PROF_ACCUMULATED)
-            return taskData->saveVec.push(TAGGED(0));
         profile_mode = mode;
     }
     // All these actions are performed by the root thread.  Only profile
@@ -514,7 +587,7 @@ Handle profilerc(TaskData *taskData, Handle mode_handle)
 // This is called from the root thread when all the ML threads have been paused.
 void ProfileRequest::Perform()
 {
-    switch (mode & ~PROF_ACCUMULATED)
+    switch (mode)
     {
     case kProfileOff:
         // Turn off old profiling mechanism and print out accumulated results 
@@ -535,7 +608,15 @@ void ProfileRequest::Perform()
     case kProfileEmulation:
         profileMode = kProfileEmulation;
         break;
-        
+
+    case kProfileLiveData:
+        profileMode = kProfileLiveData;
+        break;
+ 
+    case kProfileLiveMutables:
+        profileMode = kProfileLiveMutables;
+        break;
+       
     default: /* do nothing */
         break;
     }
@@ -556,9 +637,6 @@ void Profiling::Init(void)
 {
     // Reset profiling counts.
     profileMode = kProfileOff;
-    gc_count1 = 0;
-    gc_count2 = 0;
-    gc_count3 = 0;
+    for (unsigned k = 0; k < MTP_MAXENTRY; k++) mainThreadCounts[k] = 0;
     total_count = 0;
-    unknown_count  = 0;
 }
