@@ -5,6 +5,8 @@
     Copyright (c) 2000
         Cambridge University Technical Services Limited
 
+    Further work copyright David C.J. Matthews 2011
+
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
     License as published by the Free Software Foundation; either
@@ -21,10 +23,12 @@
 
 */
 
-#ifdef WIN32
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#elif defined(_WIN32)
 #include "winconfig.h"
 #else
-#include "config.h"
+#error "No configuration file"
 #endif
 
 #ifdef HAVE_IEEEFP_H
@@ -40,7 +44,7 @@
 #include <fenv.h>
 #endif
 
-#if (defined(WIN32))
+#if (defined(_MSC_VER))
 #define finite  _finite
 #define isnan   _isnan
 #define copysign _copysign
@@ -96,6 +100,7 @@
     DCJM: March 2000.
 */
 
+// Positive and negative infinities and (positive) NaN.
 double posInf, negInf, notANumber;
 
 /* Real numbers are represented by the address of the value. */
@@ -161,6 +166,16 @@ Handle Real_divc(TaskData *mdTaskData, Handle y, Handle x)
     return real_result(mdTaskData, dx/dy);
 }
 
+/* CALL_IO1(Real_abs, REF, NOIND) */
+// This handles +/- NaN correctly.
+// TODO: That assumes that notANumber is positive which may not be true.
+Handle Real_absc(TaskData *mdTaskData, Handle x)
+{
+    double dx = real_arg(x);
+    if (isnan(dx)) return real_result(mdTaskData, notANumber);
+    else return real_result(mdTaskData, fabs(real_arg(x)));
+}
+
 /* CALL_IO1(Real_neg, REF, NOIND) */
 Handle Real_negc(TaskData *mdTaskData, Handle x)
 {
@@ -217,8 +232,8 @@ Handle Real_neqc(TaskData *mdTaskData, Handle y, Handle x)
 /* CALL_IO1(Real_float, REF, NOIND) */
 Handle Real_floatc(TaskData *mdTaskData, Handle x) /* SHORT int to real */
 {
-    POLYSIGNED n = UNTAGGED(DEREFWORDHANDLE(x));
-    return real_result(mdTaskData, (double)n);
+    double d = get_C_real(mdTaskData, DEREFWORDHANDLE(x));
+    return real_result(mdTaskData, d);
 }
 
 /* CALL_IO1(Real_int, REF, NOIND) */
@@ -361,7 +376,7 @@ Handle Real_convc(TaskData *mdTaskData, Handle str) /* string to real */
     }
         
     /* Now convert it */
-    result = strtod(string_buffer, &finish);
+    result = poly_strtod(string_buffer, &finish);
     bool isError = *finish != '\0'; // Test before deallocating
     free(string_buffer);
     // We no longer detect overflow and underflow and instead return
@@ -416,44 +431,49 @@ static Handle powerOf(TaskData *mdTaskData, Handle args)
     return real_result(mdTaskData, pow(x, y));
 }
 
+#define POLY_ROUND_TONEAREST    0
+#define POLY_ROUND_DOWNWARD     1
+#define POLY_ROUND_UPWARD       2
+#define POLY_ROUND_TOZERO       3
+
 // It would be nice to be able to use autoconf to test for these as functions
 // but they are frequently inlined 
 #if defined(HAVE_FENV_H)
 // C99 version.  This is becoming the most common.
-static PolyWord getrounding(TaskData *)
+static int getrounding(TaskData *)
 {
     switch (fegetround())
     {
-    case FE_TONEAREST: return TAGGED(0);
-    case FE_DOWNWARD: return TAGGED(1);
-    case FE_UPWARD: return TAGGED(2);
-    case FE_TOWARDZERO: return TAGGED(3);
+    case FE_TONEAREST: return POLY_ROUND_TONEAREST;
+    case FE_DOWNWARD: return POLY_ROUND_DOWNWARD;
+    case FE_UPWARD: return POLY_ROUND_UPWARD;
+    case FE_TOWARDZERO: return POLY_ROUND_TOZERO;
     }
-    return TAGGED(0); // Keep the compiler happy
+    return 0; // Keep the compiler happy
 }
 
 static void setrounding(TaskData *taskData, Handle args)
 {
     switch (get_C_long(taskData, DEREFWORDHANDLE(args)))
     {
-    case 0: fesetround(FE_TONEAREST); break; // Choose nearest
-    case 1: fesetround(FE_DOWNWARD); break; // Towards negative infinity
-    case 2: fesetround(FE_UPWARD); break; // Towards positive infinity
-    case 3: fesetround(FE_TOWARDZERO); break; // Truncate towards zero
+    case POLY_ROUND_TONEAREST: fesetround(FE_TONEAREST); break; // Choose nearest
+    case POLY_ROUND_DOWNWARD: fesetround(FE_DOWNWARD); break; // Towards negative infinity
+    case POLY_ROUND_UPWARD: fesetround(FE_UPWARD); break; // Towards positive infinity
+    case POLY_ROUND_TOZERO: fesetround(FE_TOWARDZERO); break; // Truncate towards zero
     }
 }
 
 #elif (defined(HAVE_IEEEFP_H) && ! defined(__CYGWIN__))
 // Older FreeBSD.  Cygwin has the ieeefp.h header but not the functions!
-static PolyWord getrounding(TaskData *)
+static int getrounding(TaskData *)
 {
     switch (fpgetround())
     {
-    case FP_RN: return TAGGED(0);
-    case FP_RM: return TAGGED(1);
-    case FP_RP: return TAGGED(2);
-    case FP_RZ: return TAGGED(3);
-    default: return TAGGED(0); /* Shouldn't happen. */ 
+    case FP_RN: return POLY_ROUND_TONEAREST;
+    case FP_RM: return POLY_ROUND_DOWNWARD;
+    case FP_RP: return POLY_ROUND_UPWARD;
+    case FP_RZ: return POLY_ROUND_TOZERO;
+    default: return 0; /* Shouldn't happen. */ 
     }
 }
 
@@ -461,52 +481,52 @@ static void setrounding(TaskData *taskData, Handle args)
 {
     switch (get_C_long(taskData, DEREFWORDHANDLE(args)))
     {
-    case 0: fpsetround(FP_RN); break; /* Choose nearest */
-    case 1: fpsetround(FP_RM); break; /* Towards negative infinity */
-    case 2: fpsetround(FP_RP); break; /* Towards positive infinity */
-    case 3: fpsetround(FP_RZ); break; /* Truncate towards zero */
+    case POLY_ROUND_TONEAREST: fpsetround(FP_RN); break; /* Choose nearest */
+    case POLY_ROUND_DOWNWARD: fpsetround(FP_RM); break; /* Towards negative infinity */
+    case POLY_ROUND_UPWARD: fpsetround(FP_RP); break; /* Towards positive infinity */
+    case POLY_ROUND_TOZERO: fpsetround(FP_RZ); break; /* Truncate towards zero */
     }
 }
 
-#elif defined(WINDOWS_PC)
+#elif defined(_WIN32)
 // Windows version
-static PolyWord getrounding(TaskData *)
+static int getrounding(TaskData *)
 {
     switch (_controlfp(0,0) & _MCW_RC)
     {
-    case _RC_NEAR: return TAGGED(0);
-    case _RC_DOWN: return TAGGED(1);
-    case _RC_UP: return TAGGED(2);
-    case _RC_CHOP: return TAGGED(3);
+    case _RC_NEAR: return POLY_ROUND_TONEAREST;
+    case _RC_DOWN: return POLY_ROUND_DOWNWARD;
+    case _RC_UP: return POLY_ROUND_UPWARD;
+    case _RC_CHOP: return POLY_ROUND_TOZERO;
     }
-    return TAGGED(0); // Keep the compiler happy
+    return 0; // Keep the compiler happy
 }
 
 static void setrounding(TaskData *mdTaskData, Handle args)
 {
     switch (get_C_long(mdTaskData, DEREFWORDHANDLE(args)))
     {
-    case 0: _controlfp(_RC_NEAR, _MCW_RC); break; // Choose nearest
-    case 1: _controlfp(_RC_DOWN, _MCW_RC); break; // Towards negative infinity
-    case 2: _controlfp(_RC_UP, _MCW_RC); break; // Towards positive infinity
-    case 3: _controlfp(_RC_CHOP, _MCW_RC); break; // Truncate towards zero
+    case POLY_ROUND_TONEAREST: _controlfp(_RC_NEAR, _MCW_RC); break; // Choose nearest
+    case POLY_ROUND_DOWNWARD: _controlfp(_RC_DOWN, _MCW_RC); break; // Towards negative infinity
+    case POLY_ROUND_UPWARD: _controlfp(_RC_UP, _MCW_RC); break; // Towards positive infinity
+    case POLY_ROUND_TOZERO: _controlfp(_RC_CHOP, _MCW_RC); break; // Truncate towards zero
     }
 }
 
 #elif defined(_FPU_GETCW) && defined(_FPU_SETCW)
 // Older Linux version
-static PolyWord getrounding(TaskData *)
+static int getrounding(TaskData *)
 {
     fpu_control_t ctrl;
     _FPU_GETCW(ctrl);
     switch (ctrl & _FPU_RC_ZERO)
     {
-    case _FPU_RC_NEAREST: return TAGGED(0);
-    case _FPU_RC_DOWN: return TAGGED(1);
-    case _FPU_RC_UP: return TAGGED(2);
-    case _FPU_RC_ZERO: return TAGGED(4);
+    case _FPU_RC_NEAREST: return POLY_ROUND_TONEAREST;
+    case _FPU_RC_DOWN: return POLY_ROUND_DOWNWARD;
+    case _FPU_RC_UP: return POLY_ROUND_UPWARD;
+    case _FPU_RC_ZERO: return POLY_ROUND_TOZERO;
     }
-    return TAGGED(0); /* Never reached but this avoids warning message. */
+    return 0; /* Never reached but this avoids warning message. */
 }
 
 static void setrounding(TaskData *taskData, Handle args)
@@ -516,10 +536,10 @@ static void setrounding(TaskData *taskData, Handle args)
     ctrl &= ~_FPU_RC_ZERO; /* Mask off any existing rounding. */
     switch (get_C_long(taskData, DEREFWORDHANDLE(args)))
     {
-    case 0: ctrl |= _FPU_RC_NEAREST;
-    case 1: ctrl |= _FPU_RC_DOWN;
-    case 2: ctrl |= _FPU_RC_UP;
-    case 3: ctrl |= _FPU_RC_ZERO;
+    case POLY_ROUND_TONEAREST: ctrl |= _FPU_RC_NEAREST;
+    case POLY_ROUND_DOWNWARD: ctrl |= _FPU_RC_DOWN;
+    case POLY_ROUND_UPWARD: ctrl |= _FPU_RC_UP;
+    case POLY_ROUND_TOZERO: ctrl |= _FPU_RC_ZERO;
     }
     _FPU_SETCW(ctrl);
 }
@@ -533,40 +553,40 @@ static void getround(union db *res)
     __asm__ ("stfd f0,0(r3)");
 }
 
-static PolyWord getrounding(TaskData *)
+static int getrounding(TaskData *)
 {
     union db roundingRes;
     getround(&roundingRes);
     switch (roundingRes.wrd[1] & 3)
     {
-    case 0: return TAGGED(0); /* Choose nearest */
-    case 1: return TAGGED(3); /* Round towards zero */
-    case 2: return TAGGED(2); /* Towards positive infinity */
-    case 3: return TAGGED(1); /* Towards negative infinity */
+    case 0: return POLY_ROUND_TONEAREST; /* Choose nearest */
+    case 1: return POLY_ROUND_TOZERO; /* Round towards zero */
+    case 2: return POLY_ROUND_UPWARD; /* Towards positive infinity */
+    case 3: return POLY_ROUND_DOWNWARD; /* Towards negative infinity */
     }
-    return TAGGED(0); /* Never reached but this avoids warning message. */
+    return 0; /* Never reached but this avoids warning message. */
 }
 
 static void setrounding(TaskData *taskData, Handle args)
 {
     switch (get_C_long(taskData, DEREFWORDHANDLE(args)))
     {
-    case 0: __asm__("mtfsfi 7,0"); break; /* Choose nearest */
-    case 1: __asm__("mtfsfi 7,3"); break; /* Towards negative infinity */
-    case 2: __asm__("mtfsfi 7,2"); break; /* Towards positive infinity */
-    case 3: __asm__("mtfsfi 7,1"); break; /* Truncate towards zero */
+    case POLY_ROUND_TONEAREST: __asm__("mtfsfi 7,0"); break; /* Choose nearest */
+    case POLY_ROUND_DOWNWARD: __asm__("mtfsfi 7,3"); break; /* Towards negative infinity */
+    case POLY_ROUND_UPWARD: __asm__("mtfsfi 7,2"); break; /* Towards positive infinity */
+    case POLY_ROUND_TOZERO: __asm__("mtfsfi 7,1"); break; /* Truncate towards zero */
     }
 }
 #else
 // Give up.
-static PolyWord getrounding(TaskData *mdTaskData)
+static int getrounding(TaskData *mdTaskData)
 {
-    raise_exception_string(mdTaskData, EXC_Fail, "Unable to get flaoting point rounding control");
+    raise_exception_string(mdTaskData, EXC_Fail, "Unable to get floating point rounding control");
 }
 
 static void setrounding(TaskData *mdTaskData, Handle)
 {
-    raise_exception_string(mdTaskData, EXC_Fail, "Unable to set flaoting point rounding control");
+    raise_exception_string(mdTaskData, EXC_Fail, "Unable to set floating point rounding control");
 }
 #endif
 
@@ -574,14 +594,15 @@ Handle Real_strc(TaskData *mdTaskData, Handle hDigits, Handle hMode, Handle arg)
 {
     double  dx = real_arg(arg);
     int     decpt, sign;
-    int     mode = get_C_long(mdTaskData, DEREFWORDHANDLE(hMode));
-    int     digits = get_C_long(mdTaskData, DEREFWORDHANDLE(hDigits));
+    int     mode = get_C_int(mdTaskData, DEREFWORDHANDLE(hMode));
+    int     digits = get_C_int(mdTaskData, DEREFWORDHANDLE(hDigits));
     /* Compute the shortest string which gives the required value. */
-    /* N.B. dtoa uses static buffers and is NOT thread-safe. */
-    char *chars = dtoa(dx, mode, digits, &decpt, &sign, NULL);
+    /*  */
+    char *chars = poly_dtoa(dx, mode, digits, &decpt, &sign, NULL);
     /* We have to be careful in case an allocation causes a
        garbage collection. */
     PolyWord pStr = C_string_to_Poly(mdTaskData, chars);
+    poly_freedtoa(chars);
     Handle ppStr = mdTaskData->saveVec.push(pStr);
     /* Allocate a triple for the results. */
     PolyObject *result = alloc(mdTaskData, 3);
@@ -594,7 +615,7 @@ Handle Real_strc(TaskData *mdTaskData, Handle hDigits, Handle hMode, Handle arg)
 /* Functions added for Standard Basis Library are all indirected through here. */
 Handle Real_dispatchc(TaskData *mdTaskData, Handle args, Handle code)
 {
-    int c = get_C_long(mdTaskData, DEREFWORDHANDLE(code));
+    int c = get_C_int(mdTaskData, DEREFWORDHANDLE(code));
     switch (c)
     {
     case 0: /* tan */ return real_result(mdTaskData, tan(real_arg(args)));
@@ -631,7 +652,7 @@ Handle Real_dispatchc(TaskData *mdTaskData, Handle args, Handle code)
         setrounding(mdTaskData, args);
         return mdTaskData->saveVec.push(TAGGED(0)); /* Unit */
     case 10: /* getroundingmode */
-        return mdTaskData->saveVec.push(getrounding(mdTaskData));
+        return mdTaskData->saveVec.push(TAGGED(getrounding(mdTaskData)));
     /* Floating point representation queries. */
 #ifdef _DBL_RADIX
     case 11: /* Value of radix */ return mdTaskData->saveVec.push(TAGGED(_DBL_RADIX));
@@ -645,9 +666,9 @@ Handle Real_dispatchc(TaskData *mdTaskData, Handle args, Handle code)
        number which can be represented is DBL_MIN*2**(-DBL_MANT_DIG) */
     case 14: /* Minimum normalised number. */
         return real_result(mdTaskData, DBL_MIN);
-    case 15: /* Is finite */
+    case 15: /* Is finite */ /* No longer used - implemented in ML. */
         return mdTaskData->saveVec.push(finite(real_arg(args)) ? TAGGED(1) : TAGGED(0));
-    case 16: /* Is Nan */
+    case 16: /* Is Nan */ /* No longer used - implemented in ML. */
         return mdTaskData->saveVec.push(isnan(real_arg(args)) ? TAGGED(1) : TAGGED(0));
     case 17: /* Get sign bit.  There may be better ways to find this. */
         return mdTaskData->saveVec.push(copysign(1.0, real_arg(args)) < 0.0 ? TAGGED(1) : TAGGED(0));
@@ -675,7 +696,7 @@ Handle Real_dispatchc(TaskData *mdTaskData, Handle args, Handle code)
         }
     case 23: /* Compute ldexp */
         {
-            int exp = get_C_long(mdTaskData, DEREFHANDLE(args)->Get(1));
+            int exp = get_C_int(mdTaskData, DEREFHANDLE(args)->Get(1));
             return real_result(mdTaskData, ldexp(real_arg1(args), exp));
         }
     case 24: /* Get mantissa. */
@@ -746,14 +767,36 @@ void RealArithmetic::Init(void)
 {
     /* Some compilers object to overflow in constants so
        we compute the values here. */
-    double zero = 0.0;
-#if(defined(HAVE_IEEEFP_H) && ! defined(__CYGWIN__))
+#if (HAVE_DECL_FPSETMASK && ! defined(__CYGWIN__))
     /* In FreeBSD 3.4 at least, we sometimes get floating point
        exceptions if we don't clear the mask.  Maybe need to do
        this on other platforms as well just to be sure. */
+    // N.B.  fpsetmask is defined in the headers on Cygwin but there's no function!
     fpsetmask(0);
 #endif
-    posInf = 1.0 / zero;
-    negInf = -1.0 / zero;
-    notANumber = zero / zero;
+    // NAN and INFINITY are defined in GCC but not in Visual C++.
+#if (defined(INFINITY))
+    posInf = INFINITY;
+    negInf = -(INFINITY);
+#else
+    {
+        double zero = 0.0;
+        posInf = 1.0 / zero;
+        negInf = -1.0 / zero;
+    }
+#endif
+#if (defined(NAN))
+    notANumber = NAN;
+#else
+    {
+        double zero = 0.0;
+        notANumber = zero / zero;
+    }
+#endif
+    // Make sure this is a positive NaN since we return it from "abs".
+    // "Positive" in this context is copysign(1.0, x) > 0.0 because that's
+    // how we test the sign so we test it first and then try to change the
+    // sign if it's wrong.
+    if (copysign(1.0, notANumber) < 0)
+        notANumber = copysign(notANumber, 1.0);
 }
