@@ -2,7 +2,7 @@
     Title:      Run-time system.
     Author:     Dave Matthews, Cambridge University Computer Laboratory
 
-    Copyright (c) 2000
+    Copyright (c) 2000, 2009
         Cambridge University Technical Services Limited
 
     This library is free software; you can redistribute it and/or
@@ -21,13 +21,12 @@
 
 */
 
-/* Contains most of the routines in the interface_map vector. Others are
-   in their own modules e.g. arb.c, reals.c and persistence.c */
-
-#ifdef WIN32
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#elif defined(_WIN32)
 #include "winconfig.h"
 #else
-#include "config.h"
+#error "No configuration file"
 #endif
 
 /************************************************************************
@@ -140,6 +139,7 @@
 #include "save_vec.h"
 #include "rts_module.h"
 #include "memmgr.h"
+#include "statistics.h"
 
 #define SAVE(x) taskData->saveVec.push(x)
 #define SIZEOF(x) (sizeof(x)/sizeof(PolyWord))
@@ -166,7 +166,7 @@ PolyObject *alloc(TaskData *taskData, POLYUNSIGNED data_words, unsigned flags)
     
     if (profileMode == kProfileStoreAllocation)
     {
-        StackObject *stack = taskData->stack;
+        StackObject *stack = taskData->stack->stack();
         add_count(taskData, stack->p_pc, stack->p_sp, words);
     }
 
@@ -213,6 +213,46 @@ Handle full_gc_c(TaskData *taskData)
     return SAVE(TAGGED(0));
 }
 
+
+/******************************************************************************/
+/*                                                                            */
+/*      Error Messages                                                        */
+/*                                                                            */
+/******************************************************************************/
+
+
+// Return the handle to a string error message.  This will return
+// something like "Unknown error" from strerror if it doesn't match
+// anything.
+Handle errorMsg(TaskData *taskData, int err)
+{
+#ifdef _WIN32
+    /* In the Windows version we may have both errno values
+       and also GetLastError values.  We convert the latter into
+       negative values before returning them. */
+    if (err < 0)
+    {
+        LPTSTR lpMsg = NULL;
+        TCHAR *p;
+        if (FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM |
+                FORMAT_MESSAGE_ALLOCATE_BUFFER |
+                FORMAT_MESSAGE_IGNORE_INSERTS,
+                NULL, (DWORD)(-err), 0, (LPTSTR)&lpMsg, 1, NULL) > 0)
+        {
+            /* The message is returned with CRLF at the end.  Remove them. */
+            for (p = lpMsg; *p != '\0' && *p != '\n' && *p != '\r'; p++);
+            *p = '\0';
+            Handle res = SAVE(C_string_to_Poly(taskData, lpMsg));
+            LocalFree(lpMsg);
+            return res;
+        }
+    }
+#endif
+    // Unix and unknown Windows errors.
+    return SAVE(C_string_to_Poly(taskData, strerror(err)));
+}
+
+
 /******************************************************************************/
 /*                                                                            */
 /*      EXCEPTIONS                                                            */
@@ -246,7 +286,8 @@ Handle make_exn(TaskData *taskData, int id, Handle arg)
     DEREFEXNHANDLE(exnHandle)->ex_id   = TAGGED(id);
     DEREFEXNHANDLE(exnHandle)->ex_name = DEREFWORD(pushed_name);
     DEREFEXNHANDLE(exnHandle)->arg     = DEREFWORDHANDLE(arg);
-    
+    DEREFEXNHANDLE(exnHandle)->ex_location = TAGGED(0);
+
     return exnHandle;
 }
 
@@ -292,38 +333,41 @@ void raise_exception_string(TaskData *taskData, int id, const char *str)
     /*NOTREACHED*/
 }
 
-/******************************************************************************/
-/*                                                                            */
-/*      create_syscall_exception - called by run-time system                  */
-/*                                                                            */
-/******************************************************************************/
-// Create a syscall exception packet.
-Handle create_syscall_exception(TaskData *taskData, const char *errmsg, int err)
+// Raise a SysErr exception with a given error code.
+// The string part must match the result of OS.errorMsg
+void raiseSyscallError(TaskData *taskData, int err)
 {
-    /* exception SysErr of (string * syserror Option.option) */
-    /* If the argument is zero we don't have a suitable error number
-       so map this to NONE.  Other values are mapped to SOME(n) */
-    /* Create a pair of the error message and the error number. */
-    Handle pushed_option, pushed_name, pair;
-    if (err == 0) pushed_option = SAVE(NONE_VALUE); /* NONE */
-    else
-    {   /* SOME err */
-        Handle errornum = Make_arbitrary_precision(taskData, err);
-        pushed_option = alloc_and_save(taskData, 1);
-        DEREFHANDLE(pushed_option)->Set(0, DEREFWORDHANDLE(errornum));
-    }
-    pushed_name = SAVE(C_string_to_Poly(taskData, errmsg));
-    pair = alloc_and_save(taskData, 2);
+    Handle errornum = Make_arbitrary_precision(taskData, err);
+    Handle pushed_option = alloc_and_save(taskData, 1);
+    DEREFHANDLE(pushed_option)->Set(0, DEREFWORDHANDLE(errornum)); /* SOME err */
+    Handle pushed_name = errorMsg(taskData, err); // Generate the string.
+    Handle pair = alloc_and_save(taskData, 2);
     DEREFHANDLE(pair)->Set(0, DEREFWORDHANDLE(pushed_name));
     DEREFHANDLE(pair)->Set(1, DEREFWORDHANDLE(pushed_option));
-    return pair;
+
+    raise_exception(taskData, EXC_syserr, pair);
 }
 
+// Raise a SysErr exception which does not correspond to an error code.
+void raiseSyscallMessage(TaskData *taskData, const char *errmsg)
+{
+    Handle pushed_option = SAVE(NONE_VALUE); /* NONE */
+    Handle pushed_name = SAVE(C_string_to_Poly(taskData, errmsg));
+    Handle pair = alloc_and_save(taskData, 2);
+    DEREFHANDLE(pair)->Set(0, DEREFWORDHANDLE(pushed_name));
+    DEREFHANDLE(pair)->Set(1, DEREFWORDHANDLE(pushed_option));
 
-/* Raises a Syserr exception. */
+    raise_exception(taskData, EXC_syserr, pair);
+}
+
+// This was the previous version.  The errmsg argument is ignored unless err is zero.
+// Calls to it should really be replaced with calls to either raiseSyscallMessage
+// or raiseSyscallError but it's been left because there may be cases where errno
+// actually contains zero.
 void raise_syscall(TaskData *taskData, const char *errmsg, int err)
 {
-    raise_exception(taskData, EXC_syserr, create_syscall_exception(taskData, errmsg, err));
+    if (err == 0) raiseSyscallMessage(taskData, errmsg);
+    else raiseSyscallError(taskData, err);
 }
 
 // Raises a Fail exception.
@@ -358,9 +402,9 @@ void give_stack_trace(TaskData *taskData, PolyWord *sp, PolyWord *finish)
     // handlers which would otherwise look like return addresses.
     // Since we don't pass that in we may find it is actually out of
     // date if we are producing a trace as a result of pressing ^C.
-    StackObject *stack = taskData->stack;
+    StackObject *stack = taskData->stack->stack();
     PolyWord *exceptions = stack->p_hr;
-    PolyWord *endStack = stack->Offset(stack->Length());
+    PolyWord *endStack = taskData->stack->top;
     
 #ifdef DEBUG    
     printf("starting trace: sp = %p, finish = %p, end_of_stack = %p\n",
@@ -391,14 +435,15 @@ void give_stack_trace(TaskData *taskData, PolyWord *sp, PolyWord *finish)
         {
            /* A code pointer will be either a return address or a pointer
               to the constant section. (Or an exception handler?) */
-            PolyWord *ptr;
             // We used to have a check that this was not a constant area
             // pointer but we don't have those any more.
             
             /* Initialise ptr to points at the end-of-code marker */
-            OBJ_CODEPTR_TO_CONSTS_PTR(pc, ptr);
-            
-            PolyWord p_name = ptr[3]; /* Get procedure name */
+            // This used to use the OBJ_CODEPTR_TO_CONSTS_PTR macro which
+            // simply looked for the end-of-code marker.
+            PolyObject *ptr = ObjCodePtrToPtr(pc.AsCodePtr());
+            PolyWord *consts = ptr->ConstPtrForCode();
+            PolyWord p_name = consts[0]; /* Get procedure name */
             
             /* The name may be zero if it is anonymous */
             if (p_name == TAGGED(0)) fputs("<anon>\n",stdout);
@@ -420,9 +465,8 @@ void give_stack_trace(TaskData *taskData, PolyWord *sp, PolyWord *finish)
 /* CALL_IO0(stack_trace_, NOIND) */
 static Handle stack_trace_c(TaskData *taskData)
 {
-    StackObject *stack = taskData->stack;
-    PolyWord *endStack = stack->Offset(stack->Length());
-    give_stack_trace (taskData, stack->p_sp, endStack);
+    StackObject *stack = taskData->stack->stack();
+    give_stack_trace (taskData, stack->p_sp, taskData->stack->top);
     return SAVE(TAGGED(0));
 }
 
@@ -437,18 +481,33 @@ Handle ex_tracec(TaskData *taskData, Handle exnHandle, Handle handler_handle)
     PolyWord *handler = DEREFWORD(handler_handle).AsStackAddr();
     
     fputs("\nException trace for exception - ", stdout);
-    print_string(((poly_exn *)DEREFHANDLE(exnHandle))->ex_name);
+    print_string((DEREFEXNHANDLE(exnHandle))->ex_name);
+    // For backwards compatibility check the packet length first
+    if (DEREFHANDLE(exnHandle)->Length() == SIZEOF(poly_exn)) {
+        if (DEREFEXNHANDLE(exnHandle)->ex_location.IsDataPtr()) {
+            PolyObject *location = DEREFEXNHANDLE(exnHandle)->ex_location.AsObjPtr();
+            PolyWord fileName = location->Get(0);
+            POLYSIGNED lineNo = get_C_long(taskData, location->Get(1));
+            if (fileName.IsTagged() || ((PolyStringObject *)fileName.AsObjPtr())->length != 0) {
+                printf(" raised in ");
+                print_string(fileName);
+                if (lineNo != 0) printf(" line %" POLYSFMT, lineNo);
+            }
+            else if (lineNo != 0) printf(" raised at line %" POLYSFMT, lineNo);
+            fputs("\n", stdout);
+        }
+    }
     putc('\n',stdout);
     
     /* Trace down as far as the dummy handler on the stack. */
-    StackObject *stack = taskData->stack;
+    StackObject *stack = taskData->stack->stack();
     give_stack_trace(taskData, stack->p_sp, handler);
     fputs("End of trace\n\n",stdout);
     fflush(stdout);
     
     /* Set up the next handler so we don't come back here when we raise
        the exception again. */
-    taskData->stack->p_hr = (PolyWord*)(handler->AsStackAddr());
+    taskData->stack->stack()->p_hr = (PolyWord*)(handler->AsStackAddr());
     
     /* Set the exception data back again. */
     machineDependent->SetException(taskData, (poly_exn *)DEREFHANDLE(exnHandle));
@@ -462,58 +521,12 @@ Handle ex_tracec(TaskData *taskData, Handle exnHandle, Handle handler_handle)
 // Return the address of the iovec entry for a given index.
 Handle io_operation_c(TaskData *taskData, Handle entry)
 {
-    POLYUNSIGNED entryNo = get_C_ulong(taskData, DEREFWORD(entry));
+    unsigned entryNo = get_C_unsigned(taskData, DEREFWORD(entry));
     if (entryNo >= POLY_SYS_vecsize)
         raise_exception0(taskData, EXC_subscript);
     return SAVE((PolyObject*)IoEntry(entryNo));
 }
 
-
-/******************************************************************************/
-/*                                                                            */
-/*      INITIALISATION                                                         */
-/*                                                                            */
-/******************************************************************************/
-
-/******************************************************************************/
-/*                                                                            */
-/*      re_init_run_time_system - called from mpoly.c                         */
-/*                                                                            */
-/******************************************************************************/
-void re_init_run_time_system(void)
-{
-    ReinitModules();
-}
-
-/******************************************************************************/
-/*                                                                            */
-/*      init_run_time_system - called from mpoly.c                            */
-/*                                                                            */
-/******************************************************************************/
-void init_run_time_system(void)
-{
-    InitModules(); // Initialise other modules.
-}
-
-/******************************************************************************/
-/*                                                                            */
-/*      uninit_run_time_system - called from ???                            */
-/*                                                                            */
-/******************************************************************************/
-
-/* Release all resources.
-   This is really only needed with Windows when running as a DLL within
-   the address space of another process. */
-void uninit_run_time_system(void)
-{
-    UninitModules();
-
-#if (0)
-    uninit_xwindow_system(); /* Probably needs to be done. */
-#endif
-}
-
-
 /******************************************************************************/
 /*                                                                            */
 /*      get_flags_c - called from machine_assembly.s                          */
@@ -605,12 +618,12 @@ static Handle assign_word_long_c(TaskData *taskData, Handle value_handle, Handle
 static Handle move_bytes_long_c(TaskData *taskData, Handle len, Handle dest_offset_handle, Handle dest_handle,
                        Handle src_offset_handle, Handle src_handle)
 {
-    unsigned src_offset = get_C_ulong(taskData, DEREFWORDHANDLE(src_offset_handle));
+    POLYUNSIGNED src_offset = get_C_ulong(taskData, DEREFWORDHANDLE(src_offset_handle));
     byte *source = DEREFBYTEHANDLE(src_handle) + src_offset;
-    unsigned dest_offset = get_C_ulong(taskData, DEREFWORDHANDLE(dest_offset_handle));
+    POLYUNSIGNED dest_offset = get_C_ulong(taskData, DEREFWORDHANDLE(dest_offset_handle));
     byte *destination = DEREFBYTEHANDLE(dest_handle);
     byte *dest = destination + dest_offset;
-    unsigned bytes = get_C_ulong(taskData, DEREFWORDHANDLE(len));
+    POLYUNSIGNED bytes = get_C_ulong(taskData, DEREFWORDHANDLE(len));
     PolyObject *obj = DEREFHANDLE(dest_handle);
     ASSERT(obj->IsByteObject());
 
@@ -641,6 +654,22 @@ static Handle move_words_long_c(TaskData *taskData, Handle len, Handle dest_offs
     return taskData->saveVec.push(TAGGED(0));
 }
 
+static Handle testBytesEqual(TaskData *taskData, Handle len, Handle yOffset, Handle y,
+                             Handle xOffset, Handle x)
+{
+    POLYUNSIGNED x_offset = get_C_ulong(taskData, DEREFWORDHANDLE(xOffset));
+    byte *xAddr = DEREFBYTEHANDLE(x) + x_offset;
+
+    POLYUNSIGNED y_offset = get_C_ulong(taskData, DEREFWORDHANDLE(yOffset));
+    byte *yAddr = DEREFBYTEHANDLE(y) + y_offset;
+
+    POLYUNSIGNED bytes = get_C_ulong(taskData, DEREFWORDHANDLE(len));
+
+    int res = memcmp(xAddr, yAddr, bytes);
+    if (res == 0) return taskData->saveVec.push(TAGGED(1));
+    else return taskData->saveVec.push(TAGGED(0));
+}
+
 static Handle vec_length_c(TaskData *taskData, Handle vector)    /* Length of a vector */
 {
     POLYUNSIGNED length = vector->WordP()->Length();
@@ -665,11 +694,11 @@ static Handle load_word_long_c(TaskData *taskData, Handle word_no /* offset in W
 // return address into the assembly code.
 static Handle alloc_store_long_c(TaskData *taskData, Handle initial, Handle flags_handle, Handle size )
 {
-    POLYUNSIGNED flags = get_C_ulong(taskData, DEREFWORD(flags_handle));
+    unsigned flags = get_C_unsigned(taskData, DEREFWORD(flags_handle));
     POLYUNSIGNED usize = get_C_ulong(taskData, DEREFWORD(size));
     
-    if (usize == 0) usize = 1;
-    if (usize >= MAX_OBJECT_SIZE) raise_exception0(taskData, EXC_size);
+    if (usize == 0 || usize >= MAX_OBJECT_SIZE)
+        raise_exception0(taskData, EXC_size);
     
     PolyObject *vector = alloc(taskData, usize, flags| F_MUTABLE_BIT);
     
@@ -688,6 +717,17 @@ static Handle alloc_store_long_c(TaskData *taskData, Handle initial, Handle flag
     return taskData->saveVec.push(vector);
 }
 
+// Alloc_uninit returns the address of some memory that will be filled in later
+// so need not be initialised unlike alloc_store where the initial value may be
+// significant.  For word objects we can't risk leaving them uninitialised in case
+// we GC before we've finished filling them.  There's no harm in initialising byte
+// objects.
+static Handle alloc_uninit_c(TaskData *taskData, Handle flags_handle, Handle size )
+{
+    Handle initial = SAVE(TAGGED(0));
+    return alloc_store_long_c(taskData, initial, flags_handle, size);
+}
+
 /* Word functions. These functions assume that their arguments are tagged
    integers and treat them as unsigned values.
    These functions will almost always be implemented directly in the code
@@ -855,37 +895,36 @@ void CheckAndGrowStack(TaskData *taskData, PolyWord *lower_limit)
    a procedure to ensure that there is enough space for the maximum that can
    be allocated. */
 {
+    StackObject *oldStack = taskData->stack->stack();
     /* Get current size of new stack segment. */
-    POLYUNSIGNED old_len = OBJECT_LENGTH(taskData->stack);
+    POLYUNSIGNED old_len = taskData->stack->spaceSize();
  
     /* The minimum size must include the reserved space for the registers. */
-    POLYUNSIGNED min_size = ((PolyWord*)taskData->stack) + old_len - lower_limit + taskData->stack->p_space;
+    POLYUNSIGNED min_size = ((PolyWord*)oldStack) + old_len - lower_limit + oldStack->p_space;
     
     if (old_len >= min_size) return; /* Ok with present size. */
 
     // If it is too small double its size.
-    // BUT, the maximum size is 2^24-1 words (on 32 bit) or 2^56-1 on 64 bit.
 
-    if (old_len == MAX_OBJECT_SIZE)
+    POLYUNSIGNED new_len; /* New size */
+    for (new_len = old_len; new_len < min_size; new_len *= 2);
+
+    if (! gMem.GrowOrShrinkStack(taskData->stack, new_len))
     {
         /* Cannot expand the stack any further. */
-        fprintf(stderr, "Warning - Stack limit reached - interrupting process\n");
+        fprintf(stderr, "Warning - Unable to increase stack - interrupting thread\n");
+        if (debugOptions & DEBUG_THREADS)
+            Log("THREAD: Unable to grow stack for thread %p from %lu to %lu\n", taskData, old_len, new_len);
         // We really should do this only if the thread is handling interrupts
         // asynchronously.  On the other hand what else do we do?
         Handle exn = make_exn(taskData, EXC_interrupt, SAVE(TAGGED(0)));
         machineDependent->SetException(taskData, DEREFEXNHANDLE(exn));
-        return;
     }
-
-    POLYUNSIGNED new_len; /* New size */
-    for (new_len = old_len; new_len < min_size; new_len *= 2);
-    if (new_len > MAX_OBJECT_SIZE) new_len = MAX_OBJECT_SIZE;
-
-    /* Must make a new frame and copy the data over. */
-    StackObject *new_stack = // N.B.  May throw a C++ exception.
-        (StackObject *)alloc(taskData, new_len, F_MUTABLE_BIT|F_STACK_OBJ);
-    CopyStackFrame(taskData->stack, new_stack);
-    taskData->stack = new_stack;
+    else
+    {
+        if (debugOptions & DEBUG_THREADS)
+            Log("THREAD: Growing stack for thread %p from %lu to %lu\n", taskData, old_len, new_len);
+    }
 }
 
 // This is used after executing each top-level command to minimise the
@@ -893,35 +932,28 @@ void CheckAndGrowStack(TaskData *taskData, PolyWord *lower_limit)
 static Handle shrink_stack_c(TaskData *taskData, Handle reserved_space)
 /* Shrinks the current stack. */
 {
-    int reserved = get_C_long(taskData, DEREFWORDHANDLE(reserved_space));
-
-    int old_len; /* Current size of stack segment. */
-    int new_len; /* New size */
-    int min_size;
-    StackObject *new_stack;
-
-    if (reserved < 0)
-    {
-       raise_exception0(taskData, EXC_size);
-    }
+    unsigned reserved = get_C_unsigned(taskData, DEREFWORDHANDLE(reserved_space));
 
+    StackObject *oldStack = taskData->stack->stack();
     /* Get current size of new stack segment. */
-    old_len = OBJECT_LENGTH(taskData->stack);
+    POLYUNSIGNED old_len = taskData->stack->spaceSize();
  
     /* The minimum size must include the reserved space for the registers. */
-    min_size = (((PolyWord*)taskData->stack) + old_len - (PolyWord*)taskData->stack->p_sp) + taskData->stack->p_space + reserved;
-    
+    POLYUNSIGNED min_size = (((PolyWord*)oldStack) + old_len - (PolyWord*)oldStack->p_sp) + oldStack->p_space + reserved;
+
+    POLYUNSIGNED new_len; /* New size */
     for (new_len = machineDependent->InitialStackSize(); new_len < min_size; new_len *= 2);
 
     if (old_len <= new_len) return SAVE(TAGGED(0)); /* OK with present size. */
 
-    /* Must make a new frame and copy the data over. */
-    new_stack = (StackObject *)alloc(taskData, new_len, F_MUTABLE_BIT|F_STACK_OBJ);
-    CopyStackFrame(taskData->stack, new_stack);
-    taskData->stack = new_stack;    
+    // Try to change the stack size but ignore any error since the current size will do.
+    gMem.GrowOrShrinkStack(taskData->stack, new_len);
+
     return SAVE(TAGGED(0));
 }
 
+static unsigned long rtsCallCounts[POLY_SYS_vecsize];
+
 Handle EnterPolyCode(TaskData *taskData)
 /* Called from "main" to enter the code. */
 {
@@ -929,9 +961,16 @@ Handle EnterPolyCode(TaskData *taskData)
     while (1)
     {
         taskData->saveVec.reset(hOriginal); // Remove old RTS arguments and results.
+
         // Run the ML code and return with the function to call.
+        taskData->inML = true;
         int ioFunction = machineDependent->SwitchToPoly(taskData);
-	    try {
+        taskData->inML = false;
+
+        if ((debugOptions & DEBUG_RTSCALLS) && ioFunction >= 0 && ioFunction < POLY_SYS_vecsize)
+            rtsCallCounts[ioFunction]++;
+
+        try {
             switch (ioFunction)
             {
             case -1:
@@ -960,6 +999,10 @@ Handle EnterPolyCode(TaskData *taskData)
                 machineDependent->CallIO3(taskData, &alloc_store_long_c);
                 break;
 
+            case POLY_SYS_alloc_uninit:
+                machineDependent->CallIO2(taskData, &alloc_uninit_c);
+                break;
+
             case POLY_SYS_chdir:
                 machineDependent->CallIO1(taskData, &change_dirc);
                 break;
@@ -1010,6 +1053,10 @@ Handle EnterPolyCode(TaskData *taskData)
                 machineDependent->CallIO1(taskData, &profilerc);
                 break;
 
+            case POLY_SYS_quotrem:
+                machineDependent->CallIO3(taskData, &quot_rem_c);
+                break;
+
     //        case POLY_SYS_is_short: machineDependent->CallIO1(taskData, &is_shortc); break;
 
             case POLY_SYS_aplus:
@@ -1100,6 +1147,10 @@ Handle EnterPolyCode(TaskData *taskData)
                 machineDependent->CallIO2(taskData, &Real_divc);
                 break;
 
+            case POLY_SYS_Abs_real:
+                machineDependent->CallIO1(taskData, &Real_absc);
+                break;
+
             case POLY_SYS_Neg_real:
                 machineDependent->CallIO1(taskData, &Real_negc);
                 break;
@@ -1309,6 +1360,10 @@ Handle EnterPolyCode(TaskData *taskData)
                 machineDependent->CallIO2(taskData, &poly_dispatch_c);
                 break;
 
+            case POLY_SYS_bytevec_eq:
+                machineDependent->CallIO5(taskData, &testBytesEqual);
+                break;
+
             case POLY_SYS_set_code_constant:
                 machineDependent->CallIO4(taskData, &set_code_constant);
                 break;
@@ -1386,9 +1441,290 @@ Handle EnterPolyCode(TaskData *taskData)
             default:
                 Crash("Unknown io operation %d\n", ioFunction);
             }
-		}
-		catch (IOException) {
+        }
+        catch (IOException) {
         }
 
     }
 }
+
+// This RTS module is defined purely to allow us to print the statistics.
+class RTS: public RtsModule
+{
+public:
+    virtual void Stop(void);
+};
+
+static RTS rtsModule;
+
+static const char * const rtsName[POLY_SYS_vecsize] =
+{
+    "RTS Call   0",
+    "SYS_exit",
+    "RTS Call   2",
+    "RTS Call   3",
+    "RTS Call   4",
+    "RTS Call   5",
+    "RTS Call   6",
+    "RTS Call   7",
+    "RTS Call   8",
+    "SYS_chdir",
+    "RTS Call  10",
+    "SYS_alloc_store",
+    "SYS_alloc_unit",
+    "RTS Call  13",
+    "SYS_raisex",
+    "SYS_get_length",
+    "RTS Call  16",
+    "SYS_get_flags",
+    "RTS Call  18",
+    "RTS Call  19",
+    "RTS Call  20",
+    "RTS Call  21",
+    "RTS Call  22",
+    "SYS_str_compare",
+    "SYS_teststreq",
+    "SYS_teststrneq",
+    "SYS_teststrgtr",
+    "SYS_teststrlss",
+    "SYS_teststrgeq",
+    "SYS_teststrleq",
+    "SYS_exception_trace",
+    "SYS_give_ex_trace",
+    "RTS Call  32",
+    "RTS Call  33",
+    "RTS Call  34",
+    "RTS Call  35",
+    "RTS Call  36",
+    "RTS Call  37",
+    "RTS Call  38",
+    "RTS Call  39",
+    "RTS Call  40",
+    "RTS Call  41",
+    "RTS Call  42",
+    "RTS Call  43",
+    "RTS Call  44",
+    "RTS Call  45",
+    "RTS Call  46",
+    "SYS_lockseg",
+    "SYS_emptystring", // Not actually a call
+    "SYS_nullvector",  // Not actually a call
+    "RTS Call  50",
+    "SYS_network",
+    "SYS_os_specific",
+    "RTS Call  53",
+    "RTS Call  54",
+    "RTS Call  55",
+    "RTS Call  56",
+    "RTS Call  57",
+    "RTS Call  58",
+    "RTS Call  59",
+    "RTS Call  60",
+    "SYS_io_dispatch",
+    "SYS_signal_handler",
+    "RTS Call  63",
+    "RTS Call  64",
+    "RTS Call  65",
+    "RTS Call  66",
+    "RTS Call  67",
+    "RTS Call  68",
+    "RTS Call  69",
+    "SYS_atomic_incr",
+    "SYS_atomic_decr",
+    "SYS_thread_self",
+    "SYS_thread_dispatch",
+    "RTS Call  74",
+    "RTS Call  75",
+    "RTS Call  76",
+    "RTS Call  77",
+    "RTS Call  78",
+    "RTS Call  79",
+    "RTS Call  70",
+    "RTS Call  81",
+    "RTS Call  82",
+    "RTS Call  83",
+    "SYS_kill_self",
+    "RTS Call  85",
+    "RTS Call  86",
+    "RTS Call  87",
+    "SYS_profiler",
+    "RTS Call  89",
+    "RTS Call  90",
+    "RTS Call  91",
+    "SYS_full_gc",
+    "SYS_stack_trace",
+    "SYS_timing_dispatch",
+    "RTS Call  95",
+    "RTS Call  96",
+    "RTS Call  97",
+    "RTS Call  98",
+    "SYS_objsize",
+    "SYS_showsize",
+    "RTS Call 101",
+    "RTS Call 102",
+    "RTS Call 103",
+    "SYS_quotrem",
+    "SYS_is_short",
+    "SYS_aplus",
+    "SYS_aminus",
+    "SYS_amul",
+    "SYS_adiv",
+    "SYS_amod",
+    "SYS_aneg",
+    "SYS_xora",
+    "SYS_equala",
+    "SYS_ora",
+    "SYS_anda",
+    "RTS Call 116",
+    "SYS_Real_str",
+    "SYS_Real_geq",
+    "SYS_Real_leq",
+    "SYS_Real_gtr",
+    "SYS_Real_lss",
+    "SYS_Real_eq",
+    "SYS_Real_neq",
+    "SYS_Real_Dispatch",
+    "SYS_Add_real",
+    "SYS_Sub_real",
+    "SYS_Mul_real",
+    "SYS_Div_real",
+    "SYS_Abs_real",
+    "SYS_Neg_real",
+    "RTS Call 131",
+    "SYS_Repr_real",
+    "SYS_conv_real",
+    "SYS_real_to_int",
+    "SYS_int_to_real",
+    "SYS_sqrt_real",
+    "SYS_sin_real",
+    "SYS_cos_real",
+    "SYS_arctan_real",
+    "SYS_exp_real",
+    "SYS_ln_real",
+    "RTS Call 142",
+    "RTS Call 143",
+    "RTS Call 144",
+    "RTS Call 145",
+    "RTS Call 146",
+    "RTS Call 147",
+    "SYS_stdin",    // Not actually a call
+    "SYS_stdout",   // Not actually a call
+    "SYS_process_env",
+    "SYS_set_string_length",
+    "SYS_get_first_long_word",
+    "SYS_poly_specific",
+    "SYS_bytevec_eq",
+    "RTS Call 155",
+    "RTS Call 156",
+    "RTS Call 157",
+    "RTS Call 158",
+    "RTS Call 159",
+    "RTS Call 160",
+    "RTS Call 161",
+    "RTS Call 162",
+    "RTS Call 163",
+    "RTS Call 164",
+    "RTS Call 165",
+    "RTS Call 166",
+    "RTS Call 167",
+    "RTS Call 168",
+    "RTS Call 169",
+    "RTS Call 170",
+    "RTS Call 171",
+    "RTS Call 172",
+    "RTS Call 173",
+    "RTS Call 174",
+    "RTS Call 175",
+    "RTS Call 176",
+    "RTS Call 177",
+    "RTS Call 178",
+    "RTS Call 179",
+    "RTS Call 170",
+    "RTS Call 181",
+    "RTS Call 182",
+    "RTS Call 183",
+    "RTS Call 184",
+    "RTS Call 185",
+    "RTS Call 186",
+    "RTS Call 187",
+    "RTS Call 188",
+    "SYS_io_operation",
+    "RTS Call 190",
+    "RTS Call 191",
+    "RTS Call 192",
+    "RTS Call 193",
+    "SYS_set_code_constant",
+    "SYS_move_words",
+    "SYS_shift_right_arith_word",
+    "SYS_int_to_word",
+    "SYS_move_bytes",
+    "RTS Call 199",
+    "SYS_code_flags",
+    "SYS_shrink_stack",
+    "SYS_stderr",
+    "RTS Call 203",
+    "SYS_callcode_tupled",
+    "SYS_foreign_dispatch",
+    "SYS_foreign_null",
+    "RTS Call 207",
+    "RTS Call 208",
+    "SYS_XWindows",
+    "RTS Call 210",
+    "RTS Call 211",
+    "RTS Call 212",
+    "SYS_is_big_endian",
+    "SYS_bytes_per_word",
+    "SYS_offset_address",
+    "SYS_shift_right_word",
+    "SYS_word_neq",
+    "SYS_not_bool",
+    "RTS Call 219",
+    "RTS Call 220",
+    "RTS Call 221",
+    "RTS Call 222",
+    "SYS_string_length",
+    "RTS Call 224",
+    "RTS Call 225",
+    "RTS Call 226",
+    "RTS Call 227",
+    "RTS Call 228",
+    "SYS_int_eq",
+    "SYS_int_neq",
+    "SYS_int_geq",
+    "SYS_int_leq",
+    "SYS_int_gtr",
+    "SYS_int_lss",
+    "RTS Call 235",
+    "RTS Call 236",
+    "RTS Call 237",
+    "SYS_mul_word",
+    "SYS_plus_word",
+    "SYS_minus_word",
+    "SYS_div_word",
+    "SYS_or_word",
+    "SYS_and_word",
+    "SYS_xor_word",
+    "SYS_shift_left_word",
+    "SYS_mod_word",
+    "SYS_word_geq",
+    "SYS_word_leq",
+    "SYS_word_gtr",
+    "SYS_word_lss",
+    "SYS_word_eq",
+    "SYS_load_byte",
+    "SYS_load_word",
+    "SYS_assign_byte",
+    "SYS_assign_word"
+};
+
+void RTS::Stop()
+{
+    if (debugOptions & DEBUG_RTSCALLS)
+    {
+        for (unsigned i = 0; i < POLY_SYS_vecsize; i++)
+        {
+            if (rtsCallCounts[i] != 0)
+                Log("RTS: %s called %lu times\n", rtsName[i], rtsCallCounts[i]);
+        }
+    }
+}
