@@ -20,10 +20,12 @@
 */
 
 
-#ifdef WIN32
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#elif defined(_WIN32)
 #include "winconfig.h"
 #else
-#include "config.h"
+#error "No configuration file"
 #endif
 
 #ifdef HAVE_STDIO_H
@@ -50,6 +52,14 @@
 #include <sys/types.h>
 #endif
 
+#ifdef HAVE_SYS_STAT_H
+#include <sys/stat.h>
+#endif
+
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif
+
 #ifdef HAVE_ASSERT_H
 #include <assert.h>
 #define ASSERT(x)   assert(x)
@@ -65,23 +75,28 @@
 #include "scanaddrs.h"
 #include "arb.h"
 #include "memmgr.h"
-#include "polyexports.h"
 #include "mpoly.h" // For exportTimeStamp
 #include "exporter.h" // For CopyScan
 #include "machine_dep.h"
 #include "osmem.h"
+#include "gc.h" // For FullGC.
 
 #if(!defined(MAXPATHLEN) && defined(MAX_PATH))
 #define MAXPATHLEN MAX_PATH
 #endif
 
+#ifdef _MSC_VER
+// Don't tell me about ISO C++ changes.
+#pragma warning(disable:4996)
+#endif
+
 // Helper class to close files on exit.
 class AutoClose {
 public:
     AutoClose(FILE *f = 0): m_file(f) {}
     ~AutoClose() { if (m_file) ::fclose(m_file); }
 
-	operator FILE*() { return m_file; }
+    operator FILE*() { return m_file; }
 
 private:
     FILE *m_file;
@@ -127,10 +142,10 @@ typedef struct _savedStateHeader
     off_t       stringTable;            // Pointer to the string table (zero if none)
     size_t      stringTableSize;        // Size of string table
     unsigned    parentNameEntry;        // Position of parent name in string table (0 if top)
-    UNSIGNEDADDR timeStamp;            // The time stamp for this file.
-    UNSIGNEDADDR fileSignature;        // The signature for this file.
-    UNSIGNEDADDR parentTimeStamp;      // The time stamp for the parent.
-    UNSIGNEDADDR parentSignature;      // The signature for the parent.
+    time_t      timeStamp;            // The time stamp for this file.
+    uintptr_t   fileSignature;        // The signature for this file.
+    time_t      parentTimeStamp;      // The time stamp for the parent.
+    uintptr_t   parentSignature;      // The signature for the parent.
 } SavedStateHeader;
 
 // Entry for segment table.  This describes the segments on the disc that
@@ -150,6 +165,7 @@ typedef struct _savedStateSegmentDescr
 #define SSF_WRITABLE    1               // The segment contains mutable data
 #define SSF_OVERWRITE   2               // The segment overwrites the data (mutable) in a parent.
 #define SSF_NOOVERWRITE 4               // The segment must not be further overwritten
+#define SSF_BYTES       8               // The segment containing only byte data
 
 typedef struct _relocationEntry
 {
@@ -175,17 +191,17 @@ typedef struct _relocationEntry
 class HierarchyTable
 {
 public:
-    HierarchyTable(const char *file, UNSIGNEDADDR time):
+    HierarchyTable(const char *file, time_t time):
       fileName(strdup(file)), timeStamp(time) { }
     AutoFree<char*> fileName;
-    UNSIGNEDADDR timeStamp;
+    time_t          timeStamp;
 };
 
 HierarchyTable **hierarchyTable;
 
 static unsigned hierarchyDepth;
 
-static bool AddHierarchyEntry(const char *fileName, UNSIGNEDADDR timeStamp)
+static bool AddHierarchyEntry(const char *fileName, time_t timeStamp)
 {
     // Add an entry to the hierarchy table for this file.
     HierarchyTable *newEntry = new HierarchyTable(fileName, timeStamp);
@@ -198,6 +214,34 @@ static bool AddHierarchyEntry(const char *fileName, UNSIGNEDADDR timeStamp)
     return true;
 }
 
+// Test whether we're overwriting a parent of ourself.
+#ifdef HAVE_WINDOWS_H
+static bool sameFile(const char *x, const char *y)
+{
+    // Get the lengths and return if either does not exist.
+    LPSTR filePart;
+    DWORD dwxLen = GetFullPathName(x, 1, 0, 0);
+    if (dwxLen == 0) return false;
+    DWORD dwyLen = GetFullPathName(y, 0, 0, 0);
+    if (dwyLen == 0) return false;
+    if (dwxLen != dwyLen) return false;
+    AutoFree<char*> xName = (char*)malloc(dwxLen+1);
+    GetFullPathName(x, dwxLen+1, xName, &filePart);
+    AutoFree<char*> yName = (char*)malloc(dwyLen+1);
+    GetFullPathName(y, dwyLen+1, yName, &filePart);
+    return strcmpi(xName, yName) == 0;
+}
+#else
+static bool sameFile(const char *x, const char *y)
+{
+    struct stat xStat, yStat;
+    // If either file does not exist that's fine.
+    if (stat(x, &xStat) != 0 || stat(y, &yStat) != 0)
+        return false;
+    return (xStat.st_dev == yStat.st_dev && xStat.st_ino == yStat.st_ino);
+}
+#endif
+
 /*
  *  Saving state.
  */
@@ -243,7 +287,7 @@ PolyWord SaveStateExport::createRelocation(PolyWord p, void *relocAddr)
     void *addr = p.AsAddress();
     unsigned addrArea = findArea(addr);
     reloc.targetAddress = (char*)addr - (char*)memTable[addrArea].mtAddr;
-    reloc.targetSegment = memTable[addrArea].mtIndex;
+    reloc.targetSegment = (unsigned)memTable[addrArea].mtIndex;
     reloc.relKind = PROCESS_RELOC_DIRECT;
     fwrite(&reloc, sizeof(reloc), 1, exportFile);
     relocationCount++;
@@ -273,7 +317,7 @@ void SaveStateExport::ScanConstant(byte *addr, ScanRelocationKind code)
     RelocationEntry reloc;
     setRelocationAddress(addr, &reloc.relocAddress);
     reloc.targetAddress = (char*)a - (char*)memTable[aArea].mtAddr;
-    reloc.targetSegment = memTable[aArea].mtIndex;
+    reloc.targetSegment = (unsigned)memTable[aArea].mtIndex;
     reloc.relKind = code;
     fwrite(&reloc, sizeof(reloc), 1, exportFile);
     relocationCount++;
@@ -283,7 +327,8 @@ void SaveStateExport::ScanConstant(byte *addr, ScanRelocationKind code)
 class SaveRequest: public MainThreadRequest
 {
 public:
-    SaveRequest(const char *name, unsigned h): fileName(name), newHierarchy(h),
+    SaveRequest(const char *name, unsigned h): MainThreadRequest(MTP_SAVESTATE),
+        fileName(name), newHierarchy(h),
         errorMessage(0), errCode(0) {}
 
     virtual void Perform();
@@ -351,6 +396,16 @@ PolyWord SaveFixupAddress::GetNewAddress(PolyWord old)
 // Called by the root thread to actually save the state and write the file.
 void SaveRequest::Perform()
 {
+    // Check that we aren't overwriting our own parent.
+    for (unsigned q = 0; q < newHierarchy-1; q++) {
+        if (sameFile(hierarchyTable[q]->fileName, fileName))
+        {
+            errorMessage = "File being saved is used as a parent of this file";
+            errCode = 0;
+            return;
+        }
+    }
+
     SaveStateExport exports;
     // Open the file.  This could quite reasonably fail if the path is wrong.
     exports.exportFile = fopen(fileName, "wb");
@@ -364,12 +419,13 @@ void SaveRequest::Perform()
     // Scan over the permanent mutable area copying all reachable data that is
     // not in a lower hierarchy into new permanent segments.
     CopyScan copyScan(newHierarchy);
+    copyScan.initialise(false);
     bool success = true;
     try {
         for (unsigned i = 0; i < gMem.npSpaces; i++)
         {
             PermanentMemSpace *space = gMem.pSpaces[i];
-            if (space->isMutable && ! space->noOverwrite)
+            if (space->isMutable && ! space->noOverwrite && ! space->byteOnly)
                 copyScan.ScanAddressesInRegion(space->bottom, space->top);
         }
     }
@@ -406,6 +462,7 @@ void SaveRequest::Perform()
             {
                 entry->mtFlags = MTF_WRITEABLE;
                 if (space->noOverwrite) entry->mtFlags |= MTF_NO_OVERWRITE;
+                if (space->byteOnly) entry->mtFlags |= MTF_BYTES;
             }
             else
                 entry->mtFlags = MTF_EXECUTABLE;
@@ -425,6 +482,7 @@ void SaveRequest::Perform()
         {
             entry->mtFlags = MTF_WRITEABLE;
             if (space->noOverwrite) entry->mtFlags |= MTF_NO_OVERWRITE;
+            if (space->byteOnly) entry->mtFlags |= MTF_BYTES;
         }
         else
             entry->mtFlags = MTF_EXECUTABLE;
@@ -438,13 +496,17 @@ void SaveRequest::Perform()
     for (unsigned l = 0; l < gMem.nlSpaces; l++)
     {
         LocalMemSpace *space = gMem.lSpaces[l];
-        fixup.ScanAddressesInRegion(space->pointer, space->top);
+        fixup.ScanAddressesInRegion(space->bottom, space->lowerAllocPtr);
+        fixup.ScanAddressesInRegion(space->upperAllocPtr, space->top);
     }
     GCModules(&fixup);
 
     // Update the global memory space table.  Old segments at the same level
     // or lower are removed.  The new segments become permanent.
-    if (! success || ! gMem.PromoteExportSpaces(newHierarchy))
+    // Try to promote the spaces even if we've had a failure because export
+    // spaces are deleted in ~CopyScan and we may have already copied
+    // some objects there.
+    if (! gMem.PromoteExportSpaces(newHierarchy) || ! success)
     {
         errorMessage = "Out of Memory";
         errCode = ENOMEM;
@@ -487,7 +549,7 @@ void SaveRequest::Perform()
         memoryTableEntry *entry = &exports.memTable[j];
         memset(&descrs[j], 0, sizeof(SavedStateSegmentDescr));
         descrs[j].relocationSize = sizeof(RelocationEntry);
-        descrs[j].segmentIndex = entry->mtIndex;
+        descrs[j].segmentIndex = (unsigned)entry->mtIndex;
         descrs[j].segmentSize = entry->mtLength; // Set this even if we don't write it.
         descrs[j].originalAddress = entry->mtAddr;
         if (entry->mtFlags & MTF_WRITEABLE)
@@ -497,6 +559,8 @@ void SaveRequest::Perform()
                 descrs[j].segmentFlags |= SSF_NOOVERWRITE;
             if (j < permanentEntries && (entry->mtFlags & MTF_NO_OVERWRITE) == 0)
                 descrs[j].segmentFlags |= SSF_OVERWRITE;
+            if (entry->mtFlags & MTF_BYTES)
+                descrs[j].segmentFlags |= SSF_BYTES;
         }
     }
     // Write out temporarily. Will be overwritten at the end.
@@ -567,10 +631,16 @@ Handle SaveState(TaskData *taskData, Handle args)
     if (length > MAXPATHLEN)
         raise_syscall(taskData, "File name too long", ENAMETOOLONG);
     // The value of depth is zero for top-level save so we need to add one for hierarchy.
-    unsigned newHierarchy = get_C_ulong(taskData, DEREFHANDLE(args)->Get(1)) + 1;
-    // We don't support hierarchical saving at the moment.
+    unsigned newHierarchy = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(1)) + 1;
+
     if (newHierarchy > hierarchyDepth+1)
         raise_fail(taskData, "Depth must be no more than the current hierarchy plus one");
+
+    // Request a full GC first.  The main reason is to avoid running out of memory as a
+    // result of repeated saves.  Old export spaces are turned into local spaces and
+    // the GC will delete them if they are completely empty
+    FullGC(taskData);
+
     SaveRequest request(fileNameBuff, newHierarchy);
     processes->MakeRootRequest(taskData, &request);
     if (request.errorMessage)
@@ -585,10 +655,11 @@ Handle SaveState(TaskData *taskData, Handle args)
 class StateLoader: public MainThreadRequest
 {
 public:
-    StateLoader(const char *file): errorResult(0), errNumber(0) { strcpy(fileName, file); }
+    StateLoader(const char *file): MainThreadRequest(MTP_LOADSTATE),
+        errorResult(0), errNumber(0) { strcpy(fileName, file); }
 
     virtual void Perform(void);
-    bool LoadFile(void);
+    bool LoadFile(bool isInitial, time_t requiredStamp);
     const char *errorResult;
     // The fileName here is the last file loaded.  As well as using it
     // to load the name can also be printed out at the end to identify the
@@ -600,7 +671,7 @@ public:
 // Called by the main thread once all the ML threads have stopped.
 void StateLoader::Perform(void)
 {
-    (void)LoadFile();
+    (void)LoadFile(true, 0);
 }
 
 // This class is used to relocate addresses in areas that have been loaded.
@@ -673,46 +744,6 @@ void LoadRelocate::RelocateObject(PolyObject *p)
         // N.B. This does not deal with constants within the code.  These have
         // to be handled by real relocation entries.
     }
-    else if (p->IsStackObject())
-    {
-        StackObject *s = (StackObject*)p;
-        POLYUNSIGNED length = p->Length();
-
-        ASSERT(! p->IsMutable()); // Should have been frozen
-        /* First the standard registers, space, pc, sp, hr. */
-        // pc may be TAGGED(0) indicating a retry.
-        PolyWord pc = PolyWord::FromCodePtr(s->p_pc);
-        if (pc != TAGGED(0))
-        {
-            RelocateAddressAt(&pc);
-            s->p_pc = pc.AsCodePtr();
-        }
-
-        PolyWord *stackPtr = s->p_sp; // Save this before we change it.
-        // These point within the current stack.
-        PolyWord sp = PolyWord::FromStackAddr(s->p_sp);
-        RelocateAddressAt(&sp);
-        s->p_sp = sp.AsStackAddr();
-        PolyWord hr = PolyWord::FromStackAddr(s->p_hr);
-        RelocateAddressAt(&hr);
-        s->p_hr = hr.AsStackAddr();
-
-        /* Checked registers. */
-        PolyWord *stackStart = (PolyWord*)p;
-        PolyWord *stackEnd = stackStart+length;
-        for (POLYUNSIGNED i = 0; i < s->p_nreg; i++)
-        {
-            PolyWord r = s->p_reg[i];
-            if (r.AsStackAddr() >= stackStart && r.AsStackAddr() < stackEnd)
-                RelocateAddressAt(&s->p_reg[i]);
-            /* It seems we can have zeros in the registers, at least on the i386. */
-            else if (r == PolyWord::FromUnsigned(0)) {}
-            else RelocateAddressAt(&(s->p_reg[i]));
-        }
-        /* Now the values on the stack. */
-        for (PolyWord *q = stackPtr; q < stackEnd; q++)
-            RelocateAddressAt(q);
-    }
     else /* Ordinary objects, essentially tuples. */
     {
         POLYUNSIGNED length = p->Length();
@@ -721,7 +752,7 @@ void LoadRelocate::RelocateObject(PolyObject *p)
 }
 
 // Load a saved state file.  Calls itself to handle parent files.
-bool StateLoader::LoadFile()
+bool StateLoader::LoadFile(bool isInitial, time_t requiredStamp)
 {
     LoadRelocate relocate;
     AutoFree<char*> thisFile(strdup(fileName));
@@ -754,11 +785,20 @@ bool StateLoader::LoadFile()
         return false;
     }
 
+    // Check that we have the required stamp before loading any children.
+    // If a parent has been overwritten we could get a loop.
+    if (! isInitial && header.timeStamp != requiredStamp)
+    {
+        // Time-stamps don't match.
+        errorResult = "The parent for this saved state does not match or has been changed";
+        return false;
+    }
+
     // Have verified that this is a reasonable saved state file.  If it isn't a
     // top-level file we have to load the parents first.
     if (header.parentNameEntry != 0)
     {
-        unsigned toRead = header.stringTableSize-header.parentNameEntry;
+        size_t toRead = header.stringTableSize-header.parentNameEntry;
         if (MAXPATHLEN < toRead) toRead = MAXPATHLEN;
 
         if (header.parentNameEntry >= header.stringTableSize /* Bad entry */ ||
@@ -769,19 +809,12 @@ bool StateLoader::LoadFile()
             return false;
         }
         fileName[toRead] = 0; // Should already be null-terminated, but just in case.
-        if (! LoadFile())
+
+        if (! LoadFile(false, header.parentTimeStamp))
             return false;
 
         // Check the parent time stamp.
         ASSERT(hierarchyDepth > 0 && hierarchyTable[hierarchyDepth-1] != 0);
-
-        if (header.parentTimeStamp != hierarchyTable[hierarchyDepth-1]->timeStamp)
-        {
-            // Time-stamps don't match.
-            errorResult = "The parent for this saved state does not match or has been changed";
-            return false;
-        }
-
     }
     else // Top-level file
     {
@@ -864,9 +897,11 @@ bool StateLoader::LoadFile()
             gMem.FillUnusedSpace(mem+descr->segmentSize/sizeof(PolyWord),
                 (actualSize-descr->segmentSize)/sizeof(PolyWord));
             // At the moment we leave all segments with write access.
-            space = gMem.NewPermanentSpace(mem, actualSize / sizeof(PolyWord),
-                        (descr->segmentFlags & SSF_WRITABLE) != 0,
-                        (descr->segmentFlags & SSF_NOOVERWRITE) != 0,
+            unsigned mFlags =
+                (descr->segmentFlags & SSF_WRITABLE ? MTF_WRITEABLE : 0) |
+                (descr->segmentFlags & SSF_NOOVERWRITE ? MTF_NO_OVERWRITE : 0) |
+                (descr->segmentFlags & SSF_BYTES ? MTF_BYTES : 0);
+            space = gMem.NewPermanentSpace(mem, actualSize / sizeof(PolyWord), mFlags,
                         descr->segmentIndex, hierarchyDepth+1);
         }
     }
@@ -1101,7 +1136,7 @@ Handle ShowParent(TaskData *taskData, Handle hFileName)
     if (header.parentNameEntry != 0)
     {
         char parentFileName[MAXPATHLEN+1];
-        unsigned toRead = header.stringTableSize-header.parentNameEntry;
+        size_t toRead = header.stringTableSize-header.parentNameEntry;
         if (MAXPATHLEN < toRead) toRead = MAXPATHLEN;
 
         if (header.parentNameEntry >= header.stringTableSize /* Bad entry */ ||
