@@ -1,7 +1,7 @@
 /*
     Title:      Address scanner
 
-    Copyright (c) 2006-8 David C.J. Matthews
+    Copyright (c) 2006-8, 2012 David C.J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -18,10 +18,12 @@
     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
 */
-#ifdef WIN32
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#elif defined(_WIN32)
 #include "winconfig.h"
 #else
-#include "config.h"
+#error "No configuration file"
 #endif
 
 #ifdef HAVE_ASSERT_H
@@ -31,16 +33,20 @@
 #define ASSERT(x) 0
 #endif
 
+#include <new>
+
 #include "globals.h"
 #include "scanaddrs.h"
 #include "machine_dep.h"
 #include "check_objects.h"
 #include "diagnostics.h"
+#include "memmgr.h"
 
 // Process the value at a given location and update it as necessary.
 POLYUNSIGNED ScanAddress::ScanAddressAt(PolyWord *pt)
 {
     PolyWord val = *pt;
+    PolyWord newVal = val;
     if (IS_INT(val) || val == PolyWord::FromUnsigned(0))
     {
         // We can get zeros in the constant area if we garbage collect
@@ -59,25 +65,25 @@ POLYUNSIGNED ScanAddress::ScanAddressAt(PolyWord *pt)
         // Mustn't use ScanAddressAt here.  That's only valid if the value points
         // into the area being updated.
         PolyObject *newObject = ScanObjectAddress(oldObject);
-        *pt = PolyWord::FromCodePtr((byte*)newObject + offset);
+        newVal = PolyWord::FromCodePtr((byte*)newObject + offset);
     }
     else
     {
         ASSERT(OBJ_IS_DATAPTR(val));
         // Database pointer, local pointer or IO pointer.
         // We need to include IO area pointers when we produce an object module.
-        *pt = ScanObjectAddress(val.AsObjPtr());
+        newVal = ScanObjectAddress(val.AsObjPtr());
     }
+    if (newVal != val) // Only update if we need to.
+        *pt = newVal;
     return 0;
 }
 
 // Process a value within the stack.
-PolyWord ScanAddress::ScanStackAddress(PolyWord val, StackObject *base, bool isCode)
+PolyWord ScanAddress::ScanStackAddress(PolyWord val, StackSpace *stack, bool isCode)
 {
-    PolyWord *end = (PolyWord*)base + base->Length();
-
-    ASSERT(base->ContainsNormalLengthWord());
-    ASSERT(base->IsStackObject());
+    PolyWord *base = stack->bottom;
+    PolyWord *end = stack->top;
 
     // If isCode is set we definitely have a code address.  It may have the
     // bottom bit set or it may be word aligned.
@@ -91,11 +97,11 @@ PolyWord ScanAddress::ScanStackAddress(PolyWord val, StackObject *base, bool isC
         return PolyWord::FromCodePtr((byte*)newObject + offset);
     }
 
-    else if (val.IsTagged() || (val.AsAddress() > base && val.AsAddress() <= end))
+    else if (val.IsTagged() || val == PolyWord::FromUnsigned(0) || 
+                 (val.AsAddress() > base && val.AsAddress() <= end))
             /* We don't need to process tagged integers (now we've checked it isn't
                a code address) and we don't need to process addresses within the
-               current stack.  If the containing stack has moved the stack addresses
-               will have been updated by CopyStack. */
+               current stack. */
             /* N.B. We have "<= end" rather than "< end" because it is possible for
                the stack to be completely empty on a terminated thread. */
            return val;
@@ -115,40 +121,13 @@ void ScanAddress::ScanAddressesInObject(PolyObject *obj, POLYUNSIGNED lengthWord
     {
         ASSERT (OBJ_IS_LENGTH(lengthWord));
     
-        CheckObjectL (obj, lengthWord);
-    
         if (OBJ_IS_BYTE_OBJECT(lengthWord))
             return; /* Nothing more to do */
     
         POLYUNSIGNED length = OBJ_OBJECT_LENGTH(lengthWord);
         PolyWord *baseAddr = (PolyWord*)obj;
     
-        if (OBJ_IS_STACK_OBJECT(lengthWord))
-        {
-            StackObject *stack = (StackObject *) obj;
-            PolyWord *stackPtr = stack->p_sp; // Save this BEFORE we update
-            PolyWord *stackEnd = (PolyWord*)obj + length;
-        
-            // Either this is TAGGED(0) indicating a retry or it's a code pointer.
-            if (stack->p_pc != TAGGED(0).AsCodePtr())
-                stack->p_pc = ScanStackAddress (PolyWord::FromCodePtr(stack->p_pc), stack, true).AsCodePtr();
-
-            // Stack pointer and handler pointers
-            stack->p_sp =
-                ScanStackAddress (PolyWord::FromStackAddr(stack->p_sp), stack, false).AsStackAddr();
-            stack->p_hr =
-                ScanStackAddress (PolyWord::FromStackAddr(stack->p_hr), stack, false).AsStackAddr();
-
-            // The checked registers.
-            for (POLYUNSIGNED i = 0; i < stack->p_nreg; i++)
-                stack->p_reg[i] = ScanStackAddress(stack->p_reg[i], stack, false);
-
-            // Now the values on the stack.
-            for (PolyWord *q = stackPtr; q < stackEnd; q++)
-                *q = ScanStackAddress(*q,stack, false);
-            return; // We're done.
-        }
-        else if (OBJ_IS_CODE_OBJECT(lengthWord))
+        if (OBJ_IS_CODE_OBJECT(lengthWord))
         {
             // Scan constants within the code.
             machineDependent->ScanConstantsWithinCode(obj, obj, length, this);
@@ -160,7 +139,7 @@ void ScanAddress::ScanAddressesInObject(PolyObject *obj, POLYUNSIGNED lengthWord
 
         PolyWord *endWord = baseAddr + length;
 
-        // We want to minimise the actual recursion we preform so we try to
+        // We want to minimise the actual recursion we perform so we try to
         // use tail recursion if we can.  We first scan from the end and
         // remove any words that don't need recursion.
         POLYUNSIGNED lastLengthWord = 0;
@@ -220,6 +199,33 @@ void ScanAddress::ScanAddressesInObject(PolyObject *obj, POLYUNSIGNED lengthWord
     } while(1);
 }
 
+// Scan addresses in a stack space.
+void ScanAddress::ScanAddressesInStack(StackSpace *stackSpace)
+{
+    StackObject *stack = stackSpace->stack();
+    PolyWord *stackPtr = stack->p_sp; // Save this BEFORE we update
+    PolyWord *stackEnd = stackSpace->top;
+
+    // Either this is TAGGED(0) indicating a retry or it's a code pointer.
+    if (stack->p_pc != TAGGED(0).AsCodePtr())
+        stack->p_pc = ScanStackAddress (PolyWord::FromCodePtr(stack->p_pc), stackSpace, true).AsCodePtr();
+
+    // Stack pointer and handler pointers
+    stack->p_sp =
+        ScanStackAddress (PolyWord::FromStackAddr(stack->p_sp), stackSpace, false).AsStackAddr();
+    stack->p_hr =
+        ScanStackAddress (PolyWord::FromStackAddr(stack->p_hr), stackSpace, false).AsStackAddr();
+
+    // The checked registers.
+    for (POLYUNSIGNED i = 0; i < stack->p_nreg; i++)
+        stack->p_reg[i] = ScanStackAddress(stack->p_reg[i], stackSpace, false);
+
+    // Now the values on the stack.
+    for (PolyWord *q = stackPtr; q < stackEnd; q++)
+        *q = ScanStackAddress(*q, stackSpace, false);
+}
+
+
 void ScanAddress::ScanAddressesInRegion(PolyWord *region, PolyWord *end)
 {
     PolyWord *pt = region;
@@ -263,7 +269,7 @@ PolyWord ScanAddress::GetConstantValue(byte *addressOfConstant, ScanRelocationKi
             POLYUNSIGNED valu;
             unsigned i;
             byte *pt = addressOfConstant;
-            if (*pt & 0x80) valu = 0-1; else valu = 0;
+            if (pt[3] & 0x80) valu = 0-1; else valu = 0;
             for (i = sizeof(PolyWord); i > 0; i--)
                 valu = (valu << 8) | pt[i-1];
 
@@ -276,15 +282,15 @@ PolyWord ScanAddress::GetConstantValue(byte *addressOfConstant, ScanRelocationKi
 
             return PolyWord::FromUnsigned(valu);
         }
-    case PROCESS_RELOC_I386RELATIVE:         // 32 or 64 bit relative address
+    case PROCESS_RELOC_I386RELATIVE:         // 32 bit relative address
         {
             POLYSIGNED disp;
             byte *pt = addressOfConstant;
             // Get the displacement. This is signed.
-            if (*pt & 0x80) disp = -1; else disp = 0; // Set the sign just in case.
-            for(unsigned i = sizeof(PolyWord); i > 0; i--) disp = (disp << 8) | pt[i-1];
+            if (pt[3] & 0x80) disp = -1; else disp = 0; // Set the sign just in case.
+            for(unsigned i = 4; i > 0; i--) disp = (disp << 8) | pt[i-1];
 
-            byte *absAddr = pt + disp + sizeof(PolyWord); // The address is relative to AFTER the constant
+            byte *absAddr = pt + disp + 4; // The address is relative to AFTER the constant
 
             return PolyWord::FromCodePtr(absAddr);
         }
@@ -340,10 +346,10 @@ void ScanAddress::SetConstantValue(byte *addressOfConstant, PolyWord p, ScanRelo
             }
         }
         break;
-    case PROCESS_RELOC_I386RELATIVE:         // 32 or 64 bit relative address
+    case PROCESS_RELOC_I386RELATIVE:         // 32 bit relative address
         {
-            POLYSIGNED newDisp = p.AsCodePtr() - addressOfConstant - sizeof(PolyWord);
-            for (unsigned i = 0; i < sizeof(PolyWord); i++) {
+            POLYSIGNED newDisp = p.AsCodePtr() - addressOfConstant - 4;
+            for (unsigned i = 0; i < 4; i++) {
                 addressOfConstant[i] = (byte)(newDisp & 0xff);
                 newDisp >>= 8;
             }
@@ -396,3 +402,244 @@ void ScanAddress::ScanConstant(byte *addressOfConstant, ScanRelocationKind code)
     }
 }
 
+void ScanAddress::ScanRuntimeWord(PolyWord *w)
+{
+    if (w->IsTagged()) {} // Don't need to do anything
+    else if (w->IsCodePtr())
+    {
+        // We can have code pointers in set_code_address.
+        // Find the start of the code segment
+        PolyObject *obj = ObjCodePtrToPtr(w->AsCodePtr());
+        // Calculate the byte offset of this value within the code object.
+        POLYUNSIGNED offset = w->AsCodePtr() - (byte*)obj;
+        obj = ScanObjectAddress(obj); 
+        *w = PolyWord::FromCodePtr((byte*)obj + offset);
+
+    }
+    else {
+        ASSERT(w->IsDataPtr());
+        *w = ScanObjectAddress(w->AsObjPtr()); 
+    }
+}
+
+// This gets called in two circumstances.  It may be called for the roots
+// in which case the stack will be empty and we want to process it completely
+// or it is called for a constant address in which case it will have been
+// called from RecursiveScan::ScanAddressesInObject and that can process
+// any addresses.
+PolyObject *RecursiveScan::ScanObjectAddress(PolyObject *obj)
+{
+    PolyWord pWord = obj;
+    // Test to see if this needs to be scanned.
+    // It may update the word.
+    bool test = TestForScan(&pWord);
+    obj = pWord.AsObjPtr();
+
+    if (test)
+    {
+        MarkAsScanning(obj);
+        if (obj->IsByteObject())
+            Completed(obj); // Don't need to put it on the stack
+        // If we already have something on the stack we must being called
+        // recursively to process a constant in a code segment.  Just push
+        // it on the stack and let the caller deal with it.
+        else if (StackIsEmpty())
+            RecursiveScan::ScanAddressesInObject(obj, obj->LengthWord());
+        else
+            PushToStack(obj);
+    }
+
+    return obj;
+}
+
+// This is called via ScanAddressesInRegion to process the permanent mutables.  It is
+// also called from ScanObjectAddress to process root addresses.
+// It processes all the addresses reachable from the object.
+void RecursiveScan::ScanAddressesInObject(PolyObject *obj, POLYUNSIGNED lengthWord)
+{
+    if (OBJ_IS_BYTE_OBJECT(lengthWord))
+    {
+        Completed(obj);
+        return;
+    }
+
+    while (true)
+    {
+        ASSERT (OBJ_IS_LENGTH(lengthWord));
+
+        // Get the length and base address.  N.B.  If this is a code segment
+        // these will be side-effected by GetConstSegmentForCode.
+        POLYUNSIGNED length = OBJ_OBJECT_LENGTH(lengthWord);
+        PolyWord *baseAddr = (PolyWord*)obj;
+
+        if (OBJ_IS_CODE_OBJECT(lengthWord))
+        {
+            // It's better to process the whole code object in one go.
+            ScanAddress::ScanAddressesInObject(obj, lengthWord);
+            length = 0; // Finished
+        }
+        ASSERT(! OBJ_IS_BYTE_OBJECT(lengthWord)); // Check - remove this later
+
+        // else it's a normal object,
+
+        // If there are only two addresses in this cell that need to be
+        // followed we follow them immediately and treat this cell as done.
+        // If there are more than two we push the address of this cell on
+        // the stack, follow the first address and then rescan it.  That way
+        // list cells are processed once only but we don't overflow the
+        // stack by pushing all the addresses in a very large vector.
+        PolyWord *endWord = baseAddr + length;
+        PolyObject *firstWord = 0;
+        PolyObject *secondWord = 0;
+
+        while (baseAddr != endWord)
+        {
+            PolyWord wordAt = *baseAddr;
+
+            if (wordAt.IsDataPtr() && wordAt != PolyWord::FromUnsigned(0))
+            {
+                // Normal address.  We can have words of all zeros at least in the
+                // situation where we have a partially constructed code segment where
+                // the constants at the end of the code have not yet been filled in.
+                if (TestForScan(baseAddr)) // Test value at baseAddr (may side-effect it)
+                {
+                    PolyObject *wObj = (*baseAddr).AsObjPtr();
+                    if (wObj->IsByteObject())
+                    {
+                        // Can do this now - don't need to push it
+                        MarkAsScanning(wObj);
+                        Completed(wObj);
+                    }
+                    else if (firstWord == 0)
+                    {
+                        firstWord = wObj;
+                        // We mark the word immediately.  We can have
+                        // two words in an object that are the same
+                        // and we don't want to process it again.
+                        MarkAsScanning(firstWord);
+                    }
+                    else if (secondWord == 0)
+                        secondWord = wObj;
+                    else break;  // More than two words.
+                }
+            }
+            else if (wordAt.IsCodePtr())
+            {
+                // If we're processing the constant area of a code segment this could
+                // be a code address.
+                PolyObject *oldObject = ObjCodePtrToPtr(wordAt.AsCodePtr());
+                // Calculate the byte offset of this value within the code object.
+                POLYUNSIGNED offset = wordAt.AsCodePtr() - (byte*)oldObject;
+                wordAt = oldObject;
+                bool test = TestForScan(&wordAt);
+                // TestForScan may side-effect the word.
+                PolyObject *newObject = wordAt.AsObjPtr();
+                wordAt = PolyWord::FromCodePtr((byte*)newObject + offset);
+                if (wordAt != *baseAddr)
+                    *baseAddr = wordAt;
+                if (test)
+                {
+                    if (firstWord == 0)
+                    {
+                        firstWord = newObject;
+                        MarkAsScanning(firstWord);
+                    }
+                    else if (secondWord == 0)
+                        secondWord = newObject;
+                    else break;
+                }
+            }
+            baseAddr++;
+        }
+
+        if (baseAddr == endWord)
+        {
+            // We have done everything except possibly firstWord and secondWord.
+            Completed(obj);
+            if (secondWord != 0)
+            {
+                MarkAsScanning(secondWord);
+                // Put this on the stack.  If this is a list node we will be
+                // pushing the tail.
+                PushToStack(secondWord);
+            }
+        }
+        else // Put this back on the stack while we process the first word
+            PushToStack(obj);
+
+        if (firstWord != 0)
+            // Process it immediately.
+            obj = firstWord;
+        else if (StackIsEmpty())
+            return;
+        else
+            obj = PopFromStack();
+
+        lengthWord = obj->LengthWord();
+    }
+}
+
+// The stack is allocated as a series of blocks chained together.
+#define RSTACK_SEGMENT_SIZE 1000
+
+class RScanStack {
+public:
+    RScanStack(): nextStack(0), lastStack(0), sp(0) {}
+    ~RScanStack() { delete(nextStack); }
+
+    RScanStack *nextStack;
+    RScanStack *lastStack;
+    unsigned sp;
+    PolyObject * stack[RSTACK_SEGMENT_SIZE];
+};
+
+RecursiveScanWithStack::~RecursiveScanWithStack()
+{
+    delete(stack);
+}
+
+bool RecursiveScanWithStack::StackIsEmpty(void)
+{
+    return stack == 0 || (stack->sp == 0 && stack->lastStack == 0);
+}
+
+void RecursiveScanWithStack::PushToStack(PolyObject *obj)
+{
+    if (stack == 0 || stack->sp == RSTACK_SEGMENT_SIZE)
+    {
+        if (stack != 0 && stack->nextStack != 0)
+            stack = stack->nextStack;
+        else
+        {
+            // Need a new segment
+            try {
+                RScanStack *s = new RScanStack;
+                s->lastStack = stack;
+                if (stack != 0)
+                    stack->nextStack = s;
+                stack = s;
+            }
+            catch (std::bad_alloc a) {
+                StackOverflow();
+                return;
+            }
+        }
+    }
+    stack->stack[stack->sp++] = obj;
+}
+
+PolyObject *RecursiveScanWithStack::PopFromStack(void)
+{
+    if (stack->sp == 0)
+    {
+        // Chain to the previous stack if any
+        if (stack->lastStack == 0)
+            return 0;
+        // Before we do, delete any further one to free some memory
+        delete(stack->nextStack);
+        stack->nextStack = 0;
+        stack = stack->lastStack;
+        ASSERT(stack->sp == RSTACK_SEGMENT_SIZE);
+    }
+    return stack->stack[--stack->sp];
+}
