@@ -22,10 +22,12 @@
 */
 
 
-#ifdef WIN32
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#elif defined(_WIN32)
 #include "winconfig.h"
 #else
-#include "config.h"
+#error "No configuration file"
 #endif
 
 #ifdef HAVE_STDIO_H
@@ -94,6 +96,8 @@
 
 #ifdef HAVE_SYS_TERMIOS_H
 #include <sys/termios.h>
+#elif (defined(HAVE_TERMIOS_H))
+#include <termios.h>
 #endif
 
 #ifdef HAVE_SYS_STAT_H
@@ -146,7 +150,7 @@ static int unixConstVec[] =
 #endif
     EBUSY,
 #ifdef ECANCELED
-/* This is not defined in Linux.  Perhaps somone knows how to spell "cancelled". */
+/* This is not defined in Linux.  Perhaps someone knows how to spell "cancelled". */
     ECANCELED,
 #else
     0, /* Perhaps some other value. */
@@ -577,7 +581,7 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
         {
             /* This never returns.  When a signal is handled it will
                be interrupted. */
-            processes->BlockAndRestart(taskData, -1, true /* Interruptable. */, POLY_SYS_os_specific);
+            processes->BlockAndRestart(taskData, NULL, true /* Interruptable. */, POLY_SYS_os_specific);
         }
 
     case 22: /* Sleep until given time or until a signal.  Note: this is called
@@ -600,7 +604,9 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
                by a signal. */
             if ((unsigned long)tv.tv_sec < secs ||
                 ((unsigned long)tv.tv_sec == secs && (unsigned long)tv.tv_usec < usecs))
-                processes->BlockAndRestart(taskData, -1, true /* Interruptable. */, POLY_SYS_os_specific);
+                processes->BlockAndRestart(taskData, NULL, true /* Interruptable. */, POLY_SYS_os_specific);
+            else processes->TestAnyEvents(taskData); // Check for interrupts anyway
+
             return Make_arbitrary_precision(taskData, 0);
         }
     
@@ -1220,6 +1226,9 @@ Handle waitForProcess(TaskData *taskData, Handle args)
 /* Get result status of a child process. */
 {
 TryAgain:
+    // We should check for interrupts even if we're not going to block.
+    processes->TestAnyEvents(taskData);
+
     int kind = get_C_long(taskData, DEREFHANDLE(args)->Get(0));
     int pid = get_C_long(taskData, DEREFHANDLE(args)->Get(1));
     int callFlags = get_C_long(taskData, DEREFHANDLE(args)->Get(2));
@@ -1252,7 +1261,7 @@ TryAgain:
        wasn't a child process waiting we have to block
        and come back here later. */
     if (pres == 0 && !(callFlags & WNOHANG))
-        processes->BlockAndRestart(taskData, -1, false, POLY_SYS_os_specific);
+        processes->BlockAndRestart(taskData, NULL, false, POLY_SYS_os_specific);
 
     /* Construct the result tuple. */
     {
