@@ -18,10 +18,12 @@
     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
 */
-#ifdef WIN32
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#elif defined(_WIN32)
 #include "winconfig.h"
 #else
-#include "config.h"
+#error "No configuration file"
 #endif
 
 #ifdef HAVE_STDIO_H
@@ -105,7 +107,7 @@ typedef struct {
 
         struct {
             /* Process and IO channels. */
-            HANDLE hProcess, hInput, hOutput;
+            HANDLE hProcess, hInput, hOutput, hEvent;
             PolyObject *readToken, *writeToken;
         } process;
         HCONV hcDDEConv; /* DDE Conversation. */
@@ -133,6 +135,8 @@ static void close_handle(PHANDLETAB pTab)
             CloseHandle(pTab->entry.process.hInput);
         if (pTab->entry.process.hOutput != INVALID_HANDLE_VALUE)
             CloseHandle(pTab->entry.process.hOutput);
+        if (pTab->entry.process.hEvent)
+            CloseHandle(pTab->entry.process.hEvent);
         break;
 
     case HE_DDECONVERSATION:
@@ -160,7 +164,7 @@ static PHANDLETAB get_handle(PolyWord token, HANDENTRYTYPE heType)
 
 static Handle make_handle_entry(TaskData *taskData)
 {
-    POLYUNSIGNED handle_no;
+    unsigned handle_no;
     Handle str_token;
     bool have_collected = false;
 
@@ -180,7 +184,7 @@ static Handle make_handle_entry(TaskData *taskData)
             }
             else /* No space - expand vector. */
             {
-                int oldMax = maxHandleTab;
+                POLYUNSIGNED oldMax = maxHandleTab;
                 maxHandleTab += maxHandleTab/2;
                 handleTable =
                     (PHANDLETAB)realloc(handleTable,
@@ -213,7 +217,7 @@ static Handle deleteRegistryValue(TaskData *taskData, Handle args, HKEY hkParent
 static Handle enumerateRegistry(TaskData *taskData, Handle args, HKEY hkey, BOOL isKey);
 
 // Vector of constants returned by call1006
-static int winConstVec[] =
+static POLYUNSIGNED winConstVec[] =
 {
     KEY_ALL_ACCESS, // 0
     KEY_CREATE_LINK,
@@ -294,7 +298,7 @@ HKEY hkPredefinedKeyTab[] =
 
 Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
 {
-    int c = get_C_long(taskData, DEREFWORD(code));
+    int c = get_C_int(taskData, DEREFWORD(code));
     switch (c)
     {
     case 0: /* Return our OS type.  Not in any structure. */
@@ -319,7 +323,6 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
     case 1005: /* Get result of process. */
         {
             PHANDLETAB hnd = get_handle(DEREFWORD(args), HE_PROCESS);
-            DWORD dwResult;
             if (hnd == 0)
                 raise_syscall(taskData, "Process is closed", EINVAL);
             // Close the streams. Either of them may have been
@@ -327,6 +330,9 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
             if (hnd->entry.process.hInput != INVALID_HANDLE_VALUE)
                 CloseHandle(hnd->entry.process.hInput);
             hnd->entry.process.hInput = INVALID_HANDLE_VALUE;
+            if (hnd->entry.process.hEvent)
+                CloseHandle(hnd->entry.process.hEvent);
+            hnd->entry.process.hEvent = NULL;
             if (hnd->entry.process.readToken)
             {
                 PIOSTRUCT strm =
@@ -344,27 +350,31 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
                 if (strm != NULL) close_stream(strm);
             }
             hnd->entry.process.writeToken = 0;
+
             // See if it's finished.
-            if (GetExitCodeProcess(hnd->entry.process.hProcess,
-                    &dwResult) == 0)
-                raise_syscall(taskData, "GetExitCodeProcess failed",
-                        -(int)GetLastError());
-            if (dwResult == STILL_ACTIVE)
-                /* Run some other ML processes and come back in at
-                   the top some time later*/
-                processes->BlockAndRestart(taskData, -1, false, POLY_SYS_os_specific);
-            /* Finished - return the result. */
-            /* Note: we haven't closed the handle because we might want to ask
-               for the result again.  We only close it when we've garbage-collected
-               the token.  Doing this runs the risk of running out of handles.
-               Maybe change it and remember the result in ML. */
-            return Make_unsigned(taskData, dwResult);
+            while (true) {
+                DWORD dwResult;
+                if (GetExitCodeProcess(hnd->entry.process.hProcess, &dwResult) == 0)
+                    raise_syscall(taskData, "GetExitCodeProcess failed",
+                            -(int)GetLastError());
+                if (dwResult != STILL_ACTIVE) {
+                    /* Finished - return the result. */
+                    /* Note: we haven't closed the handle because we might want to ask
+                       for the result again.  We only close it when we've garbage-collected
+                       the token.  Doing this runs the risk of running out of handles.
+                       Maybe change it and remember the result in ML. */
+                    return Make_unsigned(taskData, dwResult);
+                }
+                // Block and try again.
+                WaitHandle waiter(hnd->entry.process.hProcess);
+                processes->ThreadPauseForIO(taskData, &waiter);
+            }
         }
 
     case 1006: /* Return a constant. */
         {
-            int i = get_C_long(taskData, DEREFWORD(args));
-            if (i < 0 || i >= (int)(sizeof(winConstVec)/sizeof(winConstVec[0])))
+            unsigned i = get_C_unsigned(taskData, DEREFWORD(args));
+            if (i >= sizeof(winConstVec)/sizeof(winConstVec[0]))
                 raise_syscall(taskData, "Invalid index", 0);
             return Make_unsigned(taskData, winConstVec[i]);
         }
@@ -372,7 +382,7 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
         /* Registry functions. */
     case 1007: // Open a key within one of the roots.
         {
-            unsigned keyIndex = get_C_ulong(taskData, DEREFWORDHANDLE(args)->Get(0));
+            unsigned keyIndex = get_C_unsigned(taskData, DEREFWORDHANDLE(args)->Get(0));
             // This should only ever happen as a result of a fault in
             // the Windows structure.
             if (keyIndex >= sizeof(hkPredefinedKeyTab)/sizeof(hkPredefinedKeyTab[0]))
@@ -391,7 +401,7 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
 
     case 1009: // Create a subkey within one of the roots.
         {
-            unsigned keyIndex = get_C_ulong(taskData, DEREFWORDHANDLE(args)->Get(0));
+            unsigned keyIndex = get_C_unsigned(taskData, DEREFWORDHANDLE(args)->Get(0));
             // This should only ever happen as a result of a fault in
             // the Windows structure.
             if (keyIndex >= sizeof(hkPredefinedKeyTab)/sizeof(hkPredefinedKeyTab[0]))
@@ -417,7 +427,7 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
 
     case 1012: // Get a value
         {
-            unsigned keyIndex = get_C_ulong(taskData, DEREFWORDHANDLE(args)->Get(0));
+            unsigned keyIndex = get_C_unsigned(taskData, DEREFWORDHANDLE(args)->Get(0));
             // This should only ever happen as a result of a fault in
             // the Windows structure.
             if (keyIndex >= sizeof(hkPredefinedKeyTab)/sizeof(hkPredefinedKeyTab[0]))
@@ -436,7 +446,7 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
 
     case 1014: // Delete a subkey
         {
-            unsigned keyIndex = get_C_ulong(taskData, DEREFWORDHANDLE(args)->Get(0));
+            unsigned keyIndex = get_C_unsigned(taskData, DEREFWORDHANDLE(args)->Get(0));
             // This should only ever happen as a result of a fault in
             // the Windows structure.
             if (keyIndex >= sizeof(hkPredefinedKeyTab)/sizeof(hkPredefinedKeyTab[0]))
@@ -455,7 +465,7 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
 
     case 1016: // Set a value
         {
-            unsigned keyIndex = get_C_ulong(taskData, DEREFWORDHANDLE(args)->Get(0));
+            unsigned keyIndex = get_C_unsigned(taskData, DEREFWORDHANDLE(args)->Get(0));
             // This should only ever happen as a result of a fault in
             // the Windows structure.
             if (keyIndex >= sizeof(hkPredefinedKeyTab)/sizeof(hkPredefinedKeyTab[0]))
@@ -474,7 +484,7 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
 
     case 1018: // Enumerate a key in the predefined keys
         {
-            unsigned keyIndex = get_C_ulong(taskData, DEREFWORDHANDLE(args)->Get(0));
+            unsigned keyIndex = get_C_unsigned(taskData, DEREFWORDHANDLE(args)->Get(0));
             if (keyIndex >= sizeof(hkPredefinedKeyTab)/sizeof(hkPredefinedKeyTab[0]))
                 raise_syscall(taskData, "Invalid index", 0);
             return enumerateRegistry(taskData, args, hkPredefinedKeyTab[keyIndex], TRUE);
@@ -491,7 +501,7 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
 
     case 1020: // Enumerate a value in the predefined keys
         {
-            unsigned keyIndex = get_C_ulong(taskData, DEREFWORDHANDLE(args)->Get(0));
+            unsigned keyIndex = get_C_unsigned(taskData, DEREFWORDHANDLE(args)->Get(0));
             if (keyIndex >= sizeof(hkPredefinedKeyTab)/sizeof(hkPredefinedKeyTab[0]))
                 raise_syscall(taskData, "Invalid index", 0);
             return enumerateRegistry(taskData, args, hkPredefinedKeyTab[keyIndex], FALSE);
@@ -508,7 +518,7 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
 
     case 1022: // Delete a value
         {
-            unsigned keyIndex = get_C_ulong(taskData, DEREFWORDHANDLE(args)->Get(0));
+            unsigned keyIndex = get_C_unsigned(taskData, DEREFWORDHANDLE(args)->Get(0));
             // This should only ever happen as a result of a fault in
             // the Windows structure.
             if (keyIndex >= sizeof(hkPredefinedKeyTab)/sizeof(hkPredefinedKeyTab[0]))
@@ -526,7 +536,7 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
         }
 
 
-    case 1030: // Convert UTC time values to local time.
+    case 1030: // Convert UTC time values to local time. -- No longer used??
         {
             FILETIME ftUTC, ftLocal;
             /* Get the file time. */
@@ -539,7 +549,7 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
                         ftLocal.dwLowDateTime);
         }
 
-    case 1031: // Convert local time values to UTC.
+    case 1031: // Convert local time values to UTC. -- No longer used??
         {
             FILETIME ftUTC, ftLocal;
             /* Get the file time. */
@@ -558,8 +568,7 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
             DWORD dwVolSerial, dwMaxComponentLen, dwFlags;
             Handle volHandle, sysHandle, serialHandle, maxCompHandle;
             Handle resultHandle;
-            int length = Poly_string_to_C(DEREFWORD(args),
-                            rootName, MAX_PATH);
+            POLYUNSIGNED length = Poly_string_to_C(DEREFWORD(args), rootName, MAX_PATH);
             if (length > MAX_PATH)
                 raise_syscall(taskData, "Root name too long", ENAMETOOLONG);
             
@@ -583,49 +592,48 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
     case 1033:
         {
             char fileName[MAX_PATH], execName[MAX_PATH];
-            int length = Poly_string_to_C(DEREFWORD(args),
-                            fileName, MAX_PATH);
+            POLYUNSIGNED length = Poly_string_to_C(DEREFWORD(args), fileName, MAX_PATH);
             HINSTANCE hInst;
             if (length > MAX_PATH)
                 raise_syscall(taskData, "File name too long", ENAMETOOLONG);
             hInst = FindExecutable(fileName, NULL, execName);
-            if ((unsigned)hInst <= 32)
+            if ((POLYUNSIGNED)hInst <= 32)
             {
-                raise_syscall(taskData, "FindExecutable failed", -(int)hInst);
+               raise_syscall(taskData, "FindExecutable failed", -(int)(POLYUNSIGNED)hInst);
             }
             return SAVE(C_string_to_Poly(taskData, execName));
         }
 
     case 1034: // Open a document
         {
-			SHELLEXECUTEINFO shellEx;
-			memset(&shellEx, 0, sizeof(shellEx));
-			shellEx.cbSize = sizeof(shellEx);
-			shellEx.lpVerb = _T("open");
-			shellEx.lpFile = Poly_string_to_T_alloc(DEREFWORD(args));
-			shellEx.hwnd = hMainWindow;
-			shellEx.nShow = SW_SHOWNORMAL;
+            SHELLEXECUTEINFO shellEx;
+            memset(&shellEx, 0, sizeof(shellEx));
+            shellEx.cbSize = sizeof(shellEx);
+            shellEx.lpVerb = _T("open");
+            shellEx.lpFile = Poly_string_to_T_alloc(DEREFWORD(args));
+            shellEx.hwnd = hMainWindow;
+            shellEx.nShow = SW_SHOWNORMAL;
             BOOL fRes = ShellExecuteEx(&shellEx);
             free((void*)shellEx.lpFile);
-			if (! fRes)
-				raise_syscall(taskData, "ShellExecuteEx failed", 0-GetLastError());
+            if (! fRes)
+                raise_syscall(taskData, "ShellExecuteEx failed", 0-GetLastError());
             return Make_arbitrary_precision(taskData, 0);
         }
 
     case 1035: // Launch an application.
         {
-			SHELLEXECUTEINFO shellEx;
-			memset(&shellEx, 0, sizeof(shellEx));
-			shellEx.cbSize = sizeof(shellEx);
-			shellEx.lpVerb = _T("open");
-			shellEx.lpFile = Poly_string_to_T_alloc(args->WordP()->Get(0));
+            SHELLEXECUTEINFO shellEx;
+            memset(&shellEx, 0, sizeof(shellEx));
+            shellEx.cbSize = sizeof(shellEx);
+            shellEx.lpVerb = _T("open");
+            shellEx.lpFile = Poly_string_to_T_alloc(args->WordP()->Get(0));
             shellEx.lpParameters = Poly_string_to_T_alloc(args->WordP()->Get(1));
-			shellEx.nShow = SW_SHOWNORMAL;
+            shellEx.nShow = SW_SHOWNORMAL;
             BOOL fRes = ShellExecuteEx(&shellEx);
             free((void*)shellEx.lpFile);
             free((void*)shellEx.lpParameters);
-			if (! fRes)
-				raise_syscall(taskData, "ShellExecuteEx failed", 0-GetLastError());
+            if (! fRes)
+                raise_syscall(taskData, "ShellExecuteEx failed", 0-GetLastError());
             return Make_arbitrary_precision(taskData, 0);
         }
 
@@ -749,8 +757,8 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
             while (1)
             {
                 HWND hwnd = (HWND)get_C_long(taskData, DEREFWORDHANDLE(args)->Get(0)); /* Handles are treated as SIGNED. */
-                UINT wMsgFilterMin = get_C_ulong(taskData, DEREFWORDHANDLE(args)->Get(1));
-                UINT wMsgFilterMax = get_C_ulong(taskData, DEREFWORDHANDLE(args)->Get(2));
+                UINT wMsgFilterMin = get_C_unsigned(taskData, DEREFWORDHANDLE(args)->Get(1));
+                UINT wMsgFilterMax = get_C_unsigned(taskData, DEREFWORDHANDLE(args)->Get(2));
                 MSG msg;
                 processes->ThreadReleaseMLMemory(taskData);
                 // N.B.  PeekMessage may directly call the window proc resulting in a
@@ -766,10 +774,10 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
     // case 1102: // Return the address of the window callback function.
 
     case 1103: // Return the application instance.
-        return Make_arbitrary_precision(taskData, (int)hApplicationInstance);
+        return Make_unsigned(taskData, (POLYUNSIGNED)hApplicationInstance);
 
     case 1104: // Return the main window handle
-        return Make_arbitrary_precision(taskData, (int)hMainWindow);
+        return Make_unsigned(taskData, (POLYUNSIGNED)hMainWindow);
 
 //    case 1105: // Set the callback function
 
@@ -778,7 +786,7 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
             char msg[100];
             sprintf(msg, "Unknown windows-specific function: %d", c);
             raise_exception_string(taskData, EXC_Fail, msg);
-			return 0;
+            return 0;
         }
     }
 }
@@ -790,16 +798,25 @@ set up in the parent.  As with Unix we create two pipes and pass one end of each
 pipe to the child.  The end we pass to the child is "inheritable" (i.e. duplicated
 in the child as with Unix file descriptors) while the ends we keep in the parent
 are non-inheritable (i.e. not duplicated in the child). 
-DCJM: December 1999.  
+DCJM: December 1999.
+This is now further complicated to improve the performance.  In Unix we can pass
+the file ID to "select" which will return immediately when input is available (we
+ignore blocking on output at the moment).  That allows the ML process to respond
+immediately.  There's no easy way to do that in Windows since the pipe handle is
+signalled whether there is input available or not.  One possibility would be to
+use overlapped IO but that requires using the ReadFile call directly and some
+contortions to create a pipe with overlapped IO.  The other, taken here, is to
+interpose a thread which can signal an event when input is available. 
 */
 static Handle execute(TaskData *taskData, Handle args)
 {
     char *commandName = 0, *arguments = 0;
-    LPSTR lpszError = "";
+    LPCSTR lpszError = "";
     HANDLE hWriteToChild = INVALID_HANDLE_VALUE,
            hReadFromParent = INVALID_HANDLE_VALUE,
            hWriteToParent = INVALID_HANDLE_VALUE,
            hReadFromChild = INVALID_HANDLE_VALUE;
+    HANDLE hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
     HANDLE hTemp;
     STARTUPINFO startupInfo;
     PROCESS_INFORMATION processInfo;
@@ -823,6 +840,13 @@ static Handle execute(TaskData *taskData, Handle args)
         lpszError = "Could not create pipe";
         goto error;
     }
+    // Create the copying thread.
+    hTemp = CreateCopyPipe(hReadFromChild, hEvent);
+    if (hTemp == NULL) {
+        lpszError = "Could not create pipe";
+        goto error;
+    }
+    hReadFromChild = hTemp;
     // Convert the handles we want to pass to the child into inheritable
     // handles by duplicating and replacing them with the duplicates.
     if (! DuplicateHandle(GetCurrentProcess(), hWriteToParent, GetCurrentProcess(),
@@ -875,6 +899,7 @@ static Handle execute(TaskData *taskData, Handle args)
     pTab->entry.process.hProcess = processInfo.hProcess;
     pTab->entry.process.hInput = hReadFromChild;
     pTab->entry.process.hOutput = hWriteToChild;
+    pTab->entry.process.hEvent = hEvent;
     pTab->entry.process.readToken = 0;
     pTab->entry.process.writeToken = 0;
 
@@ -890,6 +915,7 @@ error:
         if (hReadFromParent != INVALID_HANDLE_VALUE) CloseHandle(hReadFromParent);
         if (hWriteToParent != INVALID_HANDLE_VALUE) CloseHandle(hWriteToParent);
         if (hReadFromChild != INVALID_HANDLE_VALUE) CloseHandle(hReadFromChild);
+        if (hEvent) CloseHandle(hEvent);
         raise_syscall(taskData, lpszError, -err);
         return NULL; // Never reached.
     }
@@ -919,13 +945,13 @@ static Handle simpleExecute(TaskData *taskData, Handle args)
     startupInfo.hStdInput = hNull;
     startupInfo.hStdOutput = hNull;
     startupInfo.hStdError = hNull;
-	STARTUPINFO *start = &startupInfo;
+    STARTUPINFO *start = &startupInfo;
 
     // Try starting the process first using the given name.
     if (!CreateProcess(commandName,
             arguments, // Command line
             NULL, NULL, // Security attributes
-			TRUE, CREATE_NO_WINDOW, // Inherit handles, creation flags
+            TRUE, CREATE_NO_WINDOW, // Inherit handles, creation flags
             NULL, NULL, // Inherit our environment and directory
             start,
             &processInfo)) {
@@ -952,6 +978,7 @@ static Handle simpleExecute(TaskData *taskData, Handle args)
     // We only use the process handle entry.
     pTab->entry.process.hInput = INVALID_HANDLE_VALUE;
     pTab->entry.process.hOutput = INVALID_HANDLE_VALUE;
+    pTab->entry.process.hEvent = NULL;
     pTab->entry.process.readToken = 0;
     pTab->entry.process.writeToken = 0;
 
@@ -969,15 +996,15 @@ static Handle openProcessHandle(TaskData *taskData, Handle args, BOOL fIsRead, B
 
     if (fIsRead) hStream = hnd->entry.process.hInput;
     else hStream = hnd->entry.process.hOutput;
-	/* I had previously assumed that it wasn't possible to get the
-	   same stream twice.  The current basis library definition allows
-	   it but warns it may produce unpredictable results.  For the moment
-	   we don't allow it because we could get problems with closing
-	   the same handle twice. */
+    /* I had previously assumed that it wasn't possible to get the
+       same stream twice.  The current basis library definition allows
+       it but warns it may produce unpredictable results.  For the moment
+       we don't allow it because we could get problems with closing
+       the same handle twice. */
 #ifdef _WIN32_WCE
-	// Not possible in Windows CE.
-	raise_syscall(taskData, "Process is closed", EBADF);
-	return 0;
+    // Not possible in Windows CE.
+    raise_syscall(taskData, "Process is closed", EBADF);
+    return 0;
 #else
     if (hStream == INVALID_HANDLE_VALUE)
         raise_syscall(taskData, "Process is closed", EBADF);
@@ -988,10 +1015,11 @@ static Handle openProcessHandle(TaskData *taskData, Handle args, BOOL fIsRead, B
 
     Handle str_token = make_stream_entry(taskData);
     PIOSTRUCT strm = &basic_io_vector[STREAMID(str_token)];
-    strm->device.ioDesc = _open_osfhandle ((long)hStream, mode);
+    strm->device.ioDesc = _open_osfhandle ((POLYSIGNED) hStream, mode);
     if (strm->device.ioDesc == -1)
         raise_syscall(taskData, "_open_osfhandle failed", errno);
     strm->ioBits = ioBits | IO_BIT_OPEN | IO_BIT_PIPE;
+
     /* The responsibility for closing the handle is passed to
        the stream package.  We need to retain a pointer to the
        stream entry so that we can close the stream in "reap". */
@@ -999,6 +1027,9 @@ static Handle openProcessHandle(TaskData *taskData, Handle args, BOOL fIsRead, B
     {
         hnd->entry.process.hInput = INVALID_HANDLE_VALUE;
         hnd->entry.process.readToken = strm->token;
+        // Pass the "input available" event.
+        strm->hAvailable = hnd->entry.process.hEvent;
+        hnd->entry.process.hEvent = NULL;
     }
     else
     {
@@ -1018,8 +1049,8 @@ static Handle openRegistryKey(TaskData *taskData, Handle args, HKEY hkParent)
     Handle result;
     PHANDLETAB pTab;
     HKEY hkey;
-    REGSAM sam = get_C_ulong(taskData, DEREFWORDHANDLE(args)->Get(2));
-    int length = Poly_string_to_C(args->WordP()->Get(1), keyName, MAX_PATH);
+    REGSAM sam = get_C_unsigned(taskData, DEREFWORDHANDLE(args)->Get(2));
+    POLYUNSIGNED length = Poly_string_to_C(args->WordP()->Get(1), keyName, MAX_PATH);
     if (length > MAX_PATH)
         raise_syscall(taskData, "Key name too long", ENAMETOOLONG);
 
@@ -1045,9 +1076,9 @@ static Handle createRegistryKey(TaskData *taskData, Handle args, HKEY hkParent)
     PHANDLETAB pTab;
     HKEY hkey;
     DWORD dwDisp;
-    REGSAM sam = get_C_ulong(taskData, DEREFWORDHANDLE(args)->Get(3));
-    int opt = get_C_ulong(taskData, DEREFWORDHANDLE(args)->Get(2));
-    int length = Poly_string_to_C(args->WordP()->Get(1), keyName, MAX_PATH);
+    REGSAM sam = get_C_unsigned(taskData, DEREFWORDHANDLE(args)->Get(3));
+    unsigned opt = get_C_unsigned(taskData, DEREFWORDHANDLE(args)->Get(2));
+    POLYUNSIGNED length = Poly_string_to_C(args->WordP()->Get(1), keyName, MAX_PATH);
     if (length > MAX_PATH)
         raise_syscall(taskData, "Key name too long", ENAMETOOLONG);
 
@@ -1078,7 +1109,7 @@ static Handle deleteRegistryKey(TaskData *taskData, Handle args, HKEY hkParent)
 {
     TCHAR keyName[MAX_PATH];
     LONG lRes;
-    int length = Poly_string_to_C(args->WordP()->Get(1), keyName, MAX_PATH);
+    POLYUNSIGNED length = Poly_string_to_C(args->WordP()->Get(1), keyName, MAX_PATH);
     if (length > MAX_PATH)
         raise_syscall(taskData, "Key name too long", ENAMETOOLONG);
 
@@ -1094,7 +1125,7 @@ static Handle deleteRegistryValue(TaskData *taskData, Handle args, HKEY hkParent
 {
     TCHAR keyName[MAX_PATH];
     LONG lRes;
-    int length = Poly_string_to_C(args->WordP()->Get(1).AsObjPtr(),
+    POLYUNSIGNED length = Poly_string_to_C(args->WordP()->Get(1).AsObjPtr(),
                     keyName, MAX_PATH);
     if (length > MAX_PATH)
         raise_syscall(taskData, "Key name too long", ENAMETOOLONG);
@@ -1115,7 +1146,7 @@ static Handle queryRegistryKey(TaskData *taskData, Handle args, HKEY hkey)
     DWORD valSize;
     Handle result, resVal, resType;
     DWORD dwType;
-    int length = Poly_string_to_C(args->WordP()->Get(1), valName, MAX_PATH);
+    POLYUNSIGNED length = Poly_string_to_C(args->WordP()->Get(1), valName, MAX_PATH);
     if (length > MAX_PATH)
         raise_syscall(taskData, "Value name too long", ENAMETOOLONG);
 
@@ -1163,8 +1194,8 @@ static Handle setRegistryKey(TaskData *taskData, Handle args, HKEY hkey)
     TCHAR valName[MAX_PATH];
     LONG lRes;
     PolyWord str = args->WordP()->Get(3);
-    int length = Poly_string_to_C(args->WordP()->Get(1), valName, MAX_PATH);
-    DWORD dwType = get_C_ulong(taskData, DEREFWORDHANDLE(args)->Get(2));
+    POLYUNSIGNED length = Poly_string_to_C(args->WordP()->Get(1), valName, MAX_PATH);
+    DWORD dwType = get_C_unsigned(taskData, DEREFWORDHANDLE(args)->Get(2));
     if (length > MAX_PATH)
         raise_syscall(taskData, "Value name too long", ENAMETOOLONG);
 
@@ -1179,7 +1210,7 @@ static Handle setRegistryKey(TaskData *taskData, Handle args, HKEY hkey)
     {
         PolyStringObject *ps = (PolyStringObject*)str.AsObjPtr();
         lRes = RegSetValueEx(hkey, valName, 0, dwType,
-                            (CONST BYTE *)ps->chars, ps->length);
+                            (CONST BYTE *)ps->chars, (DWORD)ps->length);
     }
 
     if (lRes != ERROR_SUCCESS)
@@ -1192,7 +1223,7 @@ static Handle setRegistryKey(TaskData *taskData, Handle args, HKEY hkey)
 // no key/value could be found or SOME s where s is the name of the key/value.
 static Handle enumerateRegistry(TaskData *taskData, Handle args, HKEY hkey, BOOL isKey)
 {
-    DWORD num = get_C_ulong(taskData, DEREFWORDHANDLE(args)->Get(1));
+    DWORD num = get_C_unsigned(taskData, DEREFWORDHANDLE(args)->Get(1));
     LONG lRes;
     TCHAR keyName[MAX_PATH];
     DWORD dwLength = sizeof(keyName)/sizeof(keyName[0]);
@@ -1222,7 +1253,7 @@ class WindowsModule: public RtsModule
 {
 public:
     virtual void Init(void);
-    virtual void Uninit(void);
+    virtual void Stop(void);
     void GarbageCollect(ScanAddress *process);
 };
 
@@ -1258,7 +1289,7 @@ void WindowsModule::GarbageCollect(ScanAddress *process)
     }
 }
 
-void WindowsModule::Uninit(void)
+void WindowsModule::Stop(void)
 {
     if (handleTable)
     {
